@@ -39,7 +39,8 @@ const array = [
     "os": "ubuntu-latest",
     "name": "Test Linux x86_64",
     "filter": "linux-x64",
-    "isa": "x64"
+    "isa": "x64",
+    "extra_features": "--features wasmtime-wasi-nn/onnx"
   },
   {
     "os": "ubuntu-latest",
@@ -55,20 +56,23 @@ const array = [
     "isa": "x64"
   },
   {
-    "os": "macos-latest",
+    "os": "macos-13",
     "name": "Test macOS x86_64",
-    "filter": "macos-x64"
+    "filter": "macos-x64",
+    "extra_features": "--features wasmtime-wasi-nn/onnx"
   },
   {
     "os": "macos-14",
     "name": "Test macOS arm64",
     "filter": "macos-arm64",
-    "target": "aarch64-apple-darwin"
+    "target": "aarch64-apple-darwin",
+    "extra_features": "--features wasmtime-wasi-nn/onnx"
   },
   {
     "os": "windows-latest",
     "name": "Test Windows MSVC x86_64",
-    "filter": "windows-x64"
+    "filter": "windows-x64",
+    "extra_features": "--features wasmtime-wasi-nn/onnx"
   },
   {
     "os": "windows-latest",
@@ -85,7 +89,7 @@ const array = [
     "qemu_target": "aarch64-linux-user",
     "name": "Test Linux arm64",
     "filter": "linux-arm64",
-    "isa": "aarch64"
+    "isa": "aarch64",
   },
   {
     "os": "ubuntu-latest",
@@ -103,7 +107,7 @@ const array = [
     "target": "riscv64gc-unknown-linux-gnu",
     "gcc_package": "gcc-riscv64-linux-gnu",
     "gcc": "riscv64-linux-gnu-gcc",
-    "qemu": "qemu-riscv64 -cpu rv64,v=true,vlen=256,vext_spec=v1.0,zba=true,zbb=true,zbc=true,zbs=true,zbkb=true,zcb=true -L /usr/riscv64-linux-gnu",
+    "qemu": "qemu-riscv64 -cpu rv64,v=true,vlen=256,vext_spec=v1.0,Zfa=true,zba=true,zbb=true,zbc=true,zbs=true,zbkb=true,zcb=true -L /usr/riscv64-linux-gnu",
     "qemu_target": "riscv64-linux-user",
     "name": "Test Linux riscv64",
     "filter": "linux-riscv64",
