@@ -13,7 +13,7 @@ use crate::{settings, CodegenError, CodegenResult};
 pub use crate::ir::condcodes::FloatCC;
 
 use alloc::vec::Vec;
-use regalloc2::{PRegSet, RegClass, VReg};
+use regalloc2::{PRegSet, RegClass};
 use smallvec::{smallvec, SmallVec};
 use std::boxed::Box;
 use std::fmt::Write;
@@ -46,8 +46,8 @@ pub(crate) type VecU8 = Vec<u8>;
 // Instructions (top level): definition
 
 pub use crate::isa::riscv64::lower::isle::generated_code::{
-    AluOPRRI, AluOPRRR, AtomicOP, CsrImmOP, CsrRegOP, FClassResult, FFlagsException, FloatRoundOP,
-    FpuOPRR, FpuOPRRR, FpuOPRRRR, LoadOP, MInst as Inst, StoreOP, CSR, FRM,
+    AluOPRRI, AluOPRRR, AtomicOP, CsrImmOP, CsrRegOP, FClassResult, FFlagsException, FpuOPRR,
+    FpuOPRRR, FpuOPRRRR, LoadOP, MInst as Inst, StoreOP, CSR, FRM,
 };
 use crate::isa::riscv64::lower::isle::generated_code::{CjOp, MInst, VecAluOpRRImm5, VecAluOpRRR};
 
@@ -85,7 +85,6 @@ pub struct CallIndInfo {
 pub struct ReturnCallInfo {
     pub uses: CallArgList,
     pub opcode: Opcode,
-    pub old_stack_arg_size: u32,
     pub new_stack_arg_size: u32,
 }
 
@@ -318,201 +317,192 @@ impl Inst {
 
 //=============================================================================
 
-fn vec_mask_operands<F: Fn(VReg) -> VReg>(
-    mask: &VecOpMasking,
-    collector: &mut OperandCollector<'_, F>,
-) {
+fn vec_mask_operands(mask: &mut VecOpMasking, collector: &mut impl OperandVisitor) {
     match mask {
         VecOpMasking::Enabled { reg } => {
-            collector.reg_fixed_use(*reg, pv_reg(0).into());
+            collector.reg_fixed_use(reg, pv_reg(0).into());
         }
         VecOpMasking::Disabled => {}
     }
 }
-fn vec_mask_late_operands<F: Fn(VReg) -> VReg>(
-    mask: &VecOpMasking,
-    collector: &mut OperandCollector<'_, F>,
-) {
+fn vec_mask_late_operands(mask: &mut VecOpMasking, collector: &mut impl OperandVisitor) {
     match mask {
         VecOpMasking::Enabled { reg } => {
-            collector.reg_fixed_late_use(*reg, pv_reg(0).into());
+            collector.reg_fixed_late_use(reg, pv_reg(0).into());
         }
         VecOpMasking::Disabled => {}
     }
 }
 
-fn riscv64_get_operands<F: Fn(VReg) -> VReg>(inst: &Inst, collector: &mut OperandCollector<'_, F>) {
+fn riscv64_get_operands(inst: &mut Inst, collector: &mut impl OperandVisitor) {
     match inst {
-        &Inst::Nop0 => {}
-        &Inst::Nop4 => {}
-        &Inst::BrTable {
+        Inst::Nop0 | Inst::Nop4 => {}
+        Inst::BrTable {
             index, tmp1, tmp2, ..
         } => {
             collector.reg_use(index);
             collector.reg_early_def(tmp1);
             collector.reg_early_def(tmp2);
         }
-        &Inst::Auipc { rd, .. } => collector.reg_def(rd),
-        &Inst::Lui { rd, .. } => collector.reg_def(rd),
-        &Inst::LoadInlineConst { rd, .. } => collector.reg_def(rd),
-        &Inst::AluRRR { rd, rs1, rs2, .. } => {
+        Inst::Auipc { rd, .. } => collector.reg_def(rd),
+        Inst::Lui { rd, .. } => collector.reg_def(rd),
+        Inst::LoadInlineConst { rd, .. } => collector.reg_def(rd),
+        Inst::AluRRR { rd, rs1, rs2, .. } => {
             collector.reg_use(rs1);
             collector.reg_use(rs2);
             collector.reg_def(rd);
         }
-        &Inst::FpuRRR { rd, rs1, rs2, .. } => {
+        Inst::FpuRRR { rd, rs1, rs2, .. } => {
             collector.reg_use(rs1);
             collector.reg_use(rs2);
             collector.reg_def(rd);
         }
-        &Inst::AluRRImm12 { rd, rs, .. } => {
+        Inst::AluRRImm12 { rd, rs, .. } => {
             collector.reg_use(rs);
             collector.reg_def(rd);
         }
-        &Inst::CsrReg { rd, rs, .. } => {
+        Inst::CsrReg { rd, rs, .. } => {
             collector.reg_use(rs);
             collector.reg_def(rd);
         }
-        &Inst::CsrImm { rd, .. } => {
+        Inst::CsrImm { rd, .. } => {
             collector.reg_def(rd);
         }
-        &Inst::Load { rd, from, .. } => {
-            if let Some(r) = from.get_allocatable_register() {
-                collector.reg_use(r);
-            }
+        Inst::Load { rd, from, .. } => {
+            from.get_operands(collector);
             collector.reg_def(rd);
         }
-        &Inst::Store { to, src, .. } => {
-            if let Some(r) = to.get_allocatable_register() {
-                collector.reg_use(r);
-            }
+        Inst::Store { to, src, .. } => {
+            to.get_operands(collector);
             collector.reg_use(src);
         }
 
-        &Inst::Args { ref args } => {
-            for arg in args {
-                collector.reg_fixed_def(arg.vreg, arg.preg);
+        Inst::Args { args } => {
+            for ArgPair { vreg, preg } in args {
+                collector.reg_fixed_def(vreg, *preg);
             }
         }
-        &Inst::Rets { ref rets } => {
-            for ret in rets {
-                collector.reg_fixed_use(ret.vreg, ret.preg);
+        Inst::Rets { rets } => {
+            for RetPair { vreg, preg } in rets {
+                collector.reg_fixed_use(vreg, *preg);
             }
         }
-        &Inst::Ret { .. } => {}
+        Inst::Ret { .. } => {}
 
-        &Inst::Extend { rd, rn, .. } => {
+        Inst::Extend { rd, rn, .. } => {
             collector.reg_use(rn);
             collector.reg_def(rd);
         }
-        &Inst::Call { ref info } => {
-            for u in &info.uses {
-                collector.reg_fixed_use(u.vreg, u.preg);
+        Inst::Call { info } => {
+            let CallInfo { uses, defs, .. } = &mut **info;
+            for CallArgPair { vreg, preg } in uses {
+                collector.reg_fixed_use(vreg, *preg);
             }
-            for d in &info.defs {
-                collector.reg_fixed_def(d.vreg, d.preg);
+            for CallRetPair { vreg, preg } in defs {
+                collector.reg_fixed_def(vreg, *preg);
             }
             collector.reg_clobbers(info.clobbers);
         }
-        &Inst::CallInd { ref info } => {
-            if info.callee_callconv == CallConv::Tail {
-                // TODO(https://github.com/bytecodealliance/regalloc2/issues/145):
-                // This shouldn't be a fixed register constraint.
-                collector.reg_fixed_use(info.rn, x_reg(5));
-            } else {
-                collector.reg_use(info.rn);
+        Inst::CallInd { info } => {
+            let CallIndInfo { rn, uses, defs, .. } = &mut **info;
+            collector.reg_use(rn);
+            for CallArgPair { vreg, preg } in uses {
+                collector.reg_fixed_use(vreg, *preg);
             }
+            for CallRetPair { vreg, preg } in defs {
+                collector.reg_fixed_def(vreg, *preg);
+            }
+            collector.reg_clobbers(info.clobbers);
+        }
+        Inst::ReturnCall { info, .. } => {
+            for CallArgPair { vreg, preg } in &mut info.uses {
+                collector.reg_fixed_use(vreg, *preg);
+            }
+        }
+        Inst::ReturnCallInd { info, callee } => {
+            // TODO(https://github.com/bytecodealliance/regalloc2/issues/145):
+            // This shouldn't be a fixed register constraint.
+            collector.reg_fixed_use(callee, x_reg(5));
 
-            for u in &info.uses {
-                collector.reg_fixed_use(u.vreg, u.preg);
-            }
-            for d in &info.defs {
-                collector.reg_fixed_def(d.vreg, d.preg);
-            }
-            collector.reg_clobbers(info.clobbers);
-        }
-        &Inst::ReturnCall {
-            callee: _,
-            ref info,
-        } => {
-            for u in &info.uses {
-                collector.reg_fixed_use(u.vreg, u.preg);
+            for CallArgPair { vreg, preg } in &mut info.uses {
+                collector.reg_fixed_use(vreg, *preg);
             }
         }
-        &Inst::ReturnCallInd { ref info, callee } => {
-            collector.reg_use(callee);
-            for u in &info.uses {
-                collector.reg_fixed_use(u.vreg, u.preg);
-            }
-        }
-        &Inst::Jal { .. } => {
+        Inst::Jal { .. } => {
             // JAL technically has a rd register, but we currently always
             // hardcode it to x0.
         }
-        &Inst::CondBr { kind, .. } => {
-            collector.reg_use(kind.rs1);
-            collector.reg_use(kind.rs2);
+        Inst::CondBr {
+            kind: IntegerCompare { rs1, rs2, .. },
+            ..
+        } => {
+            collector.reg_use(rs1);
+            collector.reg_use(rs2);
         }
-        &Inst::LoadExtName { rd, .. } => {
+        Inst::LoadExtName { rd, .. } => {
             collector.reg_def(rd);
         }
-        &Inst::ElfTlsGetAddr { rd, .. } => {
+        Inst::ElfTlsGetAddr { rd, .. } => {
             // x10 is a0 which is both the first argument and the first return value.
             collector.reg_fixed_def(rd, a0());
             let mut clobbers = Riscv64MachineDeps::get_regs_clobbered_by_call(CallConv::SystemV);
             clobbers.remove(px_reg(10));
             collector.reg_clobbers(clobbers);
         }
-        &Inst::LoadAddr { rd, mem } => {
-            if let Some(r) = mem.get_allocatable_register() {
-                collector.reg_use(r);
-            }
+        Inst::LoadAddr { rd, mem } => {
+            mem.get_operands(collector);
             collector.reg_early_def(rd);
         }
 
-        &Inst::VirtualSPOffsetAdj { .. } => {}
-        &Inst::Mov { rd, rm, .. } => {
+        Inst::VirtualSPOffsetAdj { .. } => {}
+        Inst::Mov { rd, rm, .. } => {
             collector.reg_use(rm);
             collector.reg_def(rd);
         }
-        &Inst::MovFromPReg { rd, rm } => {
-            debug_assert!([px_reg(2), px_reg(8)].contains(&rm));
+        Inst::MovFromPReg { rd, rm } => {
+            debug_assert!([px_reg(2), px_reg(8)].contains(rm));
             collector.reg_def(rd);
         }
-        &Inst::Fence { .. } => {}
-        &Inst::EBreak => {}
-        &Inst::Udf { .. } => {}
-        &Inst::FpuRR { rd, rs, .. } => {
+        Inst::Fence { .. } => {}
+        Inst::EBreak => {}
+        Inst::Udf { .. } => {}
+        Inst::FpuRR { rd, rs, .. } => {
             collector.reg_use(rs);
             collector.reg_def(rd);
         }
-        &Inst::FpuRRRR {
+        Inst::FpuRRRR {
             rd, rs1, rs2, rs3, ..
         } => {
-            collector.reg_uses(&[rs1, rs2, rs3]);
+            collector.reg_use(rs1);
+            collector.reg_use(rs2);
+            collector.reg_use(rs3);
             collector.reg_def(rd);
         }
 
-        &Inst::Jalr { rd, base, .. } => {
+        Inst::Jalr { rd, base, .. } => {
             collector.reg_use(base);
             collector.reg_def(rd);
         }
-        &Inst::Atomic { rd, addr, src, .. } => {
+        Inst::Atomic { rd, addr, src, .. } => {
             collector.reg_use(addr);
             collector.reg_use(src);
             collector.reg_def(rd);
         }
-        &Inst::Select {
-            ref dst,
-            condition,
+        Inst::Select {
+            dst,
+            condition: IntegerCompare { rs1, rs2, .. },
             x,
             y,
             ..
         } => {
-            collector.reg_use(condition.rs1);
-            collector.reg_use(condition.rs2);
-            collector.reg_uses(x.regs());
-            collector.reg_uses(y.regs());
+            collector.reg_use(rs1);
+            collector.reg_use(rs2);
+            for reg in x.regs_mut() {
+                collector.reg_use(reg);
+            }
+            for reg in y.regs_mut() {
+                collector.reg_use(reg);
+            }
             // If there's more than one destination register then use
             // `reg_early_def` to prevent destination registers from overlapping
             // with any operands. This ensures that the lowering doesn't have to
@@ -522,15 +512,16 @@ fn riscv64_get_operands<F: Fn(VReg) -> VReg>(inst: &Inst, collector: &mut Operan
             // When there's only one destination register though don't use an
             // early def because once the register is written no other inputs
             // are read so it's ok for the destination to overlap the sources.
-            if dst.regs().len() > 1 {
-                for d in dst.regs() {
-                    collector.reg_early_def(d.clone());
+            match dst.regs_mut() {
+                [reg] => collector.reg_def(reg),
+                regs => {
+                    for d in regs {
+                        collector.reg_early_def(d);
+                    }
                 }
-            } else {
-                collector.reg_defs(dst.regs());
             }
         }
-        &Inst::AtomicCas {
+        Inst::AtomicCas {
             offset,
             t0,
             dst,
@@ -539,21 +530,24 @@ fn riscv64_get_operands<F: Fn(VReg) -> VReg>(inst: &Inst, collector: &mut Operan
             v,
             ..
         } => {
-            collector.reg_uses(&[offset, e, addr, v]);
+            collector.reg_use(offset);
+            collector.reg_use(e);
+            collector.reg_use(addr);
+            collector.reg_use(v);
             collector.reg_early_def(t0);
             collector.reg_early_def(dst);
         }
 
-        &Inst::RawData { .. } => {}
-        &Inst::AtomicStore { src, p, .. } => {
+        Inst::RawData { .. } => {}
+        Inst::AtomicStore { src, p, .. } => {
             collector.reg_use(src);
             collector.reg_use(p);
         }
-        &Inst::AtomicLoad { rd, p, .. } => {
+        Inst::AtomicLoad { rd, p, .. } => {
             collector.reg_use(p);
             collector.reg_def(rd);
         }
-        &Inst::AtomicRmwLoop {
+        Inst::AtomicRmwLoop {
             offset,
             dst,
             p,
@@ -561,31 +555,21 @@ fn riscv64_get_operands<F: Fn(VReg) -> VReg>(inst: &Inst, collector: &mut Operan
             t0,
             ..
         } => {
-            collector.reg_uses(&[offset, p, x]);
+            collector.reg_use(offset);
+            collector.reg_use(p);
+            collector.reg_use(x);
             collector.reg_early_def(t0);
             collector.reg_early_def(dst);
         }
-        &Inst::TrapIf { rs1, rs2, .. } => {
+        Inst::TrapIf { rs1, rs2, .. } => {
             collector.reg_use(rs1);
             collector.reg_use(rs2);
         }
-        &Inst::Unwind { .. } => {}
-        &Inst::DummyUse { reg } => {
+        Inst::Unwind { .. } => {}
+        Inst::DummyUse { reg } => {
             collector.reg_use(reg);
         }
-        &Inst::FloatRound {
-            rd,
-            int_tmp,
-            f_tmp,
-            rs,
-            ..
-        } => {
-            collector.reg_use(rs);
-            collector.reg_early_def(int_tmp);
-            collector.reg_early_def(f_tmp);
-            collector.reg_early_def(rd);
-        }
-        &Inst::Popcnt {
+        Inst::Popcnt {
             sum, step, rs, tmp, ..
         } => {
             collector.reg_use(rs);
@@ -593,7 +577,7 @@ fn riscv64_get_operands<F: Fn(VReg) -> VReg>(inst: &Inst, collector: &mut Operan
             collector.reg_early_def(step);
             collector.reg_early_def(sum);
         }
-        &Inst::Cltz {
+        Inst::Cltz {
             sum, step, tmp, rs, ..
         } => {
             collector.reg_use(rs);
@@ -601,7 +585,7 @@ fn riscv64_get_operands<F: Fn(VReg) -> VReg>(inst: &Inst, collector: &mut Operan
             collector.reg_early_def(step);
             collector.reg_early_def(sum);
         }
-        &Inst::Brev8 {
+        Inst::Brev8 {
             rs,
             rd,
             step,
@@ -615,19 +599,19 @@ fn riscv64_get_operands<F: Fn(VReg) -> VReg>(inst: &Inst, collector: &mut Operan
             collector.reg_early_def(tmp2);
             collector.reg_early_def(rd);
         }
-        &Inst::StackProbeLoop { .. } => {
+        Inst::StackProbeLoop { .. } => {
             // StackProbeLoop has a tmp register and StackProbeLoop used at gen_prologue.
             // t3 will do the job. (t3 is caller-save register and not used directly by compiler like writable_spilltmp_reg)
             // gen_prologue is called at emit stage.
             // no need let reg alloc know.
         }
-        &Inst::VecAluRRRR {
+        Inst::VecAluRRRR {
             op,
             vd,
             vd_src,
             vs1,
             vs2,
-            ref mask,
+            mask,
             ..
         } => {
             debug_assert_eq!(vd_src.class(), RegClass::Vector);
@@ -641,12 +625,12 @@ fn riscv64_get_operands<F: Fn(VReg) -> VReg>(inst: &Inst, collector: &mut Operan
             collector.reg_reuse_def(vd, 2); // `vd` == `vd_src`.
             vec_mask_late_operands(mask, collector);
         }
-        &Inst::VecAluRRRImm5 {
+        Inst::VecAluRRRImm5 {
             op,
             vd,
             vd_src,
             vs2,
-            ref mask,
+            mask,
             ..
         } => {
             debug_assert_eq!(vd_src.class(), RegClass::Vector);
@@ -667,12 +651,12 @@ fn riscv64_get_operands<F: Fn(VReg) -> VReg>(inst: &Inst, collector: &mut Operan
                 vec_mask_operands(mask, collector);
             }
         }
-        &Inst::VecAluRRR {
+        Inst::VecAluRRR {
             op,
             vd,
             vs1,
             vs2,
-            ref mask,
+            mask,
             ..
         } => {
             debug_assert_eq!(vd.to_reg().class(), RegClass::Vector);
@@ -693,12 +677,8 @@ fn riscv64_get_operands<F: Fn(VReg) -> VReg>(inst: &Inst, collector: &mut Operan
 
             vec_mask_operands(mask, collector);
         }
-        &Inst::VecAluRRImm5 {
-            op,
-            vd,
-            vs2,
-            ref mask,
-            ..
+        Inst::VecAluRRImm5 {
+            op, vd, vs2, mask, ..
         } => {
             debug_assert_eq!(vd.to_reg().class(), RegClass::Vector);
             debug_assert_eq!(vs2.class(), RegClass::Vector);
@@ -716,12 +696,8 @@ fn riscv64_get_operands<F: Fn(VReg) -> VReg>(inst: &Inst, collector: &mut Operan
 
             vec_mask_operands(mask, collector);
         }
-        &Inst::VecAluRR {
-            op,
-            vd,
-            vs,
-            ref mask,
-            ..
+        Inst::VecAluRR {
+            op, vd, vs, mask, ..
         } => {
             debug_assert_eq!(vd.to_reg().class(), op.dst_regclass());
             debug_assert_eq!(vs.class(), op.src_regclass());
@@ -739,39 +715,23 @@ fn riscv64_get_operands<F: Fn(VReg) -> VReg>(inst: &Inst, collector: &mut Operan
 
             vec_mask_operands(mask, collector);
         }
-        &Inst::VecAluRImm5 {
-            op, vd, ref mask, ..
-        } => {
+        Inst::VecAluRImm5 { op, vd, mask, .. } => {
             debug_assert_eq!(vd.to_reg().class(), RegClass::Vector);
             debug_assert!(!op.forbids_overlaps(mask));
 
             collector.reg_def(vd);
             vec_mask_operands(mask, collector);
         }
-        &Inst::VecSetState { rd, .. } => {
+        Inst::VecSetState { rd, .. } => {
             collector.reg_def(rd);
         }
-        &Inst::VecLoad {
-            to,
-            ref from,
-            ref mask,
-            ..
-        } => {
-            if let Some(r) = from.get_allocatable_register() {
-                collector.reg_use(r);
-            }
+        Inst::VecLoad { to, from, mask, .. } => {
+            from.get_operands(collector);
             collector.reg_def(to);
             vec_mask_operands(mask, collector);
         }
-        &Inst::VecStore {
-            ref to,
-            from,
-            ref mask,
-            ..
-        } => {
-            if let Some(r) = to.get_allocatable_register() {
-                collector.reg_use(r);
-            }
+        Inst::VecStore { to, from, mask, .. } => {
+            to.get_operands(collector);
             collector.reg_use(from);
             vec_mask_operands(mask, collector);
         }
@@ -808,7 +768,7 @@ impl MachInst for Inst {
         }
     }
 
-    fn get_operands<F: Fn(VReg) -> VReg>(&self, collector: &mut OperandCollector<'_, F>) {
+    fn get_operands(&mut self, collector: &mut impl OperandVisitor) {
         riscv64_get_operands(self, collector);
     }
 
@@ -968,30 +928,26 @@ pub fn reg_name(reg: Reg) -> String {
 }
 
 impl Inst {
-    fn print_with_state(
-        &self,
-        _state: &mut EmitState,
-        allocs: &mut AllocationConsumer<'_>,
-    ) -> String {
-        let format_reg = |reg: Reg, allocs: &mut AllocationConsumer<'_>| -> String {
+    fn print_with_state(&self, _state: &mut EmitState, allocs: &mut AllocationConsumer) -> String {
+        let format_reg = |reg: Reg, allocs: &mut AllocationConsumer| -> String {
             let reg = allocs.next(reg);
             reg_name(reg)
         };
 
-        let format_vec_amode = |amode: &VecAMode, allocs: &mut AllocationConsumer<'_>| -> String {
+        let format_vec_amode = |amode: &VecAMode, allocs: &mut AllocationConsumer| -> String {
             match amode {
                 VecAMode::UnitStride { base } => base.to_string_with_alloc(allocs),
             }
         };
 
-        let format_mask = |mask: &VecOpMasking, allocs: &mut AllocationConsumer<'_>| -> String {
+        let format_mask = |mask: &VecOpMasking, allocs: &mut AllocationConsumer| -> String {
             match mask {
                 VecOpMasking::Enabled { reg } => format!(",{}.t", format_reg(*reg, allocs)),
                 VecOpMasking::Disabled => format!(""),
             }
         };
 
-        let format_regs = |regs: &[Reg], allocs: &mut AllocationConsumer<'_>| -> String {
+        let format_regs = |regs: &[Reg], allocs: &mut AllocationConsumer| -> String {
             let mut x = if regs.len() > 1 {
                 String::from("[")
             } else {
@@ -1050,28 +1006,6 @@ impl Inst {
                     guard_size, probe_count, tmp
                 )
             }
-            &Inst::FloatRound {
-                op,
-                rd,
-                int_tmp,
-                f_tmp,
-                rs,
-                ty,
-            } => {
-                let rs = format_reg(rs, allocs);
-                let int_tmp = format_reg(int_tmp.to_reg(), allocs);
-                let f_tmp = format_reg(f_tmp.to_reg(), allocs);
-                let rd = format_reg(rd.to_reg(), allocs);
-                format!(
-                    "{} {},{}##int_tmp={} f_tmp={} ty={}",
-                    op.op_name(),
-                    rd,
-                    rs,
-                    int_tmp,
-                    f_tmp,
-                    ty
-                )
-            }
             &Inst::AtomicStore { src, ty, p } => {
                 let src = format_reg(src, allocs);
                 let p = format_reg(p, allocs);
@@ -1475,8 +1409,8 @@ impl Inst {
                 ref info,
             } => {
                 let mut s = format!(
-                    "return_call {callee:?} old_stack_arg_size:{} new_stack_arg_size:{}",
-                    info.old_stack_arg_size, info.new_stack_arg_size
+                    "return_call {callee:?} new_stack_arg_size:{}",
+                    info.new_stack_arg_size
                 );
                 for ret in &info.uses {
                     let preg = format_reg(ret.preg, &mut empty_allocs);
@@ -1488,8 +1422,8 @@ impl Inst {
             &MInst::ReturnCallInd { callee, ref info } => {
                 let callee = format_reg(callee, allocs);
                 let mut s = format!(
-                    "return_call_ind {callee} old_stack_arg_size:{} new_stack_arg_size:{}",
-                    info.old_stack_arg_size, info.new_stack_arg_size
+                    "return_call_ind {callee} new_stack_arg_size:{}",
+                    info.new_stack_arg_size
                 );
                 for ret in &info.uses {
                     let preg = format_reg(ret.preg, &mut empty_allocs);
@@ -1571,8 +1505,8 @@ impl Inst {
                 format!("virtual_sp_offset_adj {:+}", amount)
             }
             &MInst::Mov { rd, rm, ty } => {
-                let rd = format_reg(rd.to_reg(), allocs);
                 let rm = format_reg(rm, allocs);
+                let rd = format_reg(rd.to_reg(), allocs);
 
                 let op = match ty {
                     F32 => "fmv.s",
