@@ -1,10 +1,9 @@
-use crate::isa::riscv64::inst::AllocationConsumer;
 use crate::isa::riscv64::lower::isle::generated_code::VecAluOpRRRR;
 use crate::isa::riscv64::lower::isle::generated_code::{
     VecAMode, VecAluOpRImm5, VecAluOpRR, VecAluOpRRImm5, VecAluOpRRR, VecAluOpRRRImm5, VecAvl,
     VecElementWidth, VecLmul, VecMaskMode, VecOpCategory, VecOpMasking, VecTailMode,
 };
-use crate::machinst::RegClass;
+use crate::machinst::{OperandVisitor, RegClass};
 use crate::Reg;
 use core::fmt;
 
@@ -249,15 +248,6 @@ impl VecOpMasking {
             VecOpMasking::Disabled => 1,
         }
     }
-
-    pub(crate) fn with_allocs(&self, allocs: &mut AllocationConsumer<'_>) -> Self {
-        match self {
-            VecOpMasking::Enabled { reg } => VecOpMasking::Enabled {
-                reg: allocs.next(*reg),
-            },
-            VecOpMasking::Disabled => VecOpMasking::Disabled,
-        }
-    }
 }
 
 impl VecAluOpRRRR {
@@ -1070,17 +1060,9 @@ impl VecAMode {
         }
     }
 
-    pub fn get_allocatable_register(&self) -> Option<Reg> {
+    pub fn get_operands(&mut self, collector: &mut impl OperandVisitor) {
         match self {
-            VecAMode::UnitStride { base, .. } => base.get_allocatable_register(),
-        }
-    }
-
-    pub(crate) fn with_allocs(self, allocs: &mut AllocationConsumer<'_>) -> Self {
-        match self {
-            VecAMode::UnitStride { base } => VecAMode::UnitStride {
-                base: base.with_allocs(allocs),
-            },
+            VecAMode::UnitStride { base, .. } => base.get_operands(collector),
         }
     }
 
