@@ -102,42 +102,9 @@ impl MemArg {
         }
     }
 
-    pub(crate) fn can_trap(&self) -> bool {
-        !self.get_flags().notrap()
-    }
-
     /// Edit registers with allocations.
-    pub fn with_allocs(&self, allocs: &mut AllocationConsumer<'_>) -> Self {
-        match self {
-            &MemArg::BXD12 {
-                base,
-                index,
-                disp,
-                flags,
-            } => MemArg::BXD12 {
-                base: allocs.next(base),
-                index: allocs.next(index),
-                disp,
-                flags,
-            },
-            &MemArg::BXD20 {
-                base,
-                index,
-                disp,
-                flags,
-            } => MemArg::BXD20 {
-                base: allocs.next(base),
-                index: allocs.next(index),
-                disp,
-                flags,
-            },
-            &MemArg::RegOffset { reg, off, flags } => MemArg::RegOffset {
-                reg: allocs.next(reg),
-                off,
-                flags,
-            },
-            x => x.clone(),
-        }
+    pub fn with_allocs(&self, _allocs: &mut AllocationConsumer) -> Self {
+        self.clone()
     }
 }
 
@@ -187,17 +154,9 @@ impl MemArgPair {
         }
     }
 
-    pub(crate) fn can_trap(&self) -> bool {
-        !self.flags.notrap()
-    }
-
     /// Edit registers with allocations.
-    pub fn with_allocs(&self, allocs: &mut AllocationConsumer<'_>) -> Self {
-        MemArgPair {
-            base: allocs.next(self.base),
-            disp: self.disp,
-            flags: self.flags,
-        }
+    pub fn with_allocs(&self, _allocs: &mut AllocationConsumer) -> Self {
+        self.clone()
     }
 }
 
@@ -267,7 +226,7 @@ impl Cond {
 }
 
 impl PrettyPrint for MemArg {
-    fn pretty_print(&self, _: u8, allocs: &mut AllocationConsumer<'_>) -> String {
+    fn pretty_print(&self, _: u8, allocs: &mut AllocationConsumer) -> String {
         match self {
             &MemArg::BXD12 {
                 base, index, disp, ..
@@ -332,7 +291,7 @@ impl PrettyPrint for MemArg {
 }
 
 impl PrettyPrint for Cond {
-    fn pretty_print(&self, _: u8, _: &mut AllocationConsumer<'_>) -> String {
+    fn pretty_print(&self, _: u8, _: &mut AllocationConsumer) -> String {
         let s = match self.mask {
             1 => "o",
             2 => "h",
