@@ -79,12 +79,10 @@ impl<'data> ModuleEnvironment<'data> for ModuleEnv {
     }
 
     fn wasm_features(&self) -> wasmparser::WasmFeatures {
-        wasmparser::WasmFeatures {
-            memory64: true,
-            multi_memory: true,
-            relaxed_simd: true,
-            ..self.inner.wasm_features()
-        }
+        self.inner.wasm_features()
+            | wasmparser::WasmFeatures::MEMORY64
+            | wasmparser::WasmFeatures::MULTI_MEMORY
+            | wasmparser::WasmFeatures::RELAXED_SIMD
     }
 
     // ================================================================
@@ -687,4 +685,28 @@ impl<'a> FuncEnvironment for FuncEnv<'a> {
     ) -> cranelift_wasm::WasmResult<ir::Inst> {
         unimplemented!()
     }
+
+    fn translate_ref_i31(
+        &mut self,
+        _pos: cranelift_codegen::cursor::FuncCursor,
+        _val: ir::Value,
+    ) -> cranelift_wasm::WasmResult<ir::Value> {
+        unimplemented!()
+    }
+
+    fn translate_i31_get_s(
+        &mut self,
+        _pos: cranelift_codegen::cursor::FuncCursor,
+        _i31ref: ir::Value,
+    ) -> cranelift_wasm::WasmResult<ir::Value> {
+        unimplemented!()
+    }
+
+    fn translate_i31_get_u(
+        &mut self,
+        _pos: cranelift_codegen::cursor::FuncCursor,
+        _i31ref: ir::Value,
+    ) -> cranelift_wasm::WasmResult<ir::Value> {
+        unimplemented!()
+    }
 }
