@@ -98,7 +98,7 @@ where
         Ok(inputs)
     }
 
-    fn run_func_passes(&mut self, func: Function) -> Result<Function> {
+    fn run_func_passes(&mut self, func: Function, isa: &dyn TargetIsa) -> Result<Function> {
         // Do a NaN Canonicalization pass on the generated function.
         //
         // Both IEEE754 and the Wasm spec are somewhat loose about what is allowed
@@ -115,7 +115,7 @@ where
         // the interpreter won't get that version, so call that pass manually here.
 
         let mut ctx = Context::for_function(func);
-        // Assume that we are generating this function for the current ISA.
+
         // We disable the verifier here, since if it fails it prevents a test case from
         // being generated and formatted by `cargo fuzz fmt`.
         // We run the verifier before compiling the code, so it always gets verified.
@@ -125,11 +125,13 @@ where
             builder
         });
 
-        let isa = builder_with_options(false)
-            .expect("Unable to build a TargetIsa for the current host")
+        // Create a new TargetISA from the given ISA, this ensures that we copy all ISA
+        // flags, which may have an effect on the code generated by the passes below.
+        let isa = Builder::from_target_isa(isa)
             .finish(flags)
             .expect("Failed to build TargetISA");
 
+        // Finally run the NaN canonicalization pass
         ctx.canonicalize_nans(isa.as_ref())
             .expect("Failed NaN canonicalization pass");
 
@@ -156,7 +158,7 @@ where
         let func = FunctionGenerator::new(
             &mut self.u,
             &self.config,
-            isa,
+            isa.clone(),
             name,
             sig,
             usercalls,
@@ -164,7 +166,7 @@ where
         )
         .generate()?;
 
-        self.run_func_passes(func)
+        self.run_func_passes(func, &*isa)
     }
 
     /// Generate a random set of cranelift flags.
