@@ -822,18 +822,10 @@ fn inst_count(func: &Function) -> usize {
         .sum()
 }
 
-fn resolve_aliases(func: &mut Function) {
-    for block in func.stencil.layout.blocks() {
-        for inst in func.stencil.layout.block_insts(block) {
-            func.stencil.dfg.resolve_aliases_in_arguments(inst);
-        }
-    }
-}
-
 /// Resolve aliases only if function still crashes after this.
 fn try_resolve_aliases(context: &mut CrashCheckContext, func: &mut Function) {
     let mut func_with_resolved_aliases = func.clone();
-    resolve_aliases(&mut func_with_resolved_aliases);
+    func_with_resolved_aliases.dfg.resolve_all_aliases();
     if let CheckResult::Crash(_) = context.check_for_crash(&func_with_resolved_aliases) {
         *func = func_with_resolved_aliases;
     }
