@@ -1,12 +1,11 @@
 use crate::r#ref::ref_to_val;
 use crate::{
-    from_valtype, into_valtype, wasm_ref_t, wasm_valkind_t, wasmtime_valkind_t, CStoreContextMut,
-    WASM_I32,
+    from_valtype, into_valtype, wasm_ref_t, wasm_valkind_t, wasmtime_anyref_t,
+    wasmtime_externref_t, wasmtime_valkind_t, WasmtimeStoreContextMut, WASM_I32,
 };
-use std::ffi::c_void;
-use std::mem::{self, ManuallyDrop, MaybeUninit};
+use std::mem::{ManuallyDrop, MaybeUninit};
 use std::ptr;
-use wasmtime::{ExternRef, Func, HeapType, Ref, Val, ValType};
+use wasmtime::{AsContextMut, Func, HeapType, Ref, RootScope, Val, ValType};
 
 #[repr(C)]
 pub struct wasm_val_t {
@@ -85,16 +84,6 @@ impl wasm_val_t {
                 kind: from_valtype(&ValType::F64),
                 of: wasm_val_union { u64: f },
             },
-            Val::ExternRef(r) => wasm_val_t {
-                kind: from_valtype(&ValType::EXTERNREF),
-                of: wasm_val_union {
-                    ref_: r.map_or(ptr::null_mut(), |r| {
-                        Box::into_raw(Box::new(wasm_ref_t {
-                            r: Ref::Extern(Some(r)),
-                        }))
-                    }),
-                },
-            },
             Val::FuncRef(f) => wasm_val_t {
                 kind: from_valtype(&ValType::FUNCREF),
                 of: wasm_val_union {
@@ -105,7 +94,9 @@ impl wasm_val_t {
                     }),
                 },
             },
-            _ => unimplemented!("wasm_val_t::from_val {:?}", val),
+            Val::AnyRef(_) => crate::abort("creating a wasm_val_t from an anyref"),
+            Val::ExternRef(_) => crate::abort("creating a wasm_val_t from an externref"),
+            Val::V128(_) => crate::abort("creating a wasm_val_t from a v128"),
         }
     }
 
@@ -115,22 +106,16 @@ impl wasm_val_t {
             ValType::I64 => Val::from(unsafe { self.of.i64 }),
             ValType::F32 => Val::from(unsafe { self.of.f32 }),
             ValType::F64 => Val::from(unsafe { self.of.f64 }),
-            ValType::Ref(r) => match (r.is_nullable(), r.heap_type()) {
-                (true, HeapType::Extern) => unsafe {
-                    if self.of.ref_.is_null() {
-                        Val::ExternRef(None)
-                    } else {
-                        ref_to_val(&*self.of.ref_)
-                    }
-                },
-                (true, HeapType::Func) => unsafe {
+            ValType::Ref(r) => match r.heap_type() {
+                HeapType::Func => unsafe {
                     if self.of.ref_.is_null() {
+                        assert!(r.is_nullable());
                         Val::FuncRef(None)
                     } else {
                         ref_to_val(&*self.of.ref_)
                     }
                 },
-                _ => unimplemented!("wasm_val_t: non-funcref and non-externref reference types"),
+                _ => unreachable!("wasm_val_t cannot contain non-function reference values"),
             },
             ValType::V128 => unimplemented!("wasm_val_t: v128"),
         }
@@ -159,20 +144,76 @@ pub union wasmtime_val_union {
     pub i64: i64,
     pub f32: u32,
     pub f64: u64,
+    pub anyref: ManuallyDrop<wasmtime_anyref_t>,
+    pub externref: ManuallyDrop<wasmtime_externref_t>,
     pub funcref: wasmtime_func_t,
-    pub externref: ManuallyDrop<Option<ExternRef>>,
     pub v128: [u8; 16],
 }
 
+const _: () = {
+    assert!(std::mem::size_of::<wasmtime_val_union>() == 16);
+    assert!(std::mem::align_of::<wasmtime_val_union>() == 8);
+};
+
+// The raw pointers are actually optional boxes.
+unsafe impl Send for wasmtime_val_union
+where
+    Option<Box<wasmtime_anyref_t>>: Send,
+    Option<Box<wasmtime_externref_t>>: Send,
+{
+}
+unsafe impl Sync for wasmtime_val_union
+where
+    Option<Box<wasmtime_anyref_t>>: Sync,
+    Option<Box<wasmtime_externref_t>>: Sync,
+{
+}
+
 #[repr(C)]
 #[derive(Clone, Copy)]
-pub struct wasmtime_func_t {
-    pub store_id: u64,
-    pub index: usize,
+pub union wasmtime_func_t {
+    store_id: u64,
+    func: Func,
+}
+
+impl wasmtime_func_t {
+    unsafe fn as_wasmtime(&self) -> Option<Func> {
+        if self.store_id == 0 {
+            None
+        } else {
+            Some(self.func)
+        }
+    }
+}
+
+impl From<Option<Func>> for wasmtime_func_t {
+    fn from(func: Option<Func>) -> wasmtime_func_t {
+        match func {
+            Some(func) => wasmtime_func_t { func },
+            None => wasmtime_func_t { store_id: 0 },
+        }
+    }
 }
 
 impl wasmtime_val_t {
-    pub fn from_val(val: Val) -> wasmtime_val_t {
+    /// Creates a new `wasmtime_val_t` from a `wasmtime::Val`.
+    ///
+    /// Note that this requires a `RootScope` to be present to serve as proof
+    /// that `val` is not require to be rooted in the store itself which would
+    /// prevent GC. Callers should prefer this API where possible, creating a
+    /// temporary `RootScope` when needed.
+    pub fn from_val(cx: &mut RootScope<impl AsContextMut>, val: Val) -> wasmtime_val_t {
+        Self::from_val_unscoped(cx, val)
+    }
+
+    /// Equivalent of [`wasmtime_val_t::from_val`] except that a `RootScope`
+    /// is not required.
+    ///
+    /// This method should only be used when a `RootScope` is known to be
+    /// elsewhere on the stack. For example this is used when we call back out
+    /// to the embedder. In such a situation we know we previously entered with
+    /// some other call so the root scope is on the stack there.
+    pub fn from_val_unscoped(cx: impl AsContextMut, val: Val) -> wasmtime_val_t {
         match val {
             Val::I32(i) => wasmtime_val_t {
                 kind: crate::WASMTIME_I32,
@@ -190,22 +231,24 @@ impl wasmtime_val_t {
                 kind: crate::WASMTIME_F64,
                 of: wasmtime_val_union { f64: i },
             },
+            Val::AnyRef(a) => wasmtime_val_t {
+                kind: crate::WASMTIME_ANYREF,
+                of: wasmtime_val_union {
+                    anyref: ManuallyDrop::new(a.and_then(|a| a.to_manually_rooted(cx).ok()).into()),
+                },
+            },
             Val::ExternRef(e) => wasmtime_val_t {
                 kind: crate::WASMTIME_EXTERNREF,
                 of: wasmtime_val_union {
-                    externref: ManuallyDrop::new(e),
+                    externref: ManuallyDrop::new(
+                        e.and_then(|e| e.to_manually_rooted(cx).ok()).into(),
+                    ),
                 },
             },
             Val::FuncRef(func) => wasmtime_val_t {
                 kind: crate::WASMTIME_FUNCREF,
                 of: wasmtime_val_union {
-                    funcref: match func {
-                        None => wasmtime_func_t {
-                            store_id: 0,
-                            index: 0,
-                        },
-                        Some(func) => unsafe { mem::transmute::<Func, wasmtime_func_t>(func) },
-                    },
+                    funcref: func.into(),
                 },
             },
             Val::V128(val) => wasmtime_val_t {
@@ -217,91 +260,68 @@ impl wasmtime_val_t {
         }
     }
 
-    pub unsafe fn to_val(&self) -> Val {
+    /// Convert this `wasmtime_val_t` into a `wasmtime::Val`.
+    ///
+    /// See [`wasmtime_val_t::from_val`] for notes on the `RootScope`
+    /// requirement here. Note that this is particularly meaningful for this
+    /// API as the `Val` returned may contain a `Rooted<T>` which requires a
+    /// `RootScope` if we don't want the value to live for the entire lifetime
+    /// of the `Store`.
+    pub unsafe fn to_val(&self, cx: &mut RootScope<impl AsContextMut>) -> Val {
+        self.to_val_unscoped(cx)
+    }
+
+    /// Equivalent of `to_val` except doesn't require a `RootScope`.
+    ///
+    /// See notes on [`wasmtime_val_t::from_val_unscoped`] for notes on when to
+    /// use this.
+    pub unsafe fn to_val_unscoped(&self, cx: impl AsContextMut) -> Val {
         match self.kind {
             crate::WASMTIME_I32 => Val::I32(self.of.i32),
             crate::WASMTIME_I64 => Val::I64(self.of.i64),
             crate::WASMTIME_F32 => Val::F32(self.of.f32),
             crate::WASMTIME_F64 => Val::F64(self.of.f64),
             crate::WASMTIME_V128 => Val::V128(u128::from_le_bytes(self.of.v128).into()),
-            crate::WASMTIME_FUNCREF => {
-                let store = self.of.funcref.store_id;
-                let index = self.of.funcref.index;
-                Val::FuncRef(if store == 0 && index == 0 {
-                    None
-                } else {
-                    Some(mem::transmute::<wasmtime_func_t, Func>(self.of.funcref))
-                })
+            crate::WASMTIME_ANYREF => {
+                Val::AnyRef(self.of.anyref.as_wasmtime().map(|a| a.to_rooted(cx)))
             }
-            crate::WASMTIME_EXTERNREF => Val::ExternRef((*self.of.externref).clone()),
+            crate::WASMTIME_EXTERNREF => {
+                Val::ExternRef(self.of.externref.as_wasmtime().map(|e| e.to_rooted(cx)))
+            }
+            crate::WASMTIME_FUNCREF => Val::FuncRef(self.of.funcref.as_wasmtime()),
             other => panic!("unknown wasmtime_valkind_t: {}", other),
         }
     }
 }
 
-impl Drop for wasmtime_val_t {
-    fn drop(&mut self) {
-        if self.kind == crate::WASMTIME_EXTERNREF {
-            unsafe {
-                ManuallyDrop::drop(&mut self.of.externref);
+#[no_mangle]
+pub unsafe extern "C" fn wasmtime_val_unroot(
+    cx: WasmtimeStoreContextMut<'_>,
+    val: &mut MaybeUninit<wasmtime_val_t>,
+) {
+    let val = val.assume_init_read();
+    match val.kind {
+        crate::WASMTIME_ANYREF => {
+            if let Some(val) = ManuallyDrop::into_inner(val.of.anyref).as_wasmtime() {
+                val.unroot(cx);
             }
         }
+        crate::WASMTIME_EXTERNREF => {
+            if let Some(val) = ManuallyDrop::into_inner(val.of.externref).as_wasmtime() {
+                val.unroot(cx);
+            }
+        }
+        _ => {}
     }
 }
 
 #[no_mangle]
-pub unsafe extern "C" fn wasmtime_val_delete(val: &mut ManuallyDrop<wasmtime_val_t>) {
-    ManuallyDrop::drop(val)
-}
-
-#[no_mangle]
-pub unsafe extern "C" fn wasmtime_val_copy(
-    dst: &mut MaybeUninit<wasmtime_val_t>,
+pub unsafe extern "C" fn wasmtime_val_clone(
+    cx: WasmtimeStoreContextMut<'_>,
     src: &wasmtime_val_t,
+    dst: &mut MaybeUninit<wasmtime_val_t>,
 ) {
-    crate::initialize(dst, wasmtime_val_t::from_val(src.to_val()))
-}
-
-#[no_mangle]
-pub extern "C" fn wasmtime_externref_new(
-    data: *mut c_void,
-    finalizer: Option<extern "C" fn(*mut c_void)>,
-) -> ExternRef {
-    ExternRef::new(crate::ForeignData { data, finalizer })
-}
-
-#[no_mangle]
-pub extern "C" fn wasmtime_externref_data(externref: ManuallyDrop<ExternRef>) -> *mut c_void {
-    externref
-        .data()
-        .downcast_ref::<crate::ForeignData>()
-        .unwrap()
-        .data
-}
-
-#[no_mangle]
-pub extern "C" fn wasmtime_externref_clone(externref: ManuallyDrop<ExternRef>) -> ExternRef {
-    (*externref).clone()
-}
-
-#[no_mangle]
-pub extern "C" fn wasmtime_externref_delete(_val: Option<ExternRef>) {}
-
-#[no_mangle]
-pub unsafe extern "C" fn wasmtime_externref_to_raw(
-    cx: CStoreContextMut<'_>,
-    val: Option<ManuallyDrop<ExternRef>>,
-) -> *mut c_void {
-    match val {
-        Some(ptr) => ptr.to_raw(cx),
-        None => ptr::null_mut(),
-    }
-}
-
-#[no_mangle]
-pub unsafe extern "C" fn wasmtime_externref_from_raw(
-    _cx: CStoreContextMut<'_>,
-    val: *mut c_void,
-) -> Option<ExternRef> {
-    ExternRef::from_raw(val)
+    let mut scope = RootScope::new(cx);
+    let val = src.to_val(&mut scope);
+    crate::initialize(dst, wasmtime_val_t::from_val(&mut scope, val))
 }
