@@ -1,11 +1,10 @@
 use crate::debug::{DwarfSectionRelocTarget, ModuleMemoryOffset};
 use crate::func_environ::FuncEnvironment;
 use crate::{array_call_signature, native_call_signature, DEBUG_ASSERT_TRAP_CODE};
-use crate::{builder::LinkOptions, value_type, wasm_call_signature};
+use crate::{builder::LinkOptions, wasm_call_signature, BuiltinFunctionSignatures};
+use crate::{CompiledFunction, ModuleTextBuilder};
 use anyhow::{Context as _, Result};
-use cranelift_codegen::ir::{
-    self, InstBuilder, MemFlags, UserExternalName, UserExternalNameRef, UserFuncName, Value,
-};
+use cranelift_codegen::ir::{self, InstBuilder, MemFlags, UserExternalName, UserFuncName, Value};
 use cranelift_codegen::isa::{
     unwind::{UnwindInfo, UnwindInfoKind},
     OwnedTargetIsa, TargetIsa,
@@ -16,11 +15,10 @@ use cranelift_codegen::{CompiledCode, MachStackMap};
 use cranelift_entity::{EntityRef, PrimaryMap};
 use cranelift_frontend::FunctionBuilder;
 use cranelift_wasm::{
-    DefinedFuncIndex, FuncIndex, FuncTranslator, MemoryIndex, OwnedMemoryIndex, WasmFuncType,
-    WasmValType,
+    DefinedFuncIndex, FuncTranslator, MemoryIndex, OwnedMemoryIndex, WasmFuncType, WasmValType,
 };
 use object::write::{Object, StandardSegment, SymbolId};
-use object::{RelocationEncoding, RelocationKind, SectionKind};
+use object::{RelocationEncoding, RelocationFlags, RelocationKind, SectionKind};
 use std::any::Any;
 use std::cmp;
 use std::collections::HashMap;
@@ -28,11 +26,10 @@ use std::mem;
 use std::path;
 use std::sync::{Arc, Mutex};
 use wasmparser::{FuncValidatorAllocations, FunctionBody};
-use wasmtime_cranelift_shared::{CompiledFunction, ModuleTextBuilder};
 use wasmtime_environ::{
-    AddressMapSection, CacheStore, CompileError, FlagValue, FunctionBodyData, FunctionLoc,
-    ModuleTranslation, ModuleTypesBuilder, PtrSize, StackMapInformation, TrapEncodingBuilder,
-    Tunables, VMOffsets, WasmFunctionInfo,
+    AddressMapSection, BuiltinFunctionIndex, CacheStore, CompileError, FlagValue, FunctionBodyData,
+    FunctionLoc, ModuleTranslation, ModuleTypesBuilder, PtrSize, RelocationTarget,
+    StackMapInformation, TrapEncodingBuilder, Tunables, VMOffsets, WasmFunctionInfo,
 };
 
 #[cfg(feature = "component-model")]
@@ -65,7 +62,7 @@ impl Default for CompilerContext {
 
 /// A compiler that compiles a WebAssembly module with Compiler, translating
 /// the Wasm to Compiler IR, optimizing it and then translating to assembly.
-pub(crate) struct Compiler {
+pub struct Compiler {
     tunables: Tunables,
     contexts: Mutex<Vec<CompilerContext>>,
     isa: OwnedTargetIsa,
@@ -104,7 +101,7 @@ impl Drop for Compiler {
 }
 
 impl Compiler {
-    pub(crate) fn new(
+    pub fn new(
         tunables: Tunables,
         isa: OwnedTargetIsa,
         cache_store: Option<Arc<dyn CacheStore>>,
@@ -136,14 +133,14 @@ impl wasmtime_environ::Compiler for Compiler {
         let module = &translation.module;
         let func_index = module.func_index(func_index);
         let sig = translation.module.functions[func_index].signature;
-        let wasm_func_ty = &types[sig];
+        let wasm_func_ty = types[sig].unwrap_func();
 
         let mut compiler = self.function_compiler();
 
         let context = &mut compiler.cx.codegen_context;
         context.func.signature = wasm_call_signature(isa, wasm_func_ty, &self.tunables);
         context.func.name = UserFuncName::User(UserExternalName {
-            namespace: 0,
+            namespace: crate::NS_WASM_FUNC,
             index: func_index.as_u32(),
         });
 
@@ -151,8 +148,14 @@ impl wasmtime_environ::Compiler for Compiler {
             context.func.collect_debug_info();
         }
 
-        let mut func_env =
-            FuncEnvironment::new(isa, translation, types, &self.tunables, self.wmemcheck);
+        let mut func_env = FuncEnvironment::new(
+            isa,
+            translation,
+            types,
+            &self.tunables,
+            self.wmemcheck,
+            input.call_indirect_start,
+        );
 
         // The `stack_limit` global value below is the implementation of stack
         // overflow checks in Wasmtime.
@@ -206,7 +209,11 @@ impl wasmtime_environ::Compiler for Compiler {
             flags: MemFlags::trusted(),
         });
         context.func.stack_limit = Some(stack_limit);
-        let FunctionBodyData { validator, body } = input;
+        let FunctionBodyData {
+            validator,
+            body,
+            call_indirect_start: _,
+        } = input;
         let mut validator =
             validator.into_validator(mem::take(&mut compiler.cx.validator_allocations));
         compiler.cx.func_translator.translate_body(
@@ -244,7 +251,7 @@ impl wasmtime_environ::Compiler for Compiler {
     ) -> Result<Box<dyn Any + Send>, CompileError> {
         let func_index = translation.module.func_index(def_func_index);
         let sig = translation.module.functions[func_index].signature;
-        let wasm_func_ty = &types[sig];
+        let wasm_func_ty = types[sig].unwrap_func();
 
         let isa = &*self.isa;
         let pointer_type = isa.pointer_type();
@@ -312,7 +319,7 @@ impl wasmtime_environ::Compiler for Compiler {
     ) -> Result<Box<dyn Any + Send>, CompileError> {
         let func_index = translation.module.func_index(def_func_index);
         let sig = translation.module.functions[func_index].signature;
-        let wasm_func_ty = &types[sig];
+        let wasm_func_ty = types[sig].unwrap_func();
 
         let isa = &*self.isa;
         let pointer_type = isa.pointer_type();
@@ -342,7 +349,7 @@ impl wasmtime_environ::Compiler for Compiler {
             vmctx,
         );
 
-        let ret = NativeRet::classify(pointer_type, wasm_func_ty);
+        let ret = NativeRet::classify(isa, wasm_func_ty);
         let wasm_args = ret.native_args(&args);
 
         // Then call into Wasm.
@@ -373,7 +380,7 @@ impl wasmtime_environ::Compiler for Compiler {
         let callee_vmctx = args[0];
         let caller_vmctx = args[1];
 
-        let ret = NativeRet::classify(pointer_type, wasm_func_ty);
+        let ret = NativeRet::classify(isa, wasm_func_ty);
 
         // We are exiting Wasm, so save our PC and FP.
         //
@@ -426,15 +433,14 @@ impl wasmtime_environ::Compiler for Compiler {
         // use for the native call then load the results from the return pointer
         // to pass through as native return values in the wasm abi.
         let mut results = builder.func.dfg.inst_results(call).to_vec();
-        if let NativeRet::Retptr { offsets, .. } = ret {
-            let slot = *args.last().unwrap();
-            assert_eq!(offsets.len(), wasm_func_ty.returns().len() - 1);
-            for (ty, offset) in wasm_func_ty.returns()[1..].iter().zip(offsets) {
-                let ty = crate::value_type(isa, *ty);
+        if let NativeRet::Retptr { slots, .. } = ret {
+            let base = *args.last().unwrap();
+            assert_eq!(slots.len(), wasm_func_ty.returns().len() - 1);
+            for (offset, ty) in slots {
                 results.push(
                     builder
                         .ins()
-                        .load(ty, MemFlags::trusted(), slot, offset as i32),
+                        .load(ty, ir::MemFlags::trusted(), base, offset),
                 );
             }
         }
@@ -448,7 +454,7 @@ impl wasmtime_environ::Compiler for Compiler {
         &self,
         obj: &mut Object<'static>,
         funcs: &[(String, Box<dyn Any + Send>)],
-        resolve_reloc: &dyn Fn(usize, FuncIndex) -> usize,
+        resolve_reloc: &dyn Fn(usize, RelocationTarget) -> usize,
     ) -> Result<Vec<(SymbolId, FunctionLoc)>> {
         let mut builder =
             ModuleTextBuilder::new(obj, self, self.isa.text_section_builder(funcs.len()));
@@ -460,9 +466,7 @@ impl wasmtime_environ::Compiler for Compiler {
 
         let mut ret = Vec::with_capacity(funcs.len());
         for (i, (sym, func)) in funcs.iter().enumerate() {
-            let func = func
-                .downcast_ref::<CompiledFunction<CompiledFuncEnv>>()
-                .unwrap();
+            let func = func.downcast_ref::<CompiledFunction>().unwrap();
             let (sym, range) = builder.append_func(&sym, func, |idx| resolve_reloc(i, idx));
             if self.tunables.generate_address_map {
                 let addr = func.address_map();
@@ -521,11 +525,11 @@ impl wasmtime_environ::Compiler for Compiler {
     }
 
     fn flags(&self) -> Vec<(&'static str, FlagValue<'static>)> {
-        wasmtime_cranelift_shared::clif_flags_to_wasmtime(self.isa.flags().iter())
+        crate::clif_flags_to_wasmtime(self.isa.flags().iter())
     }
 
     fn isa_flags(&self) -> Vec<(&'static str, FlagValue<'static>)> {
-        wasmtime_cranelift_shared::clif_flags_to_wasmtime(self.isa.isa_flags())
+        crate::clif_flags_to_wasmtime(self.isa.isa_flags())
     }
 
     fn is_branch_protection_enabled(&self) -> bool {
@@ -542,6 +546,8 @@ impl wasmtime_environ::Compiler for Compiler {
         obj: &mut Object<'_>,
         translation: &ModuleTranslation<'_>,
         funcs: &PrimaryMap<DefinedFuncIndex, (SymbolId, &(dyn Any + Send))>,
+        dwarf_package_bytes: Option<&[u8]>,
+        tunables: &Tunables,
     ) -> Result<()> {
         let ofs = VMOffsets::new(
             self.isa
@@ -575,7 +581,7 @@ impl wasmtime_environ::Compiler for Compiler {
         let functions_info = funcs
             .iter()
             .map(|(_, (_, func))| {
-                let f: &CompiledFunction<CompiledFuncEnv> = func.downcast_ref().unwrap();
+                let f = func.downcast_ref::<CompiledFunction>().unwrap();
                 f.metadata()
             })
             .collect();
@@ -584,6 +590,8 @@ impl wasmtime_environ::Compiler for Compiler {
             &translation.debuginfo,
             &functions_info,
             &memory_offset,
+            dwarf_package_bytes,
+            tunables,
         )
         .with_context(|| "failed to emit DWARF debug information")?;
 
@@ -613,11 +621,13 @@ impl wasmtime_environ::Compiler for Compiler {
                     section_id,
                     object::write::Relocation {
                         offset: u64::from(reloc.offset),
-                        size: reloc.size << 3,
-                        kind: RelocationKind::Absolute,
-                        encoding: RelocationEncoding::Generic,
                         symbol: target_symbol,
                         addend: i64::from(reloc.addend),
+                        flags: RelocationFlags::Generic {
+                            size: reloc.size << 3,
+                            kind: RelocationKind::Absolute,
+                            encoding: RelocationEncoding::Generic,
+                        },
                     },
                 )?;
             }
@@ -629,6 +639,68 @@ impl wasmtime_environ::Compiler for Compiler {
     fn create_systemv_cie(&self) -> Option<gimli::write::CommonInformationEntry> {
         self.isa.create_systemv_cie()
     }
+
+    fn compile_wasm_to_builtin(
+        &self,
+        index: BuiltinFunctionIndex,
+    ) -> Result<Box<dyn Any + Send>, CompileError> {
+        let isa = &*self.isa;
+        let ptr_size = isa.pointer_bytes();
+        let pointer_type = isa.pointer_type();
+        let sig = BuiltinFunctionSignatures::new(isa).signature(index);
+
+        let mut compiler = self.function_compiler();
+        let func = ir::Function::with_name_signature(Default::default(), sig.clone());
+        let (mut builder, block0) = compiler.builder(func);
+        let vmctx = builder.block_params(block0)[0];
+
+        // Debug-assert that this is the right kind of vmctx, and then
+        // additionally perform the "routine of the exit trampoline" of saving
+        // fp/pc/etc.
+        debug_assert_vmctx_kind(isa, &mut builder, vmctx, wasmtime_environ::VMCONTEXT_MAGIC);
+        let limits = builder.ins().load(
+            pointer_type,
+            MemFlags::trusted(),
+            vmctx,
+            ptr_size.vmcontext_runtime_limits(),
+        );
+        save_last_wasm_exit_fp_and_pc(&mut builder, pointer_type, &ptr_size, limits);
+
+        // Now it's time to delegate to the actual builtin. Builtins are stored
+        // in an array in all `VMContext`s. First load the base pointer of the
+        // array and then load the entry of the array that correspons to this
+        // builtin.
+        let mem_flags = ir::MemFlags::trusted().with_readonly();
+        let array_addr = builder.ins().load(
+            pointer_type,
+            mem_flags,
+            vmctx,
+            i32::try_from(ptr_size.vmcontext_builtin_functions()).unwrap(),
+        );
+        let body_offset = i32::try_from(index.index() * pointer_type.bytes()).unwrap();
+        let func_addr = builder
+            .ins()
+            .load(pointer_type, mem_flags, array_addr, body_offset);
+
+        // Forward all our own arguments to the libcall itself, and then return
+        // all the same results as the libcall.
+        let block_params = builder.block_params(block0).to_vec();
+        let sig = builder.func.import_signature(sig);
+        let call = builder.ins().call_indirect(sig, func_addr, &block_params);
+        let results = builder.func.dfg.inst_results(call).to_vec();
+        builder.ins().return_(&results);
+        builder.finalize();
+
+        Ok(Box::new(compiler.finish()?))
+    }
+
+    fn compiled_function_relocation_targets<'a>(
+        &'a self,
+        func: &'a dyn Any,
+    ) -> Box<dyn Iterator<Item = RelocationTarget> + 'a> {
+        let func = func.downcast_ref::<CompiledFunction>().unwrap();
+        Box::new(func.relocations().map(|r| r.reloc_target))
+    }
 }
 
 #[cfg(feature = "incremental-cache")]
@@ -723,7 +795,7 @@ impl Compiler {
         &self,
         ty: &WasmFuncType,
         host_fn: usize,
-    ) -> Result<CompiledFunction<CompiledFuncEnv>, CompileError> {
+    ) -> Result<CompiledFunction, CompileError> {
         let isa = &*self.isa;
         let pointer_type = isa.pointer_type();
         let native_call_sig = native_call_signature(isa, ty);
@@ -734,7 +806,7 @@ impl Compiler {
         let (mut builder, block0) = compiler.builder(func);
         let args = builder.func.dfg.block_params(block0).to_vec();
 
-        let ret = NativeRet::classify(pointer_type, ty);
+        let ret = NativeRet::classify(isa, ty);
         let wasm_args = &ret.native_args(&args)[2..];
 
         let (values_vec_ptr, values_vec_len) =
@@ -786,7 +858,7 @@ impl Compiler {
         &self,
         ty: &WasmFuncType,
         host_fn: usize,
-    ) -> Result<CompiledFunction<CompiledFuncEnv>, CompileError> {
+    ) -> Result<CompiledFunction, CompileError> {
         let isa = &*self.isa;
         let pointer_type = isa.pointer_type();
         let wasm_call_sig = wasm_call_signature(isa, ty, &self.tunables);
@@ -897,21 +969,29 @@ impl Compiler {
         values_vec_ptr: Value,
         values_vec_capacity: Value,
     ) {
+        debug_assert_eq!(types.len(), values.len());
         debug_assert_enough_capacity_for_length(builder, types.len(), values_vec_capacity);
 
         // Note that loads and stores are unconditionally done in the
         // little-endian format rather than the host's native-endianness,
         // despite this load/store being unrelated to execution in wasm itself.
-        // For more details on this see the `ValRaw` type in the
-        // `wasmtime-runtime` crate.
-        let mut mflags = MemFlags::trusted();
-        mflags.set_endianness(ir::Endianness::Little);
+        // For more details on this see the `ValRaw` type in
+        // `wasmtime::runtime::vm`.
+        let flags = ir::MemFlags::new()
+            .with_notrap()
+            .with_endianness(ir::Endianness::Little);
 
         let value_size = mem::size_of::<u128>();
-        for (i, val) in values.iter().copied().enumerate() {
-            builder
-                .ins()
-                .store(mflags, val, values_vec_ptr, (i * value_size) as i32);
+        for (i, (val, ty)) in values.iter().copied().zip(types).enumerate() {
+            crate::unbarriered_store_type_at_offset(
+                &*self.isa,
+                &mut builder.cursor(),
+                *ty,
+                flags,
+                values_vec_ptr,
+                i32::try_from(i * value_size).unwrap(),
+                val,
+            );
         }
     }
 
@@ -936,18 +1016,20 @@ impl Compiler {
 
         // Note that this is little-endian like `store_values_to_array` above,
         // see notes there for more information.
-        let mut mflags = MemFlags::trusted();
-        mflags.set_endianness(ir::Endianness::Little);
+        let flags = MemFlags::new()
+            .with_notrap()
+            .with_endianness(ir::Endianness::Little);
 
         let mut results = Vec::new();
-        for (i, r) in types.iter().enumerate() {
-            let load = builder.ins().load(
-                value_type(isa, *r),
-                mflags,
+        for (i, ty) in types.iter().enumerate() {
+            results.push(crate::unbarriered_load_type_at_offset(
+                isa,
+                &mut builder.cursor(),
+                *ty,
+                flags,
                 values_vec_ptr,
-                (i * value_size) as i32,
-            );
-            results.push(load);
+                i32::try_from(i * value_size).unwrap(),
+            ));
         }
         results
     }
@@ -976,20 +1058,6 @@ impl Compiler {
     }
 }
 
-/// The compiled function environment.
-pub struct CompiledFuncEnv {
-    /// Map to resolve external name references.
-    map: PrimaryMap<UserExternalNameRef, UserExternalName>,
-}
-
-impl wasmtime_cranelift_shared::CompiledFuncEnv for CompiledFuncEnv {
-    fn resolve_user_external_name_ref(&self, external: ir::UserExternalNameRef) -> (u32, u32) {
-        let UserExternalName { index, namespace } = self.map[external];
-
-        (namespace, index)
-    }
-}
-
 struct FunctionCompiler<'a> {
     compiler: &'a Compiler,
     cx: CompilerContext,
@@ -1010,7 +1078,7 @@ impl FunctionCompiler<'_> {
         (builder, block0)
     }
 
-    fn finish(self) -> Result<CompiledFunction<CompiledFuncEnv>, CompileError> {
+    fn finish(self) -> Result<CompiledFunction, CompileError> {
         let (info, func) = self.finish_with_info(None)?;
         assert!(info.stack_maps.is_empty());
         Ok(func)
@@ -1019,7 +1087,7 @@ impl FunctionCompiler<'_> {
     fn finish_with_info(
         mut self,
         body_and_tunables: Option<(&FunctionBody<'_>, &Tunables)>,
-    ) -> Result<(WasmFunctionInfo, CompiledFunction<CompiledFuncEnv>), CompileError> {
+    ) -> Result<(WasmFunctionInfo, CompiledFunction), CompileError> {
         let context = &mut self.cx.codegen_context;
         let isa = &*self.compiler.isa;
         let (_, _code_buf) =
@@ -1036,11 +1104,11 @@ impl FunctionCompiler<'_> {
         };
 
         let alignment = compiled_code.buffer.alignment.max(preferred_alignment);
-        let env = CompiledFuncEnv {
-            map: context.func.params.user_named_funcs().clone(),
-        };
-        let mut compiled_function =
-            CompiledFunction::new(compiled_code.buffer.clone(), env, alignment);
+        let mut compiled_function = CompiledFunction::new(
+            compiled_code.buffer.clone(),
+            context.func.params.user_named_funcs().clone(),
+            alignment,
+        );
 
         if let Some((body, tunables)) = body_and_tunables {
             let data = body.get_binary_reader();
@@ -1131,7 +1199,7 @@ fn declare_and_call(
 ) -> ir::Inst {
     let name = ir::ExternalName::User(builder.func.declare_imported_user_function(
         ir::UserExternalName {
-            namespace: 0,
+            namespace: crate::NS_WASM_FUNC,
             index: func_index,
         },
     ));
@@ -1227,7 +1295,7 @@ fn save_last_wasm_exit_fp_and_pc(
         trampoline_fp,
         // The FP always points to the next older FP for all supported
         // targets. See assertion in
-        // `crates/runtime/src/traphandlers/backtrace.rs`.
+        // `crates/wasmtime/src/runtime/vm/traphandlers/backtrace.rs`.
         0,
     );
     builder.ins().store(
@@ -1248,12 +1316,15 @@ fn save_last_wasm_exit_fp_and_pc(
 
 enum NativeRet {
     Bare,
-    Retptr { offsets: Vec<u32>, size: u32 },
+    Retptr {
+        slots: Vec<(i32, ir::Type)>,
+        size: u32,
+    },
 }
 
 impl NativeRet {
-    fn classify(pointer_type: ir::Type, ty: &WasmFuncType) -> NativeRet {
-        fn align_to(val: u32, align: u32) -> u32 {
+    fn classify(isa: &dyn TargetIsa, ty: &WasmFuncType) -> NativeRet {
+        fn align_to(val: i32, align: i32) -> i32 {
             (val + (align - 1)) & !(align - 1)
         }
 
@@ -1264,20 +1335,17 @@ impl NativeRet {
                 let mut offsets = Vec::new();
                 let mut max_align = 1;
                 for ty in other[1..].iter() {
-                    let size = match ty {
-                        WasmValType::I32 | WasmValType::F32 => 4,
-                        WasmValType::I64 | WasmValType::F64 => 8,
-                        WasmValType::Ref(_) => pointer_type.bytes(),
-                        WasmValType::V128 => 16,
-                    };
+                    let ty = crate::value_type(isa, *ty);
+                    let size = ty.bytes();
+                    let size = i32::try_from(size).unwrap();
                     offset = align_to(offset, size);
-                    offsets.push(offset);
+                    offsets.push((offset, ty));
                     offset += size;
                     max_align = max_align.max(size);
                 }
                 NativeRet::Retptr {
-                    offsets,
-                    size: align_to(offset, max_align),
+                    slots: offsets,
+                    size: u32::try_from(align_to(offset, max_align)).unwrap(),
                 }
             }
         }
@@ -1300,14 +1368,13 @@ impl NativeRet {
             NativeRet::Bare => {
                 builder.ins().return_(&results);
             }
-            NativeRet::Retptr { offsets, .. } => {
+            NativeRet::Retptr { slots, .. } => {
                 let ptr = *builder.func.dfg.block_params(block0).last().unwrap();
                 let (first, rest) = results.split_first().unwrap();
-                assert_eq!(rest.len(), offsets.len());
-                for (arg, offset) in rest.iter().zip(offsets) {
-                    builder
-                        .ins()
-                        .store(MemFlags::trusted(), *arg, ptr, *offset as i32);
+                assert_eq!(rest.len(), slots.len());
+                for (arg, (offset, ty)) in rest.iter().zip(slots) {
+                    assert_eq!(builder.func.dfg.value_type(*arg), *ty);
+                    builder.ins().store(MemFlags::trusted(), *arg, ptr, *offset);
                 }
                 builder.ins().return_(&[*first]);
             }
