@@ -3,7 +3,7 @@ use super::attr::clone_attr_string;
 use super::{Reader, TransformError};
 use anyhow::{Context, Error};
 use gimli::{
-    write, DebugLine, DebugLineOffset, DebugLineStr, DebugStr, DebugStrOffsets,
+    write, AttributeValue::DebugLineRef, DebugLine, DebugLineOffset, DebugStr,
     DebuggingInformationEntry, LineEncoding, Unit,
 };
 use wasmtime_environ::{DefinedFuncIndex, EntityRef};
@@ -40,48 +40,57 @@ enum ReadLineProgramState {
 }
 
 pub(crate) fn clone_line_program<R>(
+    dwarf: &gimli::Dwarf<R>,
+    skeleton_dwarf: &gimli::Dwarf<R>,
     unit: &Unit<R, R::Offset>,
     root: &DebuggingInformationEntry<R>,
+    skeleton_die: Option<&DebuggingInformationEntry<R>>,
     addr_tr: &AddressTransform,
     out_encoding: gimli::Encoding,
     debug_str: &DebugStr<R>,
-    debug_str_offsets: &DebugStrOffsets<R>,
-    debug_line_str: &DebugLineStr<R>,
     debug_line: &DebugLine<R>,
     out_strings: &mut write::StringTable,
 ) -> Result<(write::LineProgram, DebugLineOffset, Vec<write::FileId>, u64), Error>
 where
     R: Reader,
 {
-    let offset = match root.attr_value(gimli::DW_AT_stmt_list)? {
-        Some(gimli::AttributeValue::DebugLineRef(offset)) => offset,
+    // Where are the "location" attributes
+    let (location_die, location_dwarf) = match skeleton_die {
+        Some(die) => (die, skeleton_dwarf),
+        _ => (root, dwarf),
+    };
+
+    let offset = match location_die.attr_value(gimli::DW_AT_stmt_list)? {
+        Some(DebugLineRef(offset)) => offset,
+        Some(gimli::AttributeValue::SecOffset(offset)) => DebugLineOffset(offset),
         _ => {
             return Err(TransformError("Debug line offset is not found").into());
         }
     };
-    let comp_dir = root.attr_value(gimli::DW_AT_comp_dir)?;
+    let comp_dir = location_die.attr_value(gimli::DW_AT_comp_dir)?;
     let comp_name = root.attr_value(gimli::DW_AT_name)?;
     let out_comp_dir = match &comp_dir {
         Some(comp_dir) => Some(clone_attr_string(
             comp_dir,
             gimli::DW_FORM_strp,
             unit,
-            debug_str,
-            debug_str_offsets,
-            debug_line_str,
+            location_dwarf,
             out_strings,
         )?),
         None => None,
     };
-    let out_comp_name = clone_attr_string(
-        comp_name.as_ref().context("missing DW_AT_name attribute")?,
-        gimli::DW_FORM_strp,
-        unit,
-        debug_str,
-        debug_str_offsets,
-        debug_line_str,
-        out_strings,
-    )?;
+    let out_comp_name = match comp_name {
+        Some(_) => clone_attr_string(
+            comp_name
+                .as_ref()
+                .context("failed to read DW_AT_name attribute")?,
+            gimli::DW_FORM_strp,
+            unit,
+            dwarf,
+            out_strings,
+        )?,
+        _ => gimli::write::LineString::String("missing DW_AT_name attribute".into()),
+    };
 
     let program = debug_line.program(
         offset,
@@ -114,9 +123,7 @@ where
                 dir_attr,
                 gimli::DW_FORM_string,
                 unit,
-                debug_str,
-                debug_str_offsets,
-                debug_line_str,
+                location_dwarf,
                 out_strings,
             )?);
             dirs.push(dir_id);
@@ -132,9 +139,7 @@ where
                     &file_entry.path_name(),
                     gimli::DW_FORM_string,
                     unit,
-                    debug_str,
-                    debug_str_offsets,
-                    debug_line_str,
+                    location_dwarf,
                     out_strings,
                 )?,
                 dir_id,
