@@ -3,88 +3,105 @@
 macro_rules! foreach_builtin_function {
     ($mac:ident) => {
         $mac! {
-            /// Returns an index for wasm's `memory.grow` builtin function.
+            // Returns an index for wasm's `memory.grow` builtin function.
             memory32_grow(vmctx: vmctx, delta: i64, index: i32) -> pointer;
-            /// Returns an index for wasm's `table.copy` when both tables are locally
-            /// defined.
+            // Returns an index for wasm's `table.copy` when both tables are locally
+            // defined.
             table_copy(vmctx: vmctx, dst_index: i32, src_index: i32, dst: i32, src: i32, len: i32);
-            /// Returns an index for wasm's `table.init`.
+            // Returns an index for wasm's `table.init`.
             table_init(vmctx: vmctx, table: i32, elem: i32, dst: i32, src: i32, len: i32);
-            /// Returns an index for wasm's `elem.drop`.
+            // Returns an index for wasm's `elem.drop`.
             elem_drop(vmctx: vmctx, elem: i32);
-            /// Returns an index for wasm's `memory.copy`
+            // Returns an index for wasm's `memory.copy`
             memory_copy(vmctx: vmctx, dst_index: i32, dst: i64, src_index: i32, src: i64, len: i64);
-            /// Returns an index for wasm's `memory.fill` instruction.
+            // Returns an index for wasm's `memory.fill` instruction.
             memory_fill(vmctx: vmctx, memory: i32, dst: i64, val: i32, len: i64);
-            /// Returns an index for wasm's `memory.init` instruction.
+            // Returns an index for wasm's `memory.init` instruction.
             memory_init(vmctx: vmctx, memory: i32, data: i32, dst: i64, src: i32, len: i32);
-            /// Returns a value for wasm's `ref.func` instruction.
+            // Returns a value for wasm's `ref.func` instruction.
             ref_func(vmctx: vmctx, func: i32) -> pointer;
-            /// Returns an index for wasm's `data.drop` instruction.
+            // Returns an index for wasm's `data.drop` instruction.
             data_drop(vmctx: vmctx, data: i32);
-            /// Returns a table entry after lazily initializing it.
+            // Returns a table entry after lazily initializing it.
             table_get_lazy_init_func_ref(vmctx: vmctx, table: i32, index: i32) -> pointer;
-            /// Returns an index for Wasm's `table.grow` instruction for `funcref`s.
+            // Returns an index for Wasm's `table.grow` instruction for `funcref`s.
             table_grow_func_ref(vmctx: vmctx, table: i32, delta: i32, init: pointer) -> i32;
-            /// Returns an index for Wasm's `table.fill` instruction for `funcref`s.
+            // Returns an index for Wasm's `table.fill` instruction for `funcref`s.
             table_fill_func_ref(vmctx: vmctx, table: i32, dst: i32, val: pointer, len: i32);
-            /// Returns an index for wasm's `memory.atomic.notify` instruction.
+            // Returns an index for wasm's `memory.atomic.notify` instruction.
+            #[cfg(feature = "threads")]
             memory_atomic_notify(vmctx: vmctx, memory: i32, addr: i64, count: i32) -> i32;
-            /// Returns an index for wasm's `memory.atomic.wait32` instruction.
+            // Returns an index for wasm's `memory.atomic.wait32` instruction.
+            #[cfg(feature = "threads")]
             memory_atomic_wait32(vmctx: vmctx, memory: i32, addr: i64, expected: i32, timeout: i64) -> i32;
-            /// Returns an index for wasm's `memory.atomic.wait64` instruction.
+            // Returns an index for wasm's `memory.atomic.wait64` instruction.
+            #[cfg(feature = "threads")]
             memory_atomic_wait64(vmctx: vmctx, memory: i32, addr: i64, expected: i64, timeout: i64) -> i32;
-            /// Invoked when fuel has run out while executing a function.
+            // Invoked when fuel has run out while executing a function.
             out_of_gas(vmctx: vmctx);
-            /// Invoked when we reach a new epoch.
+            // Invoked when we reach a new epoch.
             new_epoch(vmctx: vmctx) -> i64;
-            /// Invoked before malloc returns.
+            // Invoked before malloc returns.
+            #[cfg(feature = "wmemcheck")]
             check_malloc(vmctx: vmctx, addr: i32, len: i32) -> i32;
-            /// Invoked before the free returns.
+            // Invoked before the free returns.
+            #[cfg(feature = "wmemcheck")]
             check_free(vmctx: vmctx, addr: i32) -> i32;
-            /// Invoked before a load is executed.
+            // Invoked before a load is executed.
+            #[cfg(feature = "wmemcheck")]
             check_load(vmctx: vmctx, num_bytes: i32, addr: i32, offset: i32) -> i32;
-            /// Invoked before a store is executed.
+            // Invoked before a store is executed.
+            #[cfg(feature = "wmemcheck")]
             check_store(vmctx: vmctx, num_bytes: i32, addr: i32, offset: i32) -> i32;
-            /// Invoked after malloc is called.
+            // Invoked after malloc is called.
+            #[cfg(feature = "wmemcheck")]
             malloc_start(vmctx: vmctx);
-            /// Invoked after free is called.
+            // Invoked after free is called.
+            #[cfg(feature = "wmemcheck")]
             free_start(vmctx: vmctx);
-            /// Invoked when wasm stack pointer is updated.
+            // Invoked when wasm stack pointer is updated.
+            #[cfg(feature = "wmemcheck")]
             update_stack_pointer(vmctx: vmctx, value: i32);
-            /// Invoked before memory.grow is called.
+            // Invoked before memory.grow is called.
+            #[cfg(feature = "wmemcheck")]
             update_mem_size(vmctx: vmctx, num_bytes: i32);
 
-            /// Returns an index to drop a `VMExternRef`.
+            // Drop a non-stack GC reference (eg an overwritten table entry)
+            // once it will no longer be used again. (Note: `val` is not a
+            // `reference` because it needn't appear in any stack maps, as it
+            // must not be live after this call.)
             #[cfg(feature = "gc")]
-            drop_externref(vmctx: vmctx, val: pointer);
+            drop_gc_ref(vmctx: vmctx, val: pointer);
 
-            /// Returns an index to do a GC and then insert a `VMExternRef` into the
-            /// `VMExternRefActivationsTable`.
+            // Do a GC, treating the optional `root` as a GC root and returning
+            // the updated `root` (so that, in the case of moving collectors,
+            // callers have a valid version of `root` again).
             #[cfg(feature = "gc")]
-            activations_table_insert_with_gc(vmctx: vmctx, val: reference);
+            gc(vmctx: vmctx, root: reference) -> reference;
 
-            /// Returns an index for Wasm's `global.get` instruction for `externref`s.
+            // Implementation of Wasm's `global.get` instruction for globals
+            // containing GC references.
             #[cfg(feature = "gc")]
-            externref_global_get(vmctx: vmctx, global: i32) -> reference;
+            gc_ref_global_get(vmctx: vmctx, global: i32) -> reference;
 
-            /// Returns an index for Wasm's `global.get` instruction for `externref`s.
+            // Implementation of Wasm's `global.set` instruction for globals
+            // containing GC references.
             #[cfg(feature = "gc")]
-            externref_global_set(vmctx: vmctx, global: i32, val: reference);
+            gc_ref_global_set(vmctx: vmctx, global: i32, val: reference);
 
-            /// Returns an index for Wasm's `table.grow` instruction for `externref`s.
+            // Returns an index for Wasm's `table.grow` instruction for GC references.
             #[cfg(feature = "gc")]
-            table_grow_externref(vmctx: vmctx, table: i32, delta: i32, init: reference) -> i32;
+            table_grow_gc_ref(vmctx: vmctx, table: i32, delta: i32, init: reference) -> i32;
 
-            /// Returns an index for Wasm's `table.fill` instruction for `externref`s.
+            // Returns an index for Wasm's `table.fill` instruction for GC references.
             #[cfg(feature = "gc")]
-            table_fill_externref(vmctx: vmctx, table: i32, dst: i32, val: reference, len: i32);
+            table_fill_gc_ref(vmctx: vmctx, table: i32, dst: i32, val: reference, len: i32);
         }
     };
 }
 
 /// An index type for builtin functions.
-#[derive(Copy, Clone, Debug)]
+#[derive(Copy, Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
 pub struct BuiltinFunctionIndex(u32);
 
 impl BuiltinFunctionIndex {
@@ -112,6 +129,17 @@ macro_rules! declare_indexes {
                 0;
                 $( $( #[$attr] )* $name; )*
             );
+
+            /// Returns a symbol name for this builtin.
+            pub fn name(&self) -> &'static str {
+                $(
+                    $( #[$attr] )*
+                    if *self == BuiltinFunctionIndex::$name() {
+                        return stringify!($name);
+                    }
+                )*
+                unreachable!()
+            }
         }
     };
 
@@ -140,6 +168,7 @@ macro_rules! declare_indexes {
          )*
     ) => {
         $( #[$this_attr] )*
+        #[allow(missing_docs)]
         pub const fn $this_name() -> Self {
             Self($index)
         }
