@@ -119,7 +119,6 @@ use crate::component::translate::*;
 use crate::fact;
 use crate::EntityType;
 use std::collections::HashSet;
-use wasmparser::WasmFeatures;
 
 /// Metadata information about a fused adapter.
 #[derive(Debug, Clone, Hash, Eq, PartialEq)]
@@ -212,13 +211,10 @@ impl<'data> Translator<'_, 'data> {
             // specifically enabled here since the adapter module is highly
             // likely to use that if anything is actually indirected through
             // memory.
-            let mut validator = Validator::new_with_features(WasmFeatures {
-                multi_memory: true,
-                ..*self.validator.features()
-            });
+            self.validator.reset();
             let translation = ModuleEnvironment::new(
                 self.tunables,
-                &mut validator,
+                &mut self.validator,
                 self.types.module_types_builder(),
             )
             .translate(Parser::new(0), wasm)
@@ -259,7 +255,9 @@ fn fact_import_to_core_def(
 ) -> dfg::CoreDef {
     let mut simple_intrinsic = |trampoline: dfg::Trampoline| {
         let signature = ty.unwrap_func();
-        let index = dfg.trampolines.push((signature, trampoline));
+        let index = dfg
+            .trampolines
+            .push((signature.unwrap_module_type_index(), trampoline));
         dfg::CoreDef::Trampoline(index)
     };
     match import {
@@ -285,7 +283,7 @@ fn fact_import_to_core_def(
             let to = dfg.memories.push(unwrap_memory(to));
             let signature = ty.unwrap_func();
             let index = dfg.trampolines.push((
-                signature,
+                signature.unwrap_module_type_index(),
                 dfg::Trampoline::Transcoder {
                     op: *op,
                     from,
