@@ -21,9 +21,10 @@
 use crate::component::dfg::CoreDef;
 use crate::component::{
     Adapter, AdapterOptions as AdapterOptionsDfg, ComponentTypesBuilder, FlatType, InterfaceType,
-    StringEncoding, TypeFuncIndex,
+    StringEncoding, Transcode, TypeFuncIndex,
 };
 use crate::fact::transcode::Transcoder;
+use crate::prelude::*;
 use crate::{EntityRef, FuncIndex, GlobalIndex, MemoryIndex, PrimaryMap};
 use std::borrow::Cow;
 use std::collections::HashMap;
@@ -35,8 +36,6 @@ mod trampoline;
 mod transcode;
 mod traps;
 
-pub use self::transcode::{FixedEncoding, Transcode};
-
 /// Representation of an adapter module.
 pub struct Module<'a> {
     /// Whether or not debug code is inserted into the adapters themselves.
@@ -256,6 +255,7 @@ impl<'a> Module<'a> {
             GlobalType {
                 val_type: ValType::I32,
                 mutable: true,
+                shared: false,
             },
             CoreDef::InstanceFlags(*instance),
         );
@@ -268,6 +268,7 @@ impl<'a> Module<'a> {
                     maximum: None,
                     shared: false,
                     memory64: *memory64,
+                    page_size_log2: None,
                 },
                 memory.clone().into(),
             )
