@@ -5,6 +5,18 @@ fn main() {
     let mut build = cc::Build::new();
     let arch = env::var("CARGO_CFG_TARGET_ARCH").unwrap();
     let os = env::var("CARGO_CFG_TARGET_OS").unwrap();
+
+    // NB: Technically `cfg(sanitize = "address")` is not stable and requires a
+    // `#![feature]` but sort of abuse the fact that cfgs are "leaked" through
+    // into Cargo ungated via `--print cfg`. Translate that to `cfg(asan)` for
+    // us to write down in the code.
+    match env::var("CARGO_CFG_SANITIZE") {
+        Ok(s) if s == "address" => {
+            println!("cargo:rustc-cfg=asan");
+        }
+        _ => {}
+    }
+
     if os == "windows" {
         println!("cargo:rerun-if-changed=src/windows.c");
         build.file("src/windows.c");
