@@ -1,7 +1,8 @@
 //! Generate a configuration for both Wasmtime and the Wasm module to execute.
 
 use super::{
-    CodegenSettings, InstanceAllocationStrategy, MemoryConfig, ModuleConfig, UnalignedMemoryCreator,
+    AsyncConfig, CodegenSettings, InstanceAllocationStrategy, MemoryConfig, ModuleConfig,
+    NormalMemoryConfig, UnalignedMemoryCreator,
 };
 use crate::oracles::{StoreLimits, Timeout};
 use anyhow::Result;
@@ -100,12 +101,12 @@ impl Config {
         self.module_config.generate(input, default_fuel)
     }
 
-    /// Tests whether this configuration is capable of running all spec tests.
-    pub fn is_spectest_compliant(&self) -> bool {
+    /// Tests whether this configuration is capable of running all wast tests.
+    pub fn is_wast_test_compliant(&self) -> bool {
         let config = &self.module_config.config;
 
         // Check for wasm features that must be disabled to run spec tests
-        if config.memory64_enabled || config.threads_enabled {
+        if config.memory64_enabled {
             return false;
         }
 
@@ -114,6 +115,8 @@ impl Config {
             || !config.reference_types_enabled
             || !config.multi_value_enabled
             || !config.simd_enabled
+            || !config.threads_enabled
+            || config.max_memories <= 1
         {
             return false;
         }
@@ -170,7 +173,9 @@ impl Config {
                 self.wasmtime.memory_guaranteed_dense_image_size,
             ))
             .allocation_strategy(self.wasmtime.strategy.to_wasmtime())
-            .generate_address_map(self.wasmtime.generate_address_map);
+            .generate_address_map(self.wasmtime.generate_address_map)
+            .cache_call_indirects(self.wasmtime.cache_call_indirects)
+            .max_call_indirect_cache_slots(self.wasmtime.max_call_indirect_cache_slots);
 
         if !self.module_config.config.simd_enabled {
             cfg.wasm_relaxed_simd(false);
@@ -182,6 +187,13 @@ impl Config {
 
         self.wasmtime.codegen.configure(&mut cfg);
 
+        // Determine whether we will actually enable PCC -- this is
+        // disabled if the module requires memory64, which is not yet
+        // compatible (due to the need for dynamic checks).
+        let pcc = cfg!(feature = "fuzz-pcc")
+            && self.wasmtime.pcc
+            && !self.module_config.config.memory64_enabled;
+
         // Only set cranelift specific flags when the Cranelift strategy is
         // chosen.
         if cranelift_strategy {
@@ -213,8 +225,12 @@ impl Config {
                     );
                 }
             }
+
+            cfg.cranelift_pcc(pcc);
         }
 
+        self.wasmtime.async_config.configure(&mut cfg);
+
         // Vary the memory configuration, but only if threads are not enabled.
         // When the threads proposal is enabled we might generate shared memory,
         // which is less amenable to different memory configurations:
@@ -226,7 +242,20 @@ impl Config {
         //   `CustomUnaligned` variant isn't actually safe to use with a shared
         //   memory.
         if !self.module_config.config.threads_enabled {
-            match &self.wasmtime.memory_config {
+            // If PCC is enabled, force other options to be compatible: PCC is currently only
+            // supported when bounds checks are elided.
+            let memory_config = if pcc {
+                MemoryConfig::Normal(NormalMemoryConfig {
+                    static_memory_maximum_size: Some(4 << 30), // 4 GiB
+                    static_memory_guard_size: Some(2 << 30),   // 2 GiB
+                    dynamic_memory_guard_size: Some(0),
+                    guard_before_linear_memory: false,
+                })
+            } else {
+                self.wasmtime.memory_config.clone()
+            };
+
+            match &memory_config {
                 MemoryConfig::Normal(memory_config) => {
                     cfg.static_memory_maximum_size(
                         memory_config.static_memory_maximum_size.unwrap_or(0),
@@ -260,20 +289,26 @@ impl Config {
     /// Configures a store based on this configuration.
     pub fn configure_store(&self, store: &mut Store<StoreLimits>) {
         store.limiter(|s| s as &mut dyn wasmtime::ResourceLimiter);
-        if self.wasmtime.consume_fuel {
-            store.set_fuel(u64::MAX).unwrap();
-        }
-        if self.wasmtime.epoch_interruption {
-            // Without fuzzing of async execution, we can't test the
-            // "update deadline and continue" behavior, but we can at
-            // least test the codegen paths and checks with the
-            // trapping behavior, which works synchronously too. We'll
-            // set the deadline one epoch tick in the future; then
-            // this works exactly like an interrupt flag. We expect no
-            // traps/interrupts unless we bump the epoch, which we do
-            // as one particular Timeout mode (`Timeout::Epoch`).
-            store.epoch_deadline_trap();
-            store.set_epoch_deadline(1);
+        match self.wasmtime.async_config {
+            AsyncConfig::Disabled => {
+                if self.wasmtime.consume_fuel {
+                    store.set_fuel(u64::MAX).unwrap();
+                }
+                if self.wasmtime.epoch_interruption {
+                    store.epoch_deadline_trap();
+                    store.set_epoch_deadline(1);
+                }
+            }
+            AsyncConfig::YieldWithFuel(amt) => {
+                assert!(self.wasmtime.consume_fuel);
+                store.fuel_async_yield_interval(Some(amt)).unwrap();
+                store.set_fuel(amt).unwrap();
+            }
+            AsyncConfig::YieldWithEpochs { ticks, .. } => {
+                assert!(self.wasmtime.epoch_interruption);
+                store.set_epoch_deadline(ticks);
+                store.epoch_deadline_async_yield_and_update(ticks);
+            }
         }
     }
 
@@ -324,13 +359,29 @@ impl Config {
     pub fn disable_unimplemented_winch_proposals(&mut self) {
         self.module_config.config.simd_enabled = false;
         self.module_config.config.relaxed_simd_enabled = false;
-        self.module_config.config.memory64_enabled = false;
         self.module_config.config.gc_enabled = false;
         self.module_config.config.threads_enabled = false;
         self.module_config.config.tail_call_enabled = false;
         self.module_config.config.exceptions_enabled = false;
         self.module_config.config.reference_types_enabled = false;
     }
+
+    /// Updates this configuration to forcibly enable async support. Only useful
+    /// in fuzzers which do async calls.
+    pub fn enable_async(&mut self, u: &mut Unstructured<'_>) -> arbitrary::Result<()> {
+        if self.wasmtime.consume_fuel || u.arbitrary()? {
+            self.wasmtime.async_config =
+                AsyncConfig::YieldWithFuel(u.int_in_range(1000..=100_000)?);
+            self.wasmtime.consume_fuel = true;
+        } else {
+            self.wasmtime.async_config = AsyncConfig::YieldWithEpochs {
+                dur: Duration::from_millis(u.int_in_range(1..=10)?),
+                ticks: u.int_in_range(1..=10)?,
+            };
+            self.wasmtime.epoch_interruption = true;
+        }
+        Ok(())
+    }
 }
 
 impl<'a> Arbitrary<'a> for Config {
@@ -412,7 +463,7 @@ pub struct WasmtimeConfig {
     canonicalize_nans: bool,
     interruptable: bool,
     pub(crate) consume_fuel: bool,
-    epoch_interruption: bool,
+    pub(crate) epoch_interruption: bool,
     /// The Wasmtime memory configuration to use.
     pub memory_config: MemoryConfig,
     force_jump_veneers: bool,
@@ -427,6 +478,17 @@ pub struct WasmtimeConfig {
     native_unwind_info: bool,
     /// Configuration for the compiler to use.
     pub compiler_strategy: CompilerStrategy,
+    /// Whether we enable indirect-call caching.
+    cache_call_indirects: bool,
+    /// The maximum number of call-indirect cache slots.
+    max_call_indirect_cache_slots: usize,
+
+    /// Whether or not fuzzing should enable PCC.
+    pcc: bool,
+
+    /// Configuration for whether wasm is invoked in an async fashion and how
+    /// it's cooperatively time-sliced.
+    pub async_config: AsyncConfig,
 }
 
 impl WasmtimeConfig {
