@@ -1,6 +1,6 @@
 //! Dummy implementations of things that a Wasm module can import.
 
-use anyhow::bail;
+use anyhow::ensure;
 use wasmtime::*;
 
 /// Create a set of dummy functions/globals/etc for the given imports.
@@ -45,11 +45,13 @@ pub fn dummy_value(val_ty: ValType) -> Result<Val> {
         ValType::F32 => Val::F32(0),
         ValType::F64 => Val::F64(0),
         ValType::V128 => Val::V128(0.into()),
-        ValType::Ref(r) => match r.heap_type() {
-            _ if !r.is_nullable() => bail!("cannot construct a dummy value of type `{r}`"),
-            HeapType::Extern => Val::null_extern_ref(),
-            HeapType::NoFunc | HeapType::Func | HeapType::Concrete(_) => Val::null_func_ref(),
-        },
+        ValType::Ref(r) => {
+            ensure!(
+                r.is_nullable(),
+                "cannot construct a dummy value of type `{r}`"
+            );
+            Val::null_ref(r.heap_type())
+        }
     })
 }
 
