@@ -5,34 +5,39 @@ use crate::io::TokioIo;
 use crate::{
     bindings::http::types::{self, Method, Scheme},
     body::{HostIncomingBody, HyperIncomingBody, HyperOutgoingBody},
-    dns_error, hyper_request_error,
+    error::dns_error,
+    hyper_request_error,
 };
 use http_body_util::BodyExt;
 use hyper::header::HeaderName;
 use std::any::Any;
-use std::sync::Arc;
 use std::time::Duration;
 use tokio::net::TcpStream;
 use tokio::time::timeout;
 use wasmtime::component::{Resource, ResourceTable};
-use wasmtime_wasi::{AbortOnDropJoinHandle, Subscribe};
+use wasmtime_wasi::{runtime::AbortOnDropJoinHandle, Subscribe};
 
 /// Capture the state necessary for use in the wasi-http API implementation.
-pub struct WasiHttpCtx;
-
-pub struct OutgoingRequest {
-    pub use_tls: bool,
-    pub authority: String,
-    pub request: hyper::Request<HyperOutgoingBody>,
-    pub connect_timeout: Duration,
-    pub first_byte_timeout: Duration,
-    pub between_bytes_timeout: Duration,
+pub struct WasiHttpCtx {
+    _priv: (),
 }
 
-pub trait WasiHttpView: Send {
+impl WasiHttpCtx {
+    /// Create a new context.
+    pub fn new() -> Self {
+        Self { _priv: () }
+    }
+}
+
+/// A trait which provides internal WASI HTTP state.
+pub trait WasiHttpView {
+    /// Returns a mutable reference to the WASI HTTP context.
     fn ctx(&mut self) -> &mut WasiHttpCtx;
+
+    /// Returns a mutable reference to the WASI HTTP resource table.
     fn table(&mut self) -> &mut ResourceTable;
 
+    /// Create a new incoming request resource.
     fn new_incoming_request(
         &mut self,
         req: hyper::Request<HyperIncomingBody>,
@@ -50,6 +55,7 @@ pub trait WasiHttpView: Send {
         Ok(self.table().push(incoming_req)?)
     }
 
+    /// Create a new outgoing response resource.
     fn new_response_outparam(
         &mut self,
         result: tokio::sync::oneshot::Sender<
@@ -60,21 +66,52 @@ pub trait WasiHttpView: Send {
         Ok(id)
     }
 
+    /// Send an outgoing request.
     fn send_request(
         &mut self,
-        request: OutgoingRequest,
-    ) -> wasmtime::Result<Resource<HostFutureIncomingResponse>>
-    where
-        Self: Sized,
-    {
-        default_send_request(self, request)
+        request: hyper::Request<HyperOutgoingBody>,
+        config: OutgoingRequestConfig,
+    ) -> crate::HttpResult<HostFutureIncomingResponse> {
+        Ok(default_send_request(request, config))
     }
 
+    /// Whether a given header should be considered forbidden and not allowed.
     fn is_forbidden_header(&mut self, _name: &HeaderName) -> bool {
         false
     }
 }
 
+impl<T: ?Sized + WasiHttpView> WasiHttpView for &mut T {
+    fn ctx(&mut self) -> &mut WasiHttpCtx {
+        T::ctx(self)
+    }
+
+    fn table(&mut self) -> &mut ResourceTable {
+        T::table(self)
+    }
+
+    fn new_response_outparam(
+        &mut self,
+        result: tokio::sync::oneshot::Sender<
+            Result<hyper::Response<HyperOutgoingBody>, types::ErrorCode>,
+        >,
+    ) -> wasmtime::Result<Resource<HostResponseOutparam>> {
+        T::new_response_outparam(self, result)
+    }
+
+    fn send_request(
+        &mut self,
+        request: hyper::Request<HyperOutgoingBody>,
+        config: OutgoingRequestConfig,
+    ) -> crate::HttpResult<HostFutureIncomingResponse> {
+        T::send_request(self, request, config)
+    }
+
+    fn is_forbidden_header(&mut self, name: &HeaderName) -> bool {
+        T::is_forbidden_header(self, name)
+    }
+}
+
 /// Returns `true` when the header is forbidden according to this [`WasiHttpView`] implementation.
 pub(crate) fn is_forbidden_header(view: &mut dyn WasiHttpView, name: &HeaderName) -> bool {
     static FORBIDDEN_HEADERS: [HeaderName; 10] = [
@@ -111,45 +148,58 @@ pub(crate) fn remove_forbidden_headers(
     }
 }
 
+/// Configuration for an outgoing request.
+pub struct OutgoingRequestConfig {
+    /// Whether to use TLS for the request.
+    pub use_tls: bool,
+    /// The timeout for connecting.
+    pub connect_timeout: Duration,
+    /// The timeout until the first byte.
+    pub first_byte_timeout: Duration,
+    /// The timeout between chunks of a streaming body
+    pub between_bytes_timeout: Duration,
+}
+
+/// The default implementation of how an outgoing request is sent.
+///
+/// This implementation is used by the `wasi:http/outgoing-handler` interface
+/// default implementation.
 pub fn default_send_request(
-    view: &mut dyn WasiHttpView,
-    OutgoingRequest {
+    request: hyper::Request<HyperOutgoingBody>,
+    config: OutgoingRequestConfig,
+) -> HostFutureIncomingResponse {
+    let handle = wasmtime_wasi::runtime::spawn(async move {
+        Ok(default_send_request_handler(request, config).await)
+    });
+    HostFutureIncomingResponse::pending(handle)
+}
+
+/// The underlying implementation of how an outgoing request is sent. This should likely be spawned
+/// in a task.
+///
+/// This is called from [default_send_request] to actually send the request.
+pub async fn default_send_request_handler(
+    mut request: hyper::Request<HyperOutgoingBody>,
+    OutgoingRequestConfig {
         use_tls,
-        authority,
-        request,
         connect_timeout,
         first_byte_timeout,
         between_bytes_timeout,
-    }: OutgoingRequest,
-) -> wasmtime::Result<Resource<HostFutureIncomingResponse>> {
-    let handle = wasmtime_wasi::spawn(async move {
-        let resp = handler(
-            authority,
-            use_tls,
-            connect_timeout,
-            first_byte_timeout,
-            request,
-            between_bytes_timeout,
-        )
-        .await;
-        Ok(resp)
-    });
-
-    let fut = view.table().push(HostFutureIncomingResponse::new(handle))?;
-
-    Ok(fut)
-}
-
-async fn handler(
-    authority: String,
-    use_tls: bool,
-    connect_timeout: Duration,
-    first_byte_timeout: Duration,
-    mut request: http::Request<HyperOutgoingBody>,
-    between_bytes_timeout: Duration,
-) -> Result<IncomingResponseInternal, types::ErrorCode> {
-    let tcp_stream = TcpStream::connect(authority.clone())
+    }: OutgoingRequestConfig,
+) -> Result<IncomingResponse, types::ErrorCode> {
+    let authority = if let Some(authority) = request.uri().authority() {
+        if authority.port().is_some() {
+            authority.to_string()
+        } else {
+            let port = if use_tls { 443 } else { 80 };
+            format!("{}:{port}", authority.to_string())
+        }
+    } else {
+        return Err(types::ErrorCode::HttpRequestUriInvalid);
+    };
+    let tcp_stream = timeout(connect_timeout, TcpStream::connect(&authority))
         .await
+        .map_err(|_| types::ErrorCode::ConnectionTimeout)?
         .map_err(|e| match e.kind() {
             std::io::ErrorKind::AddrNotAvailable => {
                 dns_error("address not available".to_string(), 0)
@@ -176,28 +226,27 @@ async fn handler(
 
         #[cfg(not(any(target_arch = "riscv64", target_arch = "s390x")))]
         {
-            use tokio_rustls::rustls::OwnedTrustAnchor;
+            use rustls::pki_types::{ServerName, TrustAnchor};
 
             // derived from https://github.com/tokio-rs/tls/blob/master/tokio-rustls/examples/client/src/main.rs
             let mut root_cert_store = rustls::RootCertStore::empty();
-            root_cert_store.add_trust_anchors(webpki_roots::TLS_SERVER_ROOTS.iter().map(|ta| {
-                OwnedTrustAnchor::from_subject_spki_name_constraints(
-                    ta.subject,
-                    ta.spki,
-                    ta.name_constraints,
-                )
+            root_cert_store.extend(webpki_roots::TLS_SERVER_ROOTS.iter().map(|ta| TrustAnchor {
+                name_constraints: ta.name_constraints.to_owned(),
+                subject: ta.subject.to_owned(),
+                subject_public_key_info: ta.subject_public_key_info.to_owned(),
             }));
             let config = rustls::ClientConfig::builder()
-                .with_safe_defaults()
                 .with_root_certificates(root_cert_store)
                 .with_no_client_auth();
             let connector = tokio_rustls::TlsConnector::from(std::sync::Arc::new(config));
             let mut parts = authority.split(":");
             let host = parts.next().unwrap_or(&authority);
-            let domain = rustls::ServerName::try_from(host).map_err(|e| {
-                tracing::warn!("dns lookup error: {e:?}");
-                dns_error("invalid dns name".to_string(), 0)
-            })?;
+            let domain = ServerName::try_from(host)
+                .map_err(|e| {
+                    tracing::warn!("dns lookup error: {e:?}");
+                    dns_error("invalid dns name".to_string(), 0)
+                })?
+                .to_owned();
             let stream = connector.connect(domain, tcp_stream).await.map_err(|e| {
                 tracing::warn!("tls protocol error: {e:?}");
                 types::ErrorCode::TlsProtocolError
@@ -212,7 +261,7 @@ async fn handler(
             .map_err(|_| types::ErrorCode::ConnectionTimeout)?
             .map_err(hyper_request_error)?;
 
-            let worker = wasmtime_wasi::spawn(async move {
+            let worker = wasmtime_wasi::runtime::spawn(async move {
                 match conn.await {
                     Ok(()) => {}
                     // TODO: shouldn't throw away this error and ideally should
@@ -234,7 +283,7 @@ async fn handler(
         .map_err(|_| types::ErrorCode::ConnectionTimeout)?
         .map_err(hyper_request_error)?;
 
-        let worker = wasmtime_wasi::spawn(async move {
+        let worker = wasmtime_wasi::runtime::spawn(async move {
             match conn.await {
                 Ok(()) => {}
                 // TODO: same as above, shouldn't throw this error away.
@@ -265,9 +314,9 @@ async fn handler(
         .map_err(hyper_request_error)?
         .map(|body| body.map_err(hyper_request_error).boxed());
 
-    Ok(IncomingResponseInternal {
+    Ok(IncomingResponse {
         resp,
-        worker: Arc::new(worker),
+        worker: Some(worker),
         between_bytes_timeout,
     })
 }
@@ -317,12 +366,15 @@ impl TryInto<http::Method> for types::Method {
     }
 }
 
+/// The concrete type behind a `wasi:http/types/incoming-request` resource.
 pub struct HostIncomingRequest {
     pub(crate) parts: http::request::Parts,
+    /// The body of the incoming request.
     pub body: Option<HostIncomingBody>,
 }
 
 impl HostIncomingRequest {
+    /// Create a new `HostIncomingRequest`.
     pub fn new(
         view: &mut dyn WasiHttpView,
         mut parts: http::request::Parts,
@@ -333,37 +385,20 @@ impl HostIncomingRequest {
     }
 }
 
+/// The concrete type behind a `wasi:http/types/response-outparam` resource.
 pub struct HostResponseOutparam {
+    /// The sender for sending a response.
     pub result:
         tokio::sync::oneshot::Sender<Result<hyper::Response<HyperOutgoingBody>, types::ErrorCode>>,
 }
 
-pub struct HostOutgoingRequest {
-    pub method: Method,
-    pub scheme: Option<Scheme>,
-    pub path_with_query: Option<String>,
-    pub authority: Option<String>,
-    pub headers: FieldMap,
-    pub body: Option<HyperOutgoingBody>,
-}
-
-#[derive(Default)]
-pub struct HostRequestOptions {
-    pub connect_timeout: Option<std::time::Duration>,
-    pub first_byte_timeout: Option<std::time::Duration>,
-    pub between_bytes_timeout: Option<std::time::Duration>,
-}
-
-pub struct HostIncomingResponse {
-    pub status: u16,
-    pub headers: FieldMap,
-    pub body: Option<HostIncomingBody>,
-    pub worker: Arc<AbortOnDropJoinHandle<()>>,
-}
-
+/// The concrete type behind a `wasi:http/types/outgoing-response` resource.
 pub struct HostOutgoingResponse {
+    /// The status of the response.
     pub status: http::StatusCode,
+    /// The headers of the response.
     pub headers: FieldMap,
+    /// The body of the response.
     pub body: Option<HyperOutgoingBody>,
 }
 
@@ -390,50 +425,111 @@ impl TryFrom<HostOutgoingResponse> for hyper::Response<HyperOutgoingBody> {
     }
 }
 
-pub type FieldMap = hyper::HeaderMap;
+/// The concrete type behind a `wasi:http/types/outgoing-request` resource.
+pub struct HostOutgoingRequest {
+    /// The method of the request.
+    pub method: Method,
+    /// The scheme of the request.
+    pub scheme: Option<Scheme>,
+    /// The authority of the request.
+    pub authority: Option<String>,
+    /// The path and query of the request.
+    pub path_with_query: Option<String>,
+    /// The request headers.
+    pub headers: FieldMap,
+    /// The request body.
+    pub body: Option<HyperOutgoingBody>,
+}
 
+/// The concrete type behind a `wasi:http/types/request-options` resource.
+#[derive(Default)]
+pub struct HostRequestOptions {
+    /// How long to wait for a connection to be established.
+    pub connect_timeout: Option<std::time::Duration>,
+    /// How long to wait for the first byte of the response body.
+    pub first_byte_timeout: Option<std::time::Duration>,
+    /// How long to wait between frames of the response body.
+    pub between_bytes_timeout: Option<std::time::Duration>,
+}
+
+/// The concrete type behind a `wasi:http/types/incoming-response` resource.
+pub struct HostIncomingResponse {
+    /// The response status
+    pub status: u16,
+    /// The response headers
+    pub headers: FieldMap,
+    /// The response body
+    pub body: Option<HostIncomingBody>,
+}
+
+/// The concrete type behind a `wasi:http/types/fields` resource.
 pub enum HostFields {
+    /// A reference to the fields of a parent entry.
     Ref {
+        /// The parent resource rep.
         parent: u32,
 
+        /// The function to get the fields from the parent.
         // NOTE: there's not failure in the result here because we assume that HostFields will
         // always be registered as a child of the entry with the `parent` id. This ensures that the
         // entry will always exist while this `HostFields::Ref` entry exists in the table, thus we
         // don't need to account for failure when fetching the fields ref from the parent.
         get_fields: for<'a> fn(elem: &'a mut (dyn Any + 'static)) -> &'a mut FieldMap,
     },
+    /// An owned version of the fields.
     Owned {
+        /// The fields themselves.
         fields: FieldMap,
     },
 }
 
-pub struct IncomingResponseInternal {
+/// An owned version of `HostFields`
+pub type FieldMap = hyper::HeaderMap;
+
+/// A handle to a future incoming response.
+pub type FutureIncomingResponseHandle =
+    AbortOnDropJoinHandle<anyhow::Result<Result<IncomingResponse, types::ErrorCode>>>;
+
+/// A response that is in the process of being received.
+pub struct IncomingResponse {
+    /// The response itself.
     pub resp: hyper::Response<HyperIncomingBody>,
-    pub worker: Arc<AbortOnDropJoinHandle<()>>,
+    /// Optional worker task that continues to process the response.
+    pub worker: Option<AbortOnDropJoinHandle<()>>,
+    /// The timeout between chunks of the response.
     pub between_bytes_timeout: std::time::Duration,
 }
 
-type FutureIncomingResponseHandle =
-    AbortOnDropJoinHandle<anyhow::Result<Result<IncomingResponseInternal, types::ErrorCode>>>;
-
+/// The concrete type behind a `wasi:http/types/future-incoming-response` resource.
 pub enum HostFutureIncomingResponse {
+    /// A pending response
     Pending(FutureIncomingResponseHandle),
-    Ready(anyhow::Result<Result<IncomingResponseInternal, types::ErrorCode>>),
+    /// The response is ready.
+    ///
+    /// An outer error will trap while the inner error gets returned to the guest.
+    Ready(anyhow::Result<Result<IncomingResponse, types::ErrorCode>>),
+    /// The response has been consumed.
     Consumed,
 }
 
 impl HostFutureIncomingResponse {
-    pub fn new(handle: FutureIncomingResponseHandle) -> Self {
+    /// Create a new `HostFutureIncomingResponse` that is pending on the provided task handle.
+    pub fn pending(handle: FutureIncomingResponseHandle) -> Self {
         Self::Pending(handle)
     }
 
+    /// Create a new `HostFutureIncomingResponse` that is ready.
+    pub fn ready(result: anyhow::Result<Result<IncomingResponse, types::ErrorCode>>) -> Self {
+        Self::Ready(result)
+    }
+
+    /// Returns `true` if the response is ready.
     pub fn is_ready(&self) -> bool {
         matches!(self, Self::Ready(_))
     }
 
-    pub fn unwrap_ready(
-        self,
-    ) -> anyhow::Result<Result<IncomingResponseInternal, types::ErrorCode>> {
+    /// Unwrap the response, panicking if it is not ready.
+    pub fn unwrap_ready(self) -> anyhow::Result<Result<IncomingResponse, types::ErrorCode>> {
         match self {
             Self::Ready(res) => res,
             Self::Pending(_) | Self::Consumed => {
