@@ -1,19 +1,24 @@
+//! Bindings for WASIp0 aka Preview 0 aka `wasi_unstable`.
+//!
+//! This module is purely here for backwards compatibility in the Wasmtime CLI.
+//! You probably want to use [`preview1`](crate::preview1) instead.
+
 use crate::preview0::types::Error;
 use crate::preview1::types as snapshot1_types;
 use crate::preview1::wasi_snapshot_preview1::WasiSnapshotPreview1 as Snapshot1;
-use crate::preview1::WasiPreview1View;
+use crate::preview1::WasiP1Ctx;
 use wiggle::{GuestError, GuestPtr};
 
-pub fn add_to_linker_async<T: Send, W: WasiPreview1View>(
+pub fn add_to_linker_async<T: Send>(
     linker: &mut wasmtime::Linker<T>,
-    f: impl Fn(&mut T) -> &mut W + Copy + Send + Sync + 'static,
+    f: impl Fn(&mut T) -> &mut WasiP1Ctx + Copy + Send + Sync + 'static,
 ) -> anyhow::Result<()> {
     wasi_unstable::add_to_linker(linker, f)
 }
 
-pub fn add_to_linker_sync<T: Send, W: WasiPreview1View>(
+pub fn add_to_linker_sync<T: Send>(
     linker: &mut wasmtime::Linker<T>,
-    f: impl Fn(&mut T) -> &mut W + Copy + Send + Sync + 'static,
+    f: impl Fn(&mut T) -> &mut WasiP1Ctx + Copy + Send + Sync + 'static,
 ) -> anyhow::Result<()> {
     sync::add_wasi_unstable_to_linker(linker, f)
 }
@@ -54,7 +59,7 @@ mod sync {
     // Small wrapper around `in_tokio` to add a `Result` layer which is always
     // `Ok`
     fn in_tokio<F: Future>(future: F) -> Result<F::Output> {
-        Ok(crate::in_tokio(future))
+        Ok(crate::runtime::in_tokio(future))
     }
 }
 
