@@ -1,16 +1,17 @@
 use crate::component::matching::InstanceType;
 use crate::component::resources::{HostResourceData, HostResourceIndex, HostResourceTables};
 use crate::component::ResourceType;
-use crate::store::{StoreId, StoreOpaque};
-use crate::{FuncType, StoreContextMut};
-use anyhow::{bail, Result};
-use std::ptr::NonNull;
-use std::sync::Arc;
-use wasmtime_environ::component::{ComponentTypes, StringEncoding, TypeResourceTableIndex};
-use wasmtime_runtime::component::{
+use crate::prelude::*;
+use crate::runtime::vm::component::{
     CallContexts, ComponentInstance, InstanceFlags, ResourceTable, ResourceTables,
 };
-use wasmtime_runtime::{VMFuncRef, VMMemoryDefinition};
+use crate::runtime::vm::{VMFuncRef, VMMemoryDefinition};
+use crate::store::{StoreId, StoreOpaque};
+use crate::{FuncType, StoreContextMut};
+use alloc::sync::Arc;
+use anyhow::{bail, Result};
+use core::ptr::NonNull;
+use wasmtime_environ::component::{ComponentTypes, StringEncoding, TypeResourceTableIndex};
 
 /// Runtime representation of canonical ABI options in the component model.
 ///
@@ -88,26 +89,27 @@ impl Options {
 
         let realloc = self.realloc.unwrap();
 
+        let params = (
+            u32::try_from(old).err2anyhow()?,
+            u32::try_from(old_size).err2anyhow()?,
+            old_align,
+            u32::try_from(new_size).err2anyhow()?,
+        );
+
+        type ReallocFunc = crate::TypedFunc<(u32, u32, u32, u32), u32>;
+
+        // This call doesn't take any GC refs, and therefore we shouldn't ever
+        // need to GC before entering Wasm.
+        debug_assert!(!ReallocFunc::need_gc_before_call_raw(store.0, &params));
+
         // Invoke the wasm malloc function using its raw and statically known
         // signature.
-        let result = unsafe {
-            crate::TypedFunc::<(u32, u32, u32, u32), u32>::call_raw(
-                store,
-                realloc_ty,
-                realloc,
-                (
-                    u32::try_from(old)?,
-                    u32::try_from(old_size)?,
-                    old_align,
-                    u32::try_from(new_size)?,
-                ),
-            )?
-        };
+        let result = unsafe { ReallocFunc::call_raw(store, realloc_ty, realloc, params)? };
 
         if result % old_align != 0 {
             bail!("realloc return: result not aligned");
         }
-        let result = usize::try_from(result)?;
+        let result = usize::try_from(result).err2anyhow()?;
 
         let memory = self.memory_mut(store.0);
 
@@ -137,7 +139,7 @@ impl Options {
         // is an optional configuration in canonical ABI options.
         unsafe {
             let memory = self.memory.unwrap().as_ref();
-            std::slice::from_raw_parts(memory.base, memory.current_length())
+            core::slice::from_raw_parts(memory.base, memory.current_length())
         }
     }
 
@@ -148,7 +150,7 @@ impl Options {
         // See comments in `memory` about the unsafety
         unsafe {
             let memory = self.memory.unwrap().as_ref();
-            std::slice::from_raw_parts_mut(memory.base, memory.current_length())
+            core::slice::from_raw_parts_mut(memory.base, memory.current_length())
         }
     }
 
@@ -338,8 +340,14 @@ impl<'a, T> LowerContext<'a, T> {
     ///
     /// Note that this is a special case for `Resource<T>`. Most of the time a
     /// host value shouldn't be lowered with a lowering context.
-    pub fn host_resource_lower_own(&mut self, rep: u32) -> Result<HostResourceIndex> {
-        self.resource_tables().host_resource_lower_own(rep)
+    pub fn host_resource_lower_own(
+        &mut self,
+        rep: u32,
+        dtor: Option<NonNull<VMFuncRef>>,
+        flags: Option<InstanceFlags>,
+    ) -> Result<HostResourceIndex> {
+        self.resource_tables()
+            .host_resource_lower_own(rep, dtor, flags)
     }
 
     /// Returns the underlying resource type for the `ty` table specified.
@@ -491,8 +499,14 @@ impl<'a> LiftContext<'a> {
 
     /// Lowers a resource into the host-owned table, returning the index it was
     /// inserted at.
-    pub fn host_resource_lower_own(&mut self, rep: u32) -> Result<HostResourceIndex> {
-        self.resource_tables().host_resource_lower_own(rep)
+    pub fn host_resource_lower_own(
+        &mut self,
+        rep: u32,
+        dtor: Option<NonNull<VMFuncRef>>,
+        flags: Option<InstanceFlags>,
+    ) -> Result<HostResourceIndex> {
+        self.resource_tables()
+            .host_resource_lower_own(rep, dtor, flags)
     }
 
     /// Lowers a resource into the host-owned table, returning the index it was
