@@ -1,21 +1,22 @@
 use crate::component::func::{Func, LiftContext, LowerContext, Options};
 use crate::component::matching::InstanceType;
 use crate::component::storage::{storage_as_slice, storage_as_slice_mut};
+use crate::prelude::*;
+use crate::runtime::vm::component::ComponentInstance;
+use crate::runtime::vm::SendSyncPtr;
 use crate::{AsContextMut, StoreContext, StoreContextMut, ValRaw};
+use alloc::borrow::Cow;
+use alloc::sync::Arc;
 use anyhow::{anyhow, bail, Context, Result};
-use std::borrow::Cow;
-use std::fmt;
-use std::marker;
-use std::mem::{self, MaybeUninit};
-use std::ptr::NonNull;
-use std::str;
-use std::sync::Arc;
+use core::fmt;
+use core::marker;
+use core::mem::{self, MaybeUninit};
+use core::ptr::NonNull;
+use core::str;
 use wasmtime_environ::component::{
     CanonicalAbiInfo, ComponentTypes, InterfaceType, StringEncoding, VariantInfo, MAX_FLAT_PARAMS,
     MAX_FLAT_RESULTS,
 };
-use wasmtime_runtime::component::ComponentInstance;
-use wasmtime_runtime::SendSyncPtr;
 
 /// A statically-typed version of [`Func`] which takes `Params` as input and
 /// returns `Return`.
@@ -170,7 +171,6 @@ where
     /// only works with functions defined within an asynchronous store. Also
     /// panics if `store` does not own this function.
     #[cfg(feature = "async")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "async")))]
     pub async fn call_async<T>(
         &self,
         mut store: impl AsContextMut<Data = T>,
@@ -316,8 +316,8 @@ where
     ) -> Result<Return> {
         assert!(Return::flatten_count() > MAX_FLAT_RESULTS);
         // FIXME: needs to read an i64 for memory64
-        let ptr = usize::try_from(dst.get_u32())?;
-        if ptr % usize::try_from(Return::ALIGN32)? != 0 {
+        let ptr = usize::try_from(dst.get_u32()).err2anyhow()?;
+        if ptr % usize::try_from(Return::ALIGN32).err2anyhow()? != 0 {
             bail!("return pointer not aligned");
         }
 
@@ -336,7 +336,6 @@ where
 
     /// See [`Func::post_return_async`]
     #[cfg(feature = "async")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "async")))]
     pub async fn post_return_async<T: Send>(
         &self,
         store: impl AsContextMut<Data = T>,
@@ -670,8 +669,8 @@ macro_rules! forward_type_impls {
 forward_type_impls! {
     (T: ComponentType + ?Sized) &'_ T => T,
     (T: ComponentType + ?Sized) Box<T> => T,
-    (T: ComponentType + ?Sized) std::rc::Rc<T> => T,
-    (T: ComponentType + ?Sized) std::sync::Arc<T> => T,
+    (T: ComponentType + ?Sized) alloc::rc::Rc<T> => T,
+    (T: ComponentType + ?Sized) alloc::sync::Arc<T> => T,
     () String => str,
     (T: ComponentType) Vec<T> => [T],
 }
@@ -703,8 +702,8 @@ macro_rules! forward_lowers {
 forward_lowers! {
     (T: Lower + ?Sized) &'_ T => T,
     (T: Lower + ?Sized) Box<T> => T,
-    (T: Lower + ?Sized) std::rc::Rc<T> => T,
-    (T: Lower + ?Sized) std::sync::Arc<T> => T,
+    (T: Lower + ?Sized) alloc::rc::Rc<T> => T,
+    (T: Lower + ?Sized) alloc::sync::Arc<T> => T,
     () String => str,
     (T: Lower) Vec<T> => [T],
 }
@@ -727,8 +726,8 @@ macro_rules! forward_string_lifts {
 
 forward_string_lifts! {
     Box<str>,
-    std::rc::Rc<str>,
-    std::sync::Arc<str>,
+    alloc::rc::Rc<str>,
+    alloc::sync::Arc<str>,
     String,
 }
 
@@ -750,8 +749,8 @@ macro_rules! forward_list_lifts {
 
 forward_list_lifts! {
     Box<[T]>,
-    std::rc::Rc<[T]>,
-    std::sync::Arc<[T]>,
+    alloc::rc::Rc<[T]>,
+    alloc::sync::Arc<[T]>,
     Vec<T>,
 }
 
@@ -1055,7 +1054,7 @@ unsafe impl Lift for char {
     #[inline]
     fn lift(_cx: &mut LiftContext<'_>, ty: InterfaceType, src: &Self::Lower) -> Result<Self> {
         debug_assert!(matches!(ty, InterfaceType::Char));
-        Ok(char::try_from(src.get_u32())?)
+        Ok(char::try_from(src.get_u32()).err2anyhow()?)
     }
 
     #[inline]
@@ -1063,7 +1062,7 @@ unsafe impl Lift for char {
         debug_assert!(matches!(ty, InterfaceType::Char));
         debug_assert!((bytes.as_ptr() as usize) % Self::SIZE32 == 0);
         let bits = u32::from_le_bytes(bytes.try_into().unwrap());
-        Ok(char::try_from(bits)?)
+        Ok(char::try_from(bits).err2anyhow()?)
     }
 }
 
@@ -1339,18 +1338,21 @@ impl WasmStr {
         // Note that bounds-checking already happen in construction of `WasmStr`
         // so this is never expected to panic. This could theoretically be
         // unchecked indexing if we're feeling wild enough.
-        Ok(str::from_utf8(&memory[self.ptr..][..self.len])?.into())
+        Ok(str::from_utf8(&memory[self.ptr..][..self.len])
+            .err2anyhow()?
+            .into())
     }
 
     fn decode_utf16<'a>(&self, memory: &'a [u8], len: usize) -> Result<Cow<'a, str>> {
         // See notes in `decode_utf8` for why this is panicking indexing.
         let memory = &memory[self.ptr..][..len * 2];
-        Ok(std::char::decode_utf16(
+        Ok(core::char::decode_utf16(
             memory
                 .chunks(2)
                 .map(|chunk| u16::from_le_bytes(chunk.try_into().unwrap())),
         )
-        .collect::<Result<String, _>>()?
+        .collect::<Result<String, _>>()
+        .err2anyhow()?
         .into())
     }
 
@@ -1384,7 +1386,10 @@ unsafe impl Lift for WasmStr {
         // FIXME: needs memory64 treatment
         let ptr = src[0].get_u32();
         let len = src[1].get_u32();
-        let (ptr, len) = (usize::try_from(ptr)?, usize::try_from(len)?);
+        let (ptr, len) = (
+            usize::try_from(ptr).err2anyhow()?,
+            usize::try_from(len).err2anyhow()?,
+        );
         WasmStr::new(ptr, len, cx)
     }
 
@@ -1395,7 +1400,10 @@ unsafe impl Lift for WasmStr {
         // FIXME: needs memory64 treatment
         let ptr = u32::from_le_bytes(bytes[..4].try_into().unwrap());
         let len = u32::from_le_bytes(bytes[4..].try_into().unwrap());
-        let (ptr, len) = (usize::try_from(ptr)?, usize::try_from(len)?);
+        let (ptr, len) = (
+            usize::try_from(ptr).err2anyhow()?,
+            usize::try_from(len).err2anyhow()?,
+        );
         WasmStr::new(ptr, len, cx)
     }
 }
@@ -1532,7 +1540,7 @@ impl<T: Lift> WasmList<T> {
             Some(n) if n <= cx.memory().len() => {}
             _ => bail!("list pointer/length out of bounds of memory"),
         }
-        if ptr % usize::try_from(T::ALIGN32)? != 0 {
+        if ptr % usize::try_from(T::ALIGN32).err2anyhow()? != 0 {
             bail!("list pointer is not aligned")
         }
         Ok(WasmList {
@@ -1688,7 +1696,10 @@ unsafe impl<T: Lift> Lift for WasmList<T> {
         // FIXME: needs memory64 treatment
         let ptr = src[0].get_u32();
         let len = src[1].get_u32();
-        let (ptr, len) = (usize::try_from(ptr)?, usize::try_from(len)?);
+        let (ptr, len) = (
+            usize::try_from(ptr).err2anyhow()?,
+            usize::try_from(len).err2anyhow()?,
+        );
         WasmList::new(ptr, len, cx, elem)
     }
 
@@ -1701,7 +1712,10 @@ unsafe impl<T: Lift> Lift for WasmList<T> {
         // FIXME: needs memory64 treatment
         let ptr = u32::from_le_bytes(bytes[..4].try_into().unwrap());
         let len = u32::from_le_bytes(bytes[4..].try_into().unwrap());
-        let (ptr, len) = (usize::try_from(ptr)?, usize::try_from(len)?);
+        let (ptr, len) = (
+            usize::try_from(ptr).err2anyhow()?,
+            usize::try_from(len).err2anyhow()?,
+        );
         WasmList::new(ptr, len, cx, elem)
     }
 }
@@ -1787,12 +1801,12 @@ pub fn typecheck_variant(
                 );
             }
 
-            for (case, &(name, check)) in cases.iter().zip(expected) {
-                if case.name != name {
-                    bail!("expected variant case named {name}, found {}", case.name);
+            for ((case_name, case_ty), &(name, check)) in cases.iter().zip(expected) {
+                if *case_name != name {
+                    bail!("expected variant case named {name}, found {case_name}");
                 }
 
-                match (check, &case.ty) {
+                match (check, case_ty) {
                     (Some(check), Some(ty)) => check(ty, types)
                         .with_context(|| format!("type mismatch for case {name}"))?,
                     (None, None) => {}
