@@ -1,17 +1,17 @@
 use super::{invoke_wasm_and_catch_traps, HostAbi};
+use crate::runtime::vm::{VMContext, VMFuncRef, VMNativeCallFunction, VMOpaqueContext};
 use crate::store::{AutoAssertNoGc, StoreOpaque};
 use crate::{
-    AsContext, AsContextMut, Engine, ExternRef, Func, FuncType, HeapType, NoFunc, RefType,
-    StoreContextMut, ValRaw, ValType,
+    AsContext, AsContextMut, Engine, Func, FuncType, HeapType, NoFunc, RefType, StoreContextMut,
+    ValRaw, ValType,
 };
 use anyhow::{bail, Context, Result};
-use std::marker;
-use std::mem::{self, MaybeUninit};
-use std::os::raw::c_void;
-use std::ptr::{self, NonNull};
-use wasmtime_runtime::{
-    VMContext, VMFuncRef, VMNativeCallFunction, VMOpaqueContext, VMSharedTypeIndex,
-};
+use core::ffi::c_void;
+use core::marker;
+use core::mem::{self, MaybeUninit};
+use core::num::NonZeroUsize;
+use core::ptr::{self, NonNull};
+use wasmtime_environ::VMSharedTypeIndex;
 
 /// A statically typed WebAssembly function.
 ///
@@ -100,6 +100,9 @@ where
             !store.0.async_support(),
             "must use `call_async` with async stores"
         );
+        if Self::need_gc_before_call_raw(store.0, &params) {
+            store.0.gc();
+        }
         let func = self.func.vm_func_ref(store.0);
         unsafe { Self::call_raw(&mut store, &self.ty, func, params) }
     }
@@ -122,7 +125,6 @@ where
     ///
     /// [`Trap`]: crate::Trap
     #[cfg(feature = "async")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "async")))]
     pub async fn call_async<T>(
         &self,
         mut store: impl AsContextMut<Data = T>,
@@ -136,6 +138,9 @@ where
             store.0.async_support(),
             "must use `call` with non-async stores"
         );
+        if Self::need_gc_before_call_raw(store.0, &params) {
+            store.0.gc_async().await;
+        }
         store
             .on_fiber(|store| {
                 let func = self.func.vm_func_ref(store.0);
@@ -144,6 +149,31 @@ where
             .await?
     }
 
+    #[inline]
+    pub(crate) fn need_gc_before_call_raw(_store: &StoreOpaque, _params: &Params) -> bool {
+        #[cfg(feature = "gc")]
+        {
+            // See the comment in `Func::call_impl_check_args`.
+            let num_gc_refs = _params.non_i31_gc_refs_count();
+            if let Some(num_gc_refs) = NonZeroUsize::new(num_gc_refs) {
+                return _store
+                    .unwrap_gc_store()
+                    .gc_heap
+                    .need_gc_before_entering_wasm(num_gc_refs);
+            }
+        }
+
+        false
+    }
+
+    /// Do a raw call of a typed function.
+    ///
+    /// # Safety
+    ///
+    /// `func` must be of the given type.
+    ///
+    /// If `Self::need_gc_before_call_raw`, then the caller must have done a GC
+    /// just before calling this method.
     pub(crate) unsafe fn call_raw<T>(
         store: &mut StoreContextMut<'_, T>,
         ty: &FuncType,
@@ -156,29 +186,12 @@ where
             Self::debug_typecheck(store.0, func.as_ref().type_index);
         }
 
-        #[cfg(feature = "gc")]
-        {
-            // See the comment in `Func::call_impl`'s `write_params` function.
-            if params.externrefs_count()
-                > store
-                    .0
-                    .externref_activations_table()
-                    .bump_capacity_remaining()
-            {
-                store.gc();
-            }
-        }
-
         // Validate that all runtime values flowing into this store indeed
         // belong within this store, otherwise it would be unsafe for store
         // values to cross each other.
 
         let params = {
-            // GC is not safe here, since we move refs into the activations
-            // table but don't hold a strong reference onto them until we enter
-            // the Wasm frame and they get referenced from the stack maps.
-            let mut store = AutoAssertNoGc::new(&mut **store.as_context_mut().0);
-
+            let mut store = AutoAssertNoGc::new(store.0);
             params.into_abi(&mut store, ty)?
         };
 
@@ -197,10 +210,13 @@ where
             ptr::write(ret.as_mut_ptr(), result);
             *returned = true
         });
+
         let (_, ret, _, returned) = captures;
         debug_assert_eq!(result.is_ok(), returned);
         result?;
-        Ok(Results::from_abi(store.0, ret.assume_init()))
+
+        let mut store = AutoAssertNoGc::new(store.0);
+        Ok(Results::from_abi(&mut store, ret.assume_init()))
     }
 
     /// Purely a debug-mode assertion, not actually used in release builds.
@@ -232,7 +248,7 @@ pub unsafe trait WasmTy: Send {
     // The raw ABI type that values of this type can be converted to and passed
     // to Wasm, or given from Wasm and converted back from.
     #[doc(hidden)]
-    type Abi: Copy;
+    type Abi: 'static + Copy;
 
     // Do a "static" (aka at time of `func.typed::<P, R>()`) ahead-of-time type
     // check for this type at the given position. You probably don't need to
@@ -282,8 +298,8 @@ pub unsafe trait WasmTy: Send {
                 (Some(expected_ref), Some(actual_ref)) if actual_ref.heap_type().is_concrete() => {
                     expected_ref
                         .heap_type()
-                        .top(engine)
-                        .ensure_matches(engine, &actual_ref.heap_type().top(engine))
+                        .top()
+                        .ensure_matches(engine, &actual_ref.heap_type().top())
                 }
                 _ => expected.ensure_matches(engine, &actual),
             },
@@ -310,12 +326,12 @@ pub unsafe trait WasmTy: Send {
         &self,
         store: &StoreOpaque,
         nullable: bool,
-        actual: &FuncType,
+        actual: &HeapType,
     ) -> Result<()>;
 
     // Is this an externref?
     #[doc(hidden)]
-    fn is_externref(&self) -> bool;
+    fn is_non_i31_gc_ref(&self) -> bool;
 
     // Construct a `Self::Abi` from the given `ValRaw`.
     #[doc(hidden)]
@@ -326,12 +342,39 @@ pub unsafe trait WasmTy: Send {
     unsafe fn abi_into_raw(abi: Self::Abi, raw: *mut ValRaw);
 
     // Convert `self` into `Self::Abi`.
+    //
+    // NB: We _must not_ trigger a GC when passing refs from host code into Wasm
+    // (e.g. returned from a host function or passed as arguments to a Wasm
+    // function). After insertion into the activations table, the reference is
+    // no longer rooted. If multiple references are being sent from the host
+    // into Wasm and we allowed GCs during insertion, then the following events
+    // could happen:
+    //
+    // * Reference A is inserted into the activations table. This does not
+    //   trigger a GC, but does fill the table to capacity.
+    //
+    // * The caller's reference to A is removed. Now the only reference to A is
+    //   from the activations table.
+    //
+    // * Reference B is inserted into the activations table. Because the table
+    //   is at capacity, a GC is triggered.
+    //
+    // * A is reclaimed because the only reference keeping it alive was the
+    //   activation table's reference (it isn't inside any Wasm frames on the
+    //   stack yet, so stack scanning and stack maps don't increment its
+    //   reference count).
+    //
+    // * We transfer control to Wasm, giving it A and B. Wasm uses A. That's a
+    //   use-after-free bug.
+    //
+    // In conclusion, to prevent uses-after-free bugs, we cannot GC while
+    // converting types into their raw ABI forms.
     #[doc(hidden)]
-    fn into_abi(self, store: &mut StoreOpaque) -> Self::Abi;
+    fn into_abi(self, store: &mut AutoAssertNoGc<'_>) -> Result<Self::Abi>;
 
     // Convert back from `Self::Abi` into `Self`.
     #[doc(hidden)]
-    unsafe fn from_abi(abi: Self::Abi, store: &mut StoreOpaque) -> Self;
+    unsafe fn from_abi(abi: Self::Abi, store: &mut AutoAssertNoGc<'_>) -> Self;
 }
 
 macro_rules! integers {
@@ -347,11 +390,11 @@ macro_rules! integers {
                 true
             }
             #[inline]
-            fn dynamic_concrete_type_check(&self, _: &StoreOpaque, _: bool, _: &FuncType) -> Result<()> {
+            fn dynamic_concrete_type_check(&self, _: &StoreOpaque, _: bool, _: &HeapType) -> Result<()> {
                 unreachable!()
             }
             #[inline]
-            fn is_externref(&self) -> bool {
+            fn is_non_i31_gc_ref(&self) -> bool {
                 false
             }
             #[inline]
@@ -363,11 +406,12 @@ macro_rules! integers {
                 *raw = ValRaw::$primitive(abi);
             }
             #[inline]
-            fn into_abi(self, _store: &mut StoreOpaque) -> Self::Abi {
-                self
+            fn into_abi(self, _store: &mut AutoAssertNoGc<'_>) -> Result<Self::Abi>
+            {
+                Ok(self)
             }
             #[inline]
-            unsafe fn from_abi(abi: Self::Abi, _store: &mut StoreOpaque) -> Self {
+            unsafe fn from_abi(abi: Self::Abi, _store: &mut AutoAssertNoGc<'_>) -> Self {
                 abi
             }
         }
@@ -394,11 +438,11 @@ macro_rules! floats {
                 true
             }
             #[inline]
-            fn dynamic_concrete_type_check(&self, _: &StoreOpaque, _: bool, _: &FuncType) -> Result<()> {
+            fn dynamic_concrete_type_check(&self, _: &StoreOpaque, _: bool, _: &HeapType) -> Result<()> {
                 unreachable!()
             }
             #[inline]
-            fn is_externref(&self) -> bool {
+            fn is_non_i31_gc_ref(&self) -> bool {
                 false
             }
             #[inline]
@@ -410,11 +454,12 @@ macro_rules! floats {
                 *raw = ValRaw::$float(abi.to_bits());
             }
             #[inline]
-            fn into_abi(self, _store: &mut StoreOpaque) -> Self::Abi {
-                self
+            fn into_abi(self, _store: &mut AutoAssertNoGc<'_>) -> Result<Self::Abi>
+            {
+                Ok(self)
             }
             #[inline]
-            unsafe fn from_abi(abi: Self::Abi, _store: &mut StoreOpaque) -> Self {
+            unsafe fn from_abi(abi: Self::Abi, _store: &mut AutoAssertNoGc<'_>) -> Self {
                 abi
             }
         }
@@ -426,140 +471,6 @@ floats! {
     f64/u64/get_f64 => F64
 }
 
-#[cfg(feature = "gc")]
-unsafe impl WasmTy for ExternRef {
-    type Abi = NonNull<u8>;
-
-    #[inline]
-    fn valtype() -> ValType {
-        ValType::Ref(RefType::new(false, HeapType::Extern))
-    }
-
-    #[inline]
-    fn compatible_with_store(&self, _store: &StoreOpaque) -> bool {
-        true
-    }
-
-    #[inline]
-    fn dynamic_concrete_type_check(&self, _: &StoreOpaque, _: bool, _: &FuncType) -> Result<()> {
-        unreachable!()
-    }
-
-    #[inline]
-    fn is_externref(&self) -> bool {
-        true
-    }
-
-    #[inline]
-    unsafe fn abi_from_raw(raw: *mut ValRaw) -> Self::Abi {
-        let p = (*raw).get_externref().cast::<u8>();
-        debug_assert!(!p.is_null());
-        NonNull::new_unchecked(p)
-    }
-
-    #[inline]
-    unsafe fn abi_into_raw(abi: NonNull<u8>, raw: *mut ValRaw) {
-        *raw = ValRaw::externref(abi.cast::<c_void>().as_ptr());
-    }
-
-    #[inline]
-    fn into_abi(self, store: &mut StoreOpaque) -> Self::Abi {
-        let inner = self.into_vm_extern_ref();
-        let abi = inner.as_raw();
-        unsafe {
-            // NB: We _must not_ trigger a GC when passing refs from host
-            // code into Wasm (e.g. returned from a host function or passed
-            // as arguments to a Wasm function). After insertion into the
-            // table, this reference is no longer rooted. If multiple
-            // references are being sent from the host into Wasm and we
-            // allowed GCs during insertion, then the following events could
-            // happen:
-            //
-            // * Reference A is inserted into the activations
-            //   table. This does not trigger a GC, but does fill the table
-            //   to capacity.
-            //
-            // * The caller's reference to A is removed. Now the only
-            //   reference to A is from the activations table.
-            //
-            // * Reference B is inserted into the activations table. Because
-            //   the table is at capacity, a GC is triggered.
-            //
-            // * A is reclaimed because the only reference keeping it alive
-            //   was the activation table's reference (it isn't inside any
-            //   Wasm frames on the stack yet, so stack scanning and stack
-            //   maps don't increment its reference count).
-            //
-            // * We transfer control to Wasm, giving it A and B. Wasm uses
-            //   A. That's a use after free.
-            //
-            // In conclusion, to prevent uses after free, we cannot GC
-            // during this insertion.
-            let mut store = AutoAssertNoGc::new(store);
-            store.insert_vmexternref_without_gc(inner);
-
-            debug_assert!(!abi.is_null());
-            NonNull::new_unchecked(abi)
-        }
-    }
-
-    #[inline]
-    unsafe fn from_abi(abi: Self::Abi, _store: &mut StoreOpaque) -> Self {
-        let inner = wasmtime_runtime::VMExternRef::clone_from_raw(abi.as_ptr()).unwrap();
-        ExternRef::from_vm_extern_ref(inner)
-    }
-}
-
-#[cfg(feature = "gc")]
-unsafe impl WasmTy for Option<ExternRef> {
-    type Abi = *mut u8;
-
-    #[inline]
-    fn valtype() -> ValType {
-        ValType::EXTERNREF
-    }
-
-    #[inline]
-    fn compatible_with_store(&self, _store: &StoreOpaque) -> bool {
-        true
-    }
-
-    #[inline]
-    fn dynamic_concrete_type_check(&self, _: &StoreOpaque, _: bool, _: &FuncType) -> Result<()> {
-        unreachable!()
-    }
-
-    #[inline]
-    fn is_externref(&self) -> bool {
-        true
-    }
-
-    #[inline]
-    unsafe fn abi_from_raw(raw: *mut ValRaw) -> *mut u8 {
-        (*raw).get_externref() as *mut u8
-    }
-
-    #[inline]
-    unsafe fn abi_into_raw(abi: *mut u8, raw: *mut ValRaw) {
-        *raw = ValRaw::externref(abi.cast());
-    }
-
-    #[inline]
-    fn into_abi(self, store: &mut StoreOpaque) -> Self::Abi {
-        if let Some(x) = self {
-            <ExternRef as WasmTy>::into_abi(x, store).as_ptr()
-        } else {
-            ptr::null_mut()
-        }
-    }
-
-    #[inline]
-    unsafe fn from_abi(abi: Self::Abi, _store: &mut StoreOpaque) -> Self {
-        let inner = wasmtime_runtime::VMExternRef::clone_from_raw(abi)?;
-        Some(ExternRef::from_vm_extern_ref(inner))
-    }
-}
-
 unsafe impl WasmTy for NoFunc {
     type Abi = NoFunc;
 
@@ -574,12 +485,12 @@ unsafe impl WasmTy for NoFunc {
     }
 
     #[inline]
-    fn dynamic_concrete_type_check(&self, _: &StoreOpaque, _: bool, _: &FuncType) -> Result<()> {
+    fn dynamic_concrete_type_check(&self, _: &StoreOpaque, _: bool, _: &HeapType) -> Result<()> {
         match self._inner {}
     }
 
     #[inline]
-    fn is_externref(&self) -> bool {
+    fn is_non_i31_gc_ref(&self) -> bool {
         match self._inner {}
     }
 
@@ -594,12 +505,12 @@ unsafe impl WasmTy for NoFunc {
     }
 
     #[inline]
-    fn into_abi(self, _store: &mut StoreOpaque) -> Self::Abi {
+    fn into_abi(self, _store: &mut AutoAssertNoGc<'_>) -> Result<Self::Abi> {
         unreachable!("NoFunc is uninhabited")
     }
 
     #[inline]
-    unsafe fn from_abi(_abi: Self::Abi, _store: &mut StoreOpaque) -> Self {
+    unsafe fn from_abi(_abi: Self::Abi, _store: &mut AutoAssertNoGc<'_>) -> Self {
         unreachable!("NoFunc is uninhabited")
     }
 }
@@ -622,18 +533,18 @@ unsafe impl WasmTy for Option<NoFunc> {
         &self,
         _: &StoreOpaque,
         nullable: bool,
-        func_ty: &FuncType,
+        ty: &HeapType,
     ) -> Result<()> {
         if nullable {
             // `(ref null nofunc) <: (ref null $f)` for all function types `$f`.
             Ok(())
         } else {
-            bail!("argument type mismatch: expected (ref {func_ty}), found null reference")
+            bail!("argument type mismatch: expected non-nullable (ref {ty}), found null reference")
         }
     }
 
     #[inline]
-    fn is_externref(&self) -> bool {
+    fn is_non_i31_gc_ref(&self) -> bool {
         false
     }
 
@@ -648,18 +559,18 @@ unsafe impl WasmTy for Option<NoFunc> {
     }
 
     #[inline]
-    fn into_abi(self, _store: &mut StoreOpaque) -> Self::Abi {
-        ptr::null_mut()
+    fn into_abi(self, _store: &mut AutoAssertNoGc<'_>) -> Result<Self::Abi> {
+        Ok(ptr::null_mut())
     }
 
     #[inline]
-    unsafe fn from_abi(_abi: Self::Abi, _store: &mut StoreOpaque) -> Self {
+    unsafe fn from_abi(_abi: Self::Abi, _store: &mut AutoAssertNoGc<'_>) -> Self {
         None
     }
 }
 
 unsafe impl WasmTy for Func {
-    type Abi = NonNull<wasmtime_runtime::VMFuncRef>;
+    type Abi = NonNull<crate::runtime::vm::VMFuncRef>;
 
     #[inline]
     fn valtype() -> ValType {
@@ -676,14 +587,15 @@ unsafe impl WasmTy for Func {
         &self,
         store: &StoreOpaque,
         _nullable: bool,
-        actual: &FuncType,
+        expected: &HeapType,
     ) -> Result<()> {
-        self.ensure_matches_ty(store, actual)
+        let expected = expected.unwrap_concrete_func();
+        self.ensure_matches_ty(store, expected)
             .context("argument type mismatch for reference to concrete type")
     }
 
     #[inline]
-    fn is_externref(&self) -> bool {
+    fn is_non_i31_gc_ref(&self) -> bool {
         false
     }
 
@@ -691,7 +603,7 @@ unsafe impl WasmTy for Func {
     unsafe fn abi_from_raw(raw: *mut ValRaw) -> Self::Abi {
         let p = (*raw).get_funcref();
         debug_assert!(!p.is_null());
-        NonNull::new_unchecked(p.cast::<wasmtime_runtime::VMFuncRef>())
+        NonNull::new_unchecked(p.cast::<crate::runtime::vm::VMFuncRef>())
     }
 
     #[inline]
@@ -700,18 +612,18 @@ unsafe impl WasmTy for Func {
     }
 
     #[inline]
-    fn into_abi(self, store: &mut StoreOpaque) -> Self::Abi {
-        self.vm_func_ref(store)
+    fn into_abi(self, store: &mut AutoAssertNoGc<'_>) -> Result<Self::Abi> {
+        Ok(self.vm_func_ref(store))
     }
 
     #[inline]
-    unsafe fn from_abi(abi: Self::Abi, store: &mut StoreOpaque) -> Self {
+    unsafe fn from_abi(abi: Self::Abi, store: &mut AutoAssertNoGc<'_>) -> Self {
         Func::from_vm_func_ref(store, abi.as_ptr()).unwrap()
     }
 }
 
 unsafe impl WasmTy for Option<Func> {
-    type Abi = *mut wasmtime_runtime::VMFuncRef;
+    type Abi = *mut crate::runtime::vm::VMFuncRef;
 
     #[inline]
     fn valtype() -> ValType {
@@ -731,20 +643,21 @@ unsafe impl WasmTy for Option<Func> {
         &self,
         store: &StoreOpaque,
         nullable: bool,
-        func_ty: &FuncType,
+        expected: &HeapType,
     ) -> Result<()> {
         if let Some(f) = self {
-            f.ensure_matches_ty(store, func_ty)
+            let expected = expected.unwrap_concrete_func();
+            f.ensure_matches_ty(store, expected)
                 .context("argument type mismatch for reference to concrete type")
         } else if nullable {
             Ok(())
         } else {
-            bail!("argument type mismatch: expected (ref {func_ty}), found null reference")
+            bail!("argument type mismatch: expected non-nullable (ref {expected}), found null reference")
         }
     }
 
     #[inline]
-    fn is_externref(&self) -> bool {
+    fn is_non_i31_gc_ref(&self) -> bool {
         false
     }
 
@@ -759,16 +672,16 @@ unsafe impl WasmTy for Option<Func> {
     }
 
     #[inline]
-    fn into_abi(self, store: &mut StoreOpaque) -> Self::Abi {
-        if let Some(f) = self {
+    fn into_abi(self, store: &mut AutoAssertNoGc<'_>) -> Result<Self::Abi> {
+        Ok(if let Some(f) = self {
             f.vm_func_ref(store).as_ptr()
         } else {
             ptr::null_mut()
-        }
+        })
     }
 
     #[inline]
-    unsafe fn from_abi(abi: Self::Abi, store: &mut StoreOpaque) -> Self {
+    unsafe fn from_abi(abi: Self::Abi, store: &mut AutoAssertNoGc<'_>) -> Self {
         Func::from_vm_func_ref(store, abi)
     }
 }
@@ -790,10 +703,10 @@ pub unsafe trait WasmParams: Send {
     ) -> Result<()>;
 
     #[doc(hidden)]
-    fn externrefs_count(&self) -> usize;
+    fn non_i31_gc_refs_count(&self) -> usize;
 
     #[doc(hidden)]
-    fn into_abi(self, store: &mut StoreOpaque, func_ty: &FuncType) -> Result<Self::Abi>;
+    fn into_abi(self, store: &mut AutoAssertNoGc<'_>, func_ty: &FuncType) -> Result<Self::Abi>;
 
     #[doc(hidden)]
     unsafe fn invoke<R: WasmResults>(
@@ -821,12 +734,12 @@ where
     }
 
     #[inline]
-    fn externrefs_count(&self) -> usize {
-        T::is_externref(self) as usize
+    fn non_i31_gc_refs_count(&self) -> usize {
+        T::is_non_i31_gc_ref(self) as usize
     }
 
     #[inline]
-    fn into_abi(self, store: &mut StoreOpaque, func_ty: &FuncType) -> Result<Self::Abi> {
+    fn into_abi(self, store: &mut AutoAssertNoGc<'_>, func_ty: &FuncType) -> Result<Self::Abi> {
         <(T,) as WasmParams>::into_abi((self,), store, func_ty)
     }
 
@@ -873,10 +786,10 @@ macro_rules! impl_wasm_params {
             }
 
             #[inline]
-            fn externrefs_count(&self) -> usize {
+            fn non_i31_gc_refs_count(&self) -> usize {
                 let ($(ref $t,)*) = self;
                 0 $(
-                    + $t.is_externref() as usize
+                    + $t.is_non_i31_gc_ref() as usize
                 )*
             }
 
@@ -884,7 +797,7 @@ macro_rules! impl_wasm_params {
             #[inline]
             fn into_abi(
                 self,
-                _store: &mut StoreOpaque,
+                _store: &mut AutoAssertNoGc<'_>,
                 _func_ty: &FuncType,
             ) -> Result<Self::Abi> {
                 let ($($t,)*) = self;
@@ -896,14 +809,15 @@ macro_rules! impl_wasm_params {
                     }
 
                     if $t::valtype().is_ref() {
-                        let p = _func_ty.param(_i).unwrap();
-                        let r = p.unwrap_ref();
-                        if let Some(c) = r.heap_type().as_concrete() {
-                            $t.dynamic_concrete_type_check(_store, r.is_nullable(), c)?;
+                        let param_ty = _func_ty.param(_i).unwrap();
+                        let ref_ty = param_ty.unwrap_ref();
+                        let heap_ty = ref_ty.heap_type();
+                        if heap_ty.is_concrete() {
+                            $t.dynamic_concrete_type_check(_store, ref_ty.is_nullable(), heap_ty)?;
                         }
                     }
 
-                    let $t = $t.into_abi(_store);
+                    let $t = $t.into_abi(_store)?;
 
                     _i += 1;
                 )*
@@ -950,7 +864,7 @@ pub unsafe trait WasmResults: WasmParams {
     type ResultAbi: HostAbi;
 
     #[doc(hidden)]
-    unsafe fn from_abi(store: &mut StoreOpaque, abi: Self::ResultAbi) -> Self;
+    unsafe fn from_abi(store: &mut AutoAssertNoGc<'_>, abi: Self::ResultAbi) -> Self;
 }
 
 // Forwards from a bare type `T` to the 1-tuple type `(T,)`
@@ -960,7 +874,7 @@ where
 {
     type ResultAbi = <(T,) as WasmResults>::ResultAbi;
 
-    unsafe fn from_abi(store: &mut StoreOpaque, abi: Self::ResultAbi) -> Self {
+    unsafe fn from_abi(store: &mut AutoAssertNoGc<'_>, abi: Self::ResultAbi) -> Self {
         <(T,) as WasmResults>::from_abi(store, abi).0
     }
 }
@@ -974,7 +888,7 @@ macro_rules! impl_wasm_results {
             type ResultAbi = ($($t::Abi,)*);
 
             #[inline]
-            unsafe fn from_abi(store: &mut StoreOpaque, abi: Self::ResultAbi) -> Self {
+            unsafe fn from_abi(store: &mut AutoAssertNoGc<'_>, abi: Self::ResultAbi) -> Self {
                 let ($($t,)*) = abi;
                 ($($t::from_abi($t, store),)*)
             }
