@@ -1,11 +1,13 @@
-use crate::r#ref::ExternRef;
-use crate::store::StoreOpaque;
-use crate::{AsContext, AsContextMut, Func, HeapType, RefType, ValType, V128};
+use crate::runtime::vm::{TableElement, VMGcRef};
+use crate::store::{AutoAssertNoGc, StoreOpaque};
+use crate::{
+    AnyRef, AsContext, AsContextMut, ExternRef, Func, HeapType, RefType, Rooted, RootedGcRefImpl,
+    ValType, V128,
+};
 use anyhow::{bail, Context, Result};
-use std::ptr;
-use wasmtime_runtime::TableElement;
+use core::ptr;
 
-pub use wasmtime_runtime::ValRaw;
+pub use crate::runtime::vm::ValRaw;
 
 /// Possible runtime values that a WebAssembly module can either consume or
 /// produce.
@@ -42,7 +44,10 @@ pub enum Val {
     FuncRef(Option<Func>),
 
     /// An external reference.
-    ExternRef(Option<ExternRef>),
+    ExternRef(Option<Rooted<ExternRef>>),
+
+    /// An internal reference.
+    AnyRef(Option<Rooted<AnyRef>>),
 }
 
 macro_rules! accessors {
@@ -72,6 +77,12 @@ macro_rules! accessors {
 }
 
 impl Val {
+    /// Returns the null reference for the given heap type.
+    #[inline]
+    pub fn null_ref(heap_type: &HeapType) -> Val {
+        Ref::null(&heap_type).into()
+    }
+
     /// Returns the null function reference value.
     ///
     /// The return value has type `(ref null nofunc)` aka `nullfuncref` and is a
@@ -90,6 +101,15 @@ impl Val {
         Val::ExternRef(None)
     }
 
+    /// Returns the null function reference value.
+    ///
+    /// The return value has type `(ref null any)` aka `nullref` and is a
+    /// subtype of all internal references.
+    #[inline]
+    pub const fn null_any_ref() -> Val {
+        Val::AnyRef(None)
+    }
+
     /// Returns the corresponding [`ValType`] for this `Val`.
     #[inline]
     pub fn ty(&self, store: impl AsContext) -> ValType {
@@ -106,25 +126,33 @@ impl Val {
             Val::V128(_) => ValType::V128,
             Val::ExternRef(_) => ValType::EXTERNREF,
             Val::FuncRef(None) => ValType::NULLFUNCREF,
-            Val::FuncRef(Some(f)) => {
-                ValType::Ref(RefType::new(false, HeapType::Concrete(f.load_ty(store))))
+            Val::FuncRef(Some(f)) => ValType::Ref(RefType::new(
+                false,
+                HeapType::ConcreteFunc(f.load_ty(store)),
+            )),
+            Val::AnyRef(None) => ValType::NULLREF,
+            Val::AnyRef(Some(_)) => {
+                assert!(VMGcRef::ONLY_EXTERN_REF_AND_I31);
+                ValType::Ref(RefType::new(false, HeapType::I31))
             }
         }
     }
 
     /// Does this value match the given type?
     ///
+    /// Returns an error is an underlying `Rooted` has been unrooted.
+    ///
     /// # Panics
     ///
     /// Panics if this value is not associated with the given store.
-    pub fn matches_ty(&self, store: impl AsContext, ty: &ValType) -> bool {
+    pub fn matches_ty(&self, store: impl AsContext, ty: &ValType) -> Result<bool> {
         self._matches_ty(&store.as_context().0, ty)
     }
 
-    pub(crate) fn _matches_ty(&self, store: &StoreOpaque, ty: &ValType) -> bool {
+    pub(crate) fn _matches_ty(&self, store: &StoreOpaque, ty: &ValType) -> Result<bool> {
         assert!(self.comes_from_same_store(store));
         assert!(ty.comes_from_same_engine(store.engine()));
-        match (self, ty) {
+        Ok(match (self, ty) {
             (Val::I32(_), ValType::I32)
             | (Val::I64(_), ValType::I64)
             | (Val::F32(_), ValType::F32)
@@ -132,11 +160,12 @@ impl Val {
             | (Val::V128(_), ValType::V128) => true,
 
             (Val::FuncRef(f), ValType::Ref(ref_ty)) => {
-                Ref::from(f.clone())._matches_ty(store, ref_ty)
+                Ref::from(f.clone())._matches_ty(store, ref_ty)?
             }
             (Val::ExternRef(e), ValType::Ref(ref_ty)) => {
-                Ref::from(e.clone())._matches_ty(store, ref_ty)
+                Ref::from(*e)._matches_ty(store, ref_ty)?
             }
+            (Val::AnyRef(a), ValType::Ref(ref_ty)) => Ref::from(*a)._matches_ty(store, ref_ty)?,
 
             (Val::I32(_), _)
             | (Val::I64(_), _)
@@ -144,8 +173,9 @@ impl Val {
             | (Val::F64(_), _)
             | (Val::V128(_), _)
             | (Val::FuncRef(_), _)
-            | (Val::ExternRef(_), _) => false,
-        }
+            | (Val::ExternRef(_), _)
+            | (Val::AnyRef(_), _) => false,
+        })
     }
 
     pub(crate) fn ensure_matches_ty(&self, store: &StoreOpaque, ty: &ValType) -> Result<()> {
@@ -155,7 +185,7 @@ impl Val {
         if !ty.comes_from_same_engine(store.engine()) {
             bail!("type used with wrong engine")
         }
-        if self._matches_ty(store, ty) {
+        if self._matches_ty(store, ty)? {
             Ok(())
         } else {
             let actual_ty = self.load_ty(store);
@@ -165,31 +195,32 @@ impl Val {
 
     /// Convenience method to convert this [`Val`] into a [`ValRaw`].
     ///
+    /// Returns an error if this value is a GC reference and the GC reference
+    /// has been unrooted.
+    ///
     /// # Unsafety
     ///
     /// This method is unsafe for the reasons that [`ExternRef::to_raw`] and
     /// [`Func::to_raw`] are unsafe.
-    pub unsafe fn to_raw(&self, store: impl AsContextMut) -> ValRaw {
+    pub unsafe fn to_raw(&self, store: impl AsContextMut) -> Result<ValRaw> {
         match self {
-            Val::I32(i) => ValRaw::i32(*i),
-            Val::I64(i) => ValRaw::i64(*i),
-            Val::F32(u) => ValRaw::f32(*u),
-            Val::F64(u) => ValRaw::f64(*u),
-            Val::V128(b) => ValRaw::v128(b.as_u128()),
-            Val::ExternRef(e) => {
-                let externref = match e {
-                    None => ptr::null_mut(),
-                    Some(e) => e.to_raw(store),
-                };
-                ValRaw::externref(externref)
-            }
-            Val::FuncRef(f) => {
-                let funcref = match f {
-                    Some(f) => f.to_raw(store),
-                    None => ptr::null_mut(),
-                };
-                ValRaw::funcref(funcref)
-            }
+            Val::I32(i) => Ok(ValRaw::i32(*i)),
+            Val::I64(i) => Ok(ValRaw::i64(*i)),
+            Val::F32(u) => Ok(ValRaw::f32(*u)),
+            Val::F64(u) => Ok(ValRaw::f64(*u)),
+            Val::V128(b) => Ok(ValRaw::v128(b.as_u128())),
+            Val::ExternRef(e) => Ok(ValRaw::externref(match e {
+                None => 0,
+                Some(e) => e.to_raw(store)?,
+            })),
+            Val::AnyRef(e) => Ok(ValRaw::anyref(match e {
+                None => 0,
+                Some(e) => e.to_raw(store)?,
+            })),
+            Val::FuncRef(f) => Ok(ValRaw::funcref(match f {
+                Some(f) => f.to_raw(store),
+                None => ptr::null_mut(),
+            })),
         }
     }
 
@@ -209,15 +240,32 @@ impl Val {
             ValType::V128 => Val::V128(raw.get_v128().into()),
             ValType::Ref(ref_ty) => {
                 let ref_ = match ref_ty.heap_type() {
-                    HeapType::Func | HeapType::Concrete(_) => {
+                    HeapType::Func | HeapType::ConcreteFunc(_) => {
                         Func::from_raw(store, raw.get_funcref()).into()
                     }
+
                     HeapType::NoFunc => Ref::Func(None),
-                    HeapType::Extern => ExternRef::from_raw(raw.get_externref()).into(),
+
+                    HeapType::Extern => ExternRef::from_raw(store, raw.get_externref()).into(),
+
+                    HeapType::NoExtern => Ref::Extern(None),
+
+                    HeapType::Any
+                    | HeapType::Eq
+                    | HeapType::I31
+                    | HeapType::Array
+                    | HeapType::ConcreteArray(_)
+                    | HeapType::Struct
+                    | HeapType::ConcreteStruct(_) => {
+                        AnyRef::from_raw(store, raw.get_anyref()).into()
+                    }
+
+                    HeapType::None => Ref::Any(None),
                 };
                 assert!(
                     ref_ty.is_nullable() || !ref_.is_null(),
-                    "if the type is not nullable, we shouldn't get null"
+                    "if the type is not nullable, we shouldn't get null; got \
+                     type = {ref_ty}, ref = {ref_:?}"
                 );
                 ref_.into()
             }
@@ -231,7 +279,8 @@ impl Val {
         (F32(f32) f32 unwrap_f32 f32::from_bits(*e))
         (F64(f64) f64 unwrap_f64 f64::from_bits(*e))
         (FuncRef(Option<&Func>) func_ref unwrap_func_ref e.as_ref())
-        (ExternRef(Option<&ExternRef>) extern_ref unwrap_extern_ref e.as_ref())
+        (ExternRef(Option<&Rooted<ExternRef>>) extern_ref unwrap_extern_ref e.as_ref())
+        (AnyRef(Option<&Rooted<AnyRef>>) any_ref unwrap_any_ref e.as_ref())
         (V128(V128) v128 unwrap_v128 *e)
     }
 
@@ -241,6 +290,7 @@ impl Val {
         match self {
             Val::FuncRef(f) => Some(Ref::Func(f)),
             Val::ExternRef(e) => Some(Ref::Extern(e)),
+            Val::AnyRef(a) => Some(Ref::Any(a)),
             Val::I32(_) | Val::I64(_) | Val::F32(_) | Val::F64(_) | Val::V128(_) => None,
         }
     }
@@ -253,7 +303,7 @@ impl Val {
     ///
     /// If this is a non-null `externref`, then `Some(Some(..))` is returned.
     #[inline]
-    pub fn externref(&self) -> Option<Option<&ExternRef>> {
+    pub fn externref(&self) -> Option<Option<&Rooted<ExternRef>>> {
         match self {
             Val::ExternRef(None) => Some(None),
             Val::ExternRef(Some(e)) => Some(Some(e)),
@@ -272,10 +322,41 @@ impl Val {
     ///
     /// Panics if `self` is not a (nullable) `externref`.
     #[inline]
-    pub fn unwrap_externref(&self) -> Option<&ExternRef> {
+    pub fn unwrap_externref(&self) -> Option<&Rooted<ExternRef>> {
         self.externref().expect("expected externref")
     }
 
+    /// Attempt to access the underlying `anyref` value of this `Val`.
+    ///
+    /// If this is not an `anyref`, then `None` is returned.
+    ///
+    /// If this is a null `anyref`, then `Some(None)` is returned.
+    ///
+    /// If this is a non-null `anyref`, then `Some(Some(..))` is returned.
+    #[inline]
+    pub fn anyref(&self) -> Option<Option<&Rooted<AnyRef>>> {
+        match self {
+            Val::AnyRef(None) => Some(None),
+            Val::AnyRef(Some(e)) => Some(Some(e)),
+            _ => None,
+        }
+    }
+
+    /// Returns the underlying `anyref` value of this `Val`, panicking if it's the
+    /// wrong type.
+    ///
+    /// If this is a null `anyref`, then `None` is returned.
+    ///
+    /// If this is a non-null `anyref`, then `Some(..)` is returned.
+    ///
+    /// # Panics
+    ///
+    /// Panics if `self` is not a (nullable) `anyref`.
+    #[inline]
+    pub fn unwrap_anyref(&self) -> Option<&Rooted<AnyRef>> {
+        self.anyref().expect("expected anyref")
+    }
+
     /// Attempt to access the underlying `funcref` value of this `Val`.
     ///
     /// If this is not an `funcref`, then `None` is returned.
@@ -313,15 +394,16 @@ impl Val {
             Val::FuncRef(Some(f)) => f.comes_from_same_store(store),
             Val::FuncRef(None) => true,
 
-            // Integers, floats, vectors, and `externref`s have no association
-            // with any particular store, so they're always considered as "yes I
-            // came from that store",
-            Val::I32(_)
-            | Val::I64(_)
-            | Val::F32(_)
-            | Val::F64(_)
-            | Val::V128(_)
-            | Val::ExternRef(_) => true,
+            Val::ExternRef(Some(x)) => x.comes_from_same_store(store),
+            Val::ExternRef(None) => true,
+
+            Val::AnyRef(Some(a)) => a.comes_from_same_store(store),
+            Val::AnyRef(None) => true,
+
+            // Integers, floats, and vectors have no association with any
+            // particular store, so they're always considered as "yes I came
+            // from that store",
+            Val::I32(_) | Val::I64(_) | Val::F32(_) | Val::F64(_) | Val::V128(_) => true,
         }
     }
 }
@@ -360,24 +442,39 @@ impl From<Ref> for Val {
         match val {
             Ref::Extern(e) => Val::ExternRef(e),
             Ref::Func(f) => Val::FuncRef(f),
+            Ref::Any(a) => Val::AnyRef(a),
         }
     }
 }
 
-impl From<ExternRef> for Val {
+impl From<Rooted<ExternRef>> for Val {
     #[inline]
-    fn from(val: ExternRef) -> Val {
+    fn from(val: Rooted<ExternRef>) -> Val {
         Val::ExternRef(Some(val))
     }
 }
 
-impl From<Option<ExternRef>> for Val {
+impl From<Option<Rooted<ExternRef>>> for Val {
     #[inline]
-    fn from(val: Option<ExternRef>) -> Val {
+    fn from(val: Option<Rooted<ExternRef>>) -> Val {
         Val::ExternRef(val)
     }
 }
 
+impl From<Rooted<AnyRef>> for Val {
+    #[inline]
+    fn from(val: Rooted<AnyRef>) -> Val {
+        Val::AnyRef(Some(val))
+    }
+}
+
+impl From<Option<Rooted<AnyRef>>> for Val {
+    #[inline]
+    fn from(val: Option<Rooted<AnyRef>>) -> Val {
+        Val::AnyRef(val)
+    }
+}
+
 impl From<Func> for Val {
     #[inline]
     fn from(val: Func) -> Val {
@@ -479,7 +576,17 @@ pub enum Ref {
     ///
     /// Wasm can create null external references via the `ref.null extern`
     /// instruction.
-    Extern(Option<ExternRef>),
+    Extern(Option<Rooted<ExternRef>>),
+
+    /// An internal reference.
+    ///
+    /// The `AnyRef` type represents WebAssembly `anyref` values. These can be
+    /// references to `struct`s and `array`s or inline/unboxed 31-bit
+    /// integers.
+    ///
+    /// Unlike `externref`, Wasm guests can directly allocate `anyref`s, and
+    /// does not need to rely on the host to do that.
+    Any(Option<Rooted<AnyRef>>),
 }
 
 impl From<Func> for Ref {
@@ -496,27 +603,52 @@ impl From<Option<Func>> for Ref {
     }
 }
 
-impl From<ExternRef> for Ref {
+impl From<Rooted<ExternRef>> for Ref {
     #[inline]
-    fn from(e: ExternRef) -> Ref {
+    fn from(e: Rooted<ExternRef>) -> Ref {
         Ref::Extern(Some(e))
     }
 }
 
-impl From<Option<ExternRef>> for Ref {
+impl From<Option<Rooted<ExternRef>>> for Ref {
     #[inline]
-    fn from(e: Option<ExternRef>) -> Ref {
+    fn from(e: Option<Rooted<ExternRef>>) -> Ref {
         Ref::Extern(e)
     }
 }
 
+impl From<Rooted<AnyRef>> for Ref {
+    #[inline]
+    fn from(e: Rooted<AnyRef>) -> Ref {
+        Ref::Any(Some(e))
+    }
+}
+
+impl From<Option<Rooted<AnyRef>>> for Ref {
+    #[inline]
+    fn from(e: Option<Rooted<AnyRef>>) -> Ref {
+        Ref::Any(e)
+    }
+}
+
 impl Ref {
+    /// Create a null reference to the given heap type.
+    #[inline]
+    pub fn null(heap_type: &HeapType) -> Self {
+        match heap_type.top() {
+            HeapType::Any => Ref::Any(None),
+            HeapType::Extern => Ref::Extern(None),
+            HeapType::Func => Ref::Func(None),
+            ty => unreachable!("not a heap type: {ty:?}"),
+        }
+    }
+
     /// Is this a null reference?
     #[inline]
     pub fn is_null(&self) -> bool {
         match self {
-            Ref::Extern(None) | Ref::Func(None) => true,
-            Ref::Extern(Some(_)) | Ref::Func(Some(_)) => false,
+            Ref::Any(None) | Ref::Extern(None) | Ref::Func(None) => true,
+            Ref::Any(Some(_)) | Ref::Extern(Some(_)) | Ref::Func(Some(_)) => false,
         }
     }
 
@@ -541,7 +673,7 @@ impl Ref {
     ///
     /// Returns `Some(Some(_))` if this `Ref` is a non-null `extern` reference.
     #[inline]
-    pub fn as_extern(&self) -> Option<Option<&ExternRef>> {
+    pub fn as_extern(&self) -> Option<Option<&Rooted<ExternRef>>> {
         match self {
             Ref::Extern(e) => Some(e.as_ref()),
             _ => None,
@@ -555,11 +687,44 @@ impl Ref {
     ///
     /// Returns `Some(_)` if this `Ref` is a non-null `extern` reference.
     #[inline]
-    pub fn unwrap_extern(&self) -> Option<&ExternRef> {
+    pub fn unwrap_extern(&self) -> Option<&Rooted<ExternRef>> {
         self.as_extern()
             .expect("Ref::unwrap_extern on non-extern reference")
     }
 
+    /// Is this an `any` reference?
+    #[inline]
+    pub fn is_any(&self) -> bool {
+        matches!(self, Ref::Any(_))
+    }
+
+    /// Get the underlying `any` reference, if any.
+    ///
+    /// Returns `None` if this `Ref` is not an `any` reference, eg it is a
+    /// `func` reference.
+    ///
+    /// Returns `Some(None)` if this `Ref` is a null `any` reference.
+    ///
+    /// Returns `Some(Some(_))` if this `Ref` is a non-null `any` reference.
+    #[inline]
+    pub fn as_any(&self) -> Option<Option<&Rooted<AnyRef>>> {
+        match self {
+            Ref::Any(e) => Some(e.as_ref()),
+            _ => None,
+        }
+    }
+
+    /// Get the underlying `any` reference, panicking if this is a different
+    /// kind of reference.
+    ///
+    /// Returns `None` if this `Ref` is a null `any` reference.
+    ///
+    /// Returns `Some(_)` if this `Ref` is a non-null `any` reference.
+    #[inline]
+    pub fn unwrap_any(&self) -> Option<&Rooted<AnyRef>> {
+        self.as_any().expect("Ref::unwrap_any on non-any reference")
+    }
+
     /// Is this a `func` reference?
     #[inline]
     pub fn is_func(&self) -> bool {
@@ -606,42 +771,56 @@ impl Ref {
     pub(crate) fn load_ty(&self, store: &StoreOpaque) -> RefType {
         assert!(self.comes_from_same_store(store));
         RefType::new(
-            false,
+            self.is_null(),
             match self {
                 Ref::Extern(_) => HeapType::Extern,
 
                 // NB: We choose the most-specific heap type we can here and let
                 // subtyping do its thing if callers are matching against a
                 // `HeapType::Func`.
-                Ref::Func(Some(f)) => HeapType::Concrete(f.load_ty(store)),
+                Ref::Func(Some(f)) => HeapType::ConcreteFunc(f.load_ty(store)),
                 Ref::Func(None) => HeapType::NoFunc,
+
+                Ref::Any(Some(_)) => {
+                    assert!(VMGcRef::ONLY_EXTERN_REF_AND_I31);
+                    HeapType::I31
+                }
+                Ref::Any(None) => HeapType::None,
             },
         )
     }
 
     /// Does this reference value match the given type?
     ///
+    /// Returns an error if the underlying `Rooted` has been unrooted.
+    ///
     /// # Panics
     ///
     /// Panics if this reference is not associated with the given store.
-    pub fn matches_ty(&self, store: impl AsContext, ty: &RefType) -> bool {
+    pub fn matches_ty(&self, store: impl AsContext, ty: &RefType) -> Result<bool> {
         self._matches_ty(&store.as_context().0, ty)
     }
 
-    pub(crate) fn _matches_ty(&self, store: &StoreOpaque, ty: &RefType) -> bool {
+    pub(crate) fn _matches_ty(&self, store: &StoreOpaque, ty: &RefType) -> Result<bool> {
         assert!(self.comes_from_same_store(store));
         assert!(ty.comes_from_same_engine(store.engine()));
         if self.is_null() && !ty.is_nullable() {
-            return false;
+            return Ok(false);
         }
-        match (self, ty.heap_type()) {
+        Ok(match (self, ty.heap_type()) {
             (Ref::Extern(_), HeapType::Extern) => true,
             (Ref::Extern(_), _) => false,
-            (Ref::Func(_), HeapType::Func | HeapType::NoFunc) => true,
-            (Ref::Func(None), HeapType::Concrete(_)) => true,
-            (Ref::Func(Some(f)), HeapType::Concrete(func_ty)) => f._matches_ty(store, func_ty),
+
+            (Ref::Func(_), HeapType::Func) => true,
+            (Ref::Func(None), HeapType::NoFunc | HeapType::ConcreteFunc(_)) => true,
+            (Ref::Func(Some(f)), HeapType::ConcreteFunc(func_ty)) => f._matches_ty(store, func_ty),
             (Ref::Func(_), _) => false,
-        }
+
+            (Ref::Any(_), HeapType::Any) => true,
+            (Ref::Any(Some(a)), HeapType::I31) => a._is_i31(store)?,
+            (Ref::Any(None), HeapType::None | HeapType::I31) => true,
+            (Ref::Any(_), _) => false,
+        })
     }
 
     pub(crate) fn ensure_matches_ty(&self, store: &StoreOpaque, ty: &RefType) -> Result<()> {
@@ -651,7 +830,7 @@ impl Ref {
         if !ty.comes_from_same_engine(store.engine()) {
             bail!("type used with wrong engine")
         }
-        if self._matches_ty(store, ty) {
+        if self._matches_ty(store, ty)? {
             Ok(())
         } else {
             let actual_ty = self.load_ty(store);
@@ -663,10 +842,10 @@ impl Ref {
         match self {
             Ref::Func(Some(f)) => f.comes_from_same_store(store),
             Ref::Func(None) => true,
-
-            // `ExternRef`s aren't associated with any single store right
-            // now. That may change in the future.
-            Ref::Extern(_) => true,
+            Ref::Extern(Some(x)) => x.comes_from_same_store(store),
+            Ref::Extern(None) => true,
+            Ref::Any(Some(a)) => a.comes_from_same_store(store),
+            Ref::Any(None) => true,
         }
     }
 
@@ -675,27 +854,43 @@ impl Ref {
         store: &mut StoreOpaque,
         ty: &RefType,
     ) -> Result<TableElement> {
-        self.ensure_matches_ty(store, &ty)
+        let mut store = AutoAssertNoGc::new(store);
+        self.ensure_matches_ty(&store, &ty)
             .context("type mismatch: value does not match table element type")?;
-        match (self, ty.heap_type()) {
-            (Ref::Func(None), HeapType::NoFunc | HeapType::Func | HeapType::Concrete(_)) => {
+
+        match (self, ty.heap_type().top()) {
+            (Ref::Func(None), HeapType::Func) => {
                 assert!(ty.is_nullable());
                 Ok(TableElement::FuncRef(ptr::null_mut()))
             }
-            (Ref::Func(Some(f)), HeapType::Func | HeapType::Concrete(_)) => {
+            (Ref::Func(Some(f)), HeapType::Func) => {
                 debug_assert!(
-                    f.comes_from_same_store(store),
+                    f.comes_from_same_store(&store),
                     "checked in `ensure_matches_ty`"
                 );
-                Ok(TableElement::FuncRef(f.vm_func_ref(store).as_ptr()))
+                Ok(TableElement::FuncRef(f.vm_func_ref(&mut store).as_ptr()))
             }
 
             (Ref::Extern(e), HeapType::Extern) => match e {
                 None => {
                     assert!(ty.is_nullable());
-                    Ok(TableElement::ExternRef(None))
+                    Ok(TableElement::GcRef(None))
+                }
+                Some(e) => {
+                    let gc_ref = e.try_clone_gc_ref(&mut store)?;
+                    Ok(TableElement::GcRef(Some(gc_ref)))
+                }
+            },
+
+            (Ref::Any(a), HeapType::Any) => match a {
+                None => {
+                    assert!(ty.is_nullable());
+                    Ok(TableElement::GcRef(None))
+                }
+                Some(a) => {
+                    let gc_ref = a.try_clone_gc_ref(&mut store)?;
+                    Ok(TableElement::GcRef(Some(gc_ref)))
                 }
-                Some(e) => Ok(TableElement::ExternRef(Some(e.into_vm_extern_ref()))),
             },
 
             _ => unreachable!("checked that the value matches the type above"),
@@ -749,7 +944,7 @@ mod tests {
         // Should panic.
         let _ = Val::FuncRef(Some(f)).matches_ty(
             &s1,
-            &ValType::Ref(RefType::new(true, HeapType::Concrete(t2))),
+            &ValType::Ref(RefType::new(true, HeapType::ConcreteFunc(t2))),
         );
     }
 
@@ -766,6 +961,6 @@ mod tests {
         let f = Func::new(&mut s1, t1.clone(), |_caller, _args, _results| Ok(()));
 
         // Should panic.
-        let _ = Ref::Func(Some(f)).matches_ty(&s1, &RefType::new(true, HeapType::Concrete(t2)));
+        let _ = Ref::Func(Some(f)).matches_ty(&s1, &RefType::new(true, HeapType::ConcreteFunc(t2)));
     }
 }
