@@ -5,13 +5,13 @@ use std::any::Any;
 use std::mem;
 use std::sync::Mutex;
 use wasmparser::FuncValidatorAllocations;
-use wasmtime_cranelift_shared::{CompiledFunction, ModuleTextBuilder};
+use wasmtime_cranelift::{CompiledFunction, ModuleTextBuilder};
 use wasmtime_environ::{
-    CompileError, DefinedFuncIndex, FilePos, FuncIndex, FunctionBodyData, FunctionLoc,
-    ModuleTranslation, ModuleTypesBuilder, PrimaryMap, TrapEncodingBuilder, VMOffsets,
-    WasmFunctionInfo,
+    AddressMapSection, BuiltinFunctionIndex, CompileError, DefinedFuncIndex, FunctionBodyData,
+    FunctionLoc, ModuleTranslation, ModuleTypesBuilder, PrimaryMap, RelocationTarget,
+    TrapEncodingBuilder, Tunables, VMOffsets, WasmFunctionInfo,
 };
-use winch_codegen::{BuiltinFunctions, TargetIsa, TrampolineKind};
+use winch_codegen::{BuiltinFunctions, TargetIsa};
 
 /// Function compilation context.
 /// This struct holds information that can be shared globally across
@@ -25,25 +25,22 @@ struct CompilationContext {
 
 pub(crate) struct Compiler {
     isa: Box<dyn TargetIsa>,
+    trampolines: Box<dyn wasmtime_environ::Compiler>,
     contexts: Mutex<Vec<CompilationContext>>,
-}
-
-/// The compiled function environment.
-pub struct CompiledFuncEnv;
-impl wasmtime_cranelift_shared::CompiledFuncEnv for CompiledFuncEnv {
-    fn resolve_user_external_name_ref(
-        &self,
-        external: cranelift_codegen::ir::UserExternalNameRef,
-    ) -> (u32, u32) {
-        (0, external.as_u32())
-    }
+    tunables: Tunables,
 }
 
 impl Compiler {
-    pub fn new(isa: Box<dyn TargetIsa>) -> Self {
+    pub fn new(
+        isa: Box<dyn TargetIsa>,
+        trampolines: Box<dyn wasmtime_environ::Compiler>,
+        tunables: Tunables,
+    ) -> Self {
         Self {
             isa,
+            trampolines,
             contexts: Mutex::new(Vec::new()),
+            tunables,
         }
     }
 
@@ -68,7 +65,7 @@ impl Compiler {
     /// Emit unwind info into the [`CompiledFunction`].
     fn emit_unwind_info(
         &self,
-        compiled_function: &mut CompiledFunction<CompiledFuncEnv>,
+        compiled_function: &mut CompiledFunction,
     ) -> Result<(), CompileError> {
         let kind = match self.isa.triple().operating_system {
             target_lexicon::OperatingSystem::Windows => UnwindInfoKind::Windows,
@@ -97,17 +94,13 @@ impl wasmtime_environ::Compiler for Compiler {
     ) -> Result<(WasmFunctionInfo, Box<dyn Any + Send>), CompileError> {
         let index = translation.module.func_index(index);
         let sig = translation.module.functions[index].signature;
-        let ty = &types[sig];
-        let FunctionBodyData { body, validator } = data;
-        let start_srcloc = FilePos::new(
-            body.get_binary_reader()
-                .original_position()
-                .try_into()
-                .unwrap(),
-        );
+        let ty = types[sig].unwrap_func();
+        let FunctionBodyData {
+            body, validator, ..
+        } = data;
         let mut context = self.get_context(translation);
         let mut validator = validator.into_validator(mem::take(&mut context.allocations));
-        let buffer = self
+        let func = self
             .isa
             .compile_function(
                 ty,
@@ -119,21 +112,25 @@ impl wasmtime_environ::Compiler for Compiler {
             )
             .map_err(|e| CompileError::Codegen(format!("{e:?}")));
         self.save_context(context, validator.into_allocations());
-        let buffer = buffer?;
+        let mut func = func?;
 
-        let mut compiled_function =
-            CompiledFunction::new(buffer, CompiledFuncEnv {}, self.isa.function_alignment());
+        let reader = body.get_binary_reader();
+        func.set_address_map(
+            reader.original_position() as u32,
+            reader.bytes_remaining() as u32,
+            self.tunables.generate_address_map,
+        );
 
         if self.isa.flags().unwind_info() {
-            self.emit_unwind_info(&mut compiled_function)?;
+            self.emit_unwind_info(&mut func)?;
         }
 
         Ok((
             WasmFunctionInfo {
-                start_srcloc,
+                start_srcloc: func.metadata().address_map.start_srcloc,
                 stack_maps: Box::new([]),
             },
-            Box::new(compiled_function),
+            Box::new(func),
         ))
     }
 
@@ -143,22 +140,8 @@ impl wasmtime_environ::Compiler for Compiler {
         types: &ModuleTypesBuilder,
         index: DefinedFuncIndex,
     ) -> Result<Box<dyn Any + Send>, CompileError> {
-        let func_index = translation.module.func_index(index);
-        let sig = translation.module.functions[func_index].signature;
-        let ty = &types[sig];
-        let buffer = self
-            .isa
-            .compile_trampoline(&ty, TrampolineKind::ArrayToWasm(func_index))
-            .map_err(|e| CompileError::Codegen(format!("{:?}", e)))?;
-
-        let mut compiled_function =
-            CompiledFunction::new(buffer, CompiledFuncEnv {}, self.isa.function_alignment());
-
-        if self.isa.flags().unwind_info() {
-            self.emit_unwind_info(&mut compiled_function)?;
-        }
-
-        Ok(Box::new(compiled_function))
+        self.trampolines
+            .compile_array_to_wasm_trampoline(translation, types, index)
     }
 
     fn compile_native_to_wasm_trampoline(
@@ -167,61 +150,37 @@ impl wasmtime_environ::Compiler for Compiler {
         types: &ModuleTypesBuilder,
         index: DefinedFuncIndex,
     ) -> Result<Box<dyn Any + Send>, CompileError> {
-        let func_index = translation.module.func_index(index);
-        let sig = translation.module.functions[func_index].signature;
-        let ty = &types[sig];
-
-        let buffer = self
-            .isa
-            .compile_trampoline(ty, TrampolineKind::NativeToWasm(func_index))
-            .map_err(|e| CompileError::Codegen(format!("{:?}", e)))?;
-
-        let mut compiled_function =
-            CompiledFunction::new(buffer, CompiledFuncEnv {}, self.isa.function_alignment());
-
-        if self.isa.flags().unwind_info() {
-            self.emit_unwind_info(&mut compiled_function)?;
-        }
-
-        Ok(Box::new(compiled_function))
+        self.trampolines
+            .compile_native_to_wasm_trampoline(translation, types, index)
     }
 
     fn compile_wasm_to_native_trampoline(
         &self,
         wasm_func_ty: &wasmtime_environ::WasmFuncType,
     ) -> Result<Box<dyn Any + Send>, CompileError> {
-        let buffer = self
-            .isa
-            .compile_trampoline(wasm_func_ty, TrampolineKind::WasmToNative)
-            .map_err(|e| CompileError::Codegen(format!("{:?}", e)))?;
-
-        let mut compiled_function =
-            CompiledFunction::new(buffer, CompiledFuncEnv {}, self.isa.function_alignment());
-
-        if self.isa.flags().unwind_info() {
-            self.emit_unwind_info(&mut compiled_function)?;
-        }
-
-        Ok(Box::new(compiled_function))
+        self.trampolines
+            .compile_wasm_to_native_trampoline(wasm_func_ty)
     }
 
     fn append_code(
         &self,
         obj: &mut Object<'static>,
         funcs: &[(String, Box<dyn Any + Send>)],
-        resolve_reloc: &dyn Fn(usize, FuncIndex) -> usize,
+        resolve_reloc: &dyn Fn(usize, wasmtime_environ::RelocationTarget) -> usize,
     ) -> Result<Vec<(SymbolId, FunctionLoc)>> {
         let mut builder =
             ModuleTextBuilder::new(obj, self, self.isa.text_section_builder(funcs.len()));
         let mut traps = TrapEncodingBuilder::default();
+        let mut addrs = AddressMapSection::default();
 
         let mut ret = Vec::with_capacity(funcs.len());
         for (i, (sym, func)) in funcs.iter().enumerate() {
-            let func = func
-                .downcast_ref::<CompiledFunction<CompiledFuncEnv>>()
-                .unwrap();
+            let func = func.downcast_ref::<CompiledFunction>().unwrap();
 
             let (sym, range) = builder.append_func(&sym, func, |idx| resolve_reloc(i, idx));
+            if self.tunables.generate_address_map {
+                addrs.push(range.clone(), &func.address_map().instructions);
+            }
             traps.push(range.clone(), &func.traps().collect::<Vec<_>>());
 
             let info = FunctionLoc {
@@ -231,6 +190,9 @@ impl wasmtime_environ::Compiler for Compiler {
             ret.push((sym, info));
         }
         builder.finish();
+        if self.tunables.generate_address_map {
+            addrs.append_to(obj);
+        }
         traps.append_to(obj);
         Ok(ret)
     }
@@ -252,11 +214,11 @@ impl wasmtime_environ::Compiler for Compiler {
     }
 
     fn flags(&self) -> Vec<(&'static str, wasmtime_environ::FlagValue<'static>)> {
-        wasmtime_cranelift_shared::clif_flags_to_wasmtime(self.isa.flags().iter())
+        wasmtime_cranelift::clif_flags_to_wasmtime(self.isa.flags().iter())
     }
 
     fn isa_flags(&self) -> Vec<(&'static str, wasmtime_environ::FlagValue<'static>)> {
-        wasmtime_cranelift_shared::clif_flags_to_wasmtime(self.isa.isa_flags())
+        wasmtime_cranelift::clif_flags_to_wasmtime(self.isa.isa_flags())
     }
 
     fn is_branch_protection_enabled(&self) -> bool {
@@ -265,7 +227,7 @@ impl wasmtime_environ::Compiler for Compiler {
 
     #[cfg(feature = "component-model")]
     fn component_compiler(&self) -> &dyn wasmtime_environ::component::ComponentCompiler {
-        todo!()
+        self.trampolines.component_compiler()
     }
 
     fn append_dwarf(
@@ -273,6 +235,8 @@ impl wasmtime_environ::Compiler for Compiler {
         _obj: &mut Object<'_>,
         _translation: &ModuleTranslation<'_>,
         _funcs: &PrimaryMap<DefinedFuncIndex, (SymbolId, &(dyn Any + Send))>,
+        _dwarf_package: Option<&[u8]>,
+        _tunables: &wasmtime_environ::Tunables,
     ) -> Result<()> {
         todo!()
     }
@@ -280,4 +244,18 @@ impl wasmtime_environ::Compiler for Compiler {
     fn create_systemv_cie(&self) -> Option<gimli::write::CommonInformationEntry> {
         self.isa.create_systemv_cie()
     }
+
+    fn compile_wasm_to_builtin(
+        &self,
+        index: BuiltinFunctionIndex,
+    ) -> Result<Box<dyn Any + Send>, CompileError> {
+        self.trampolines.compile_wasm_to_builtin(index)
+    }
+
+    fn compiled_function_relocation_targets<'a>(
+        &'a self,
+        func: &'a dyn Any,
+    ) -> Box<dyn Iterator<Item = RelocationTarget> + 'a> {
+        self.trampolines.compiled_function_relocation_targets(func)
+    }
 }
