@@ -39,8 +39,8 @@ pub trait RustGenerator<'a> {
             Type::S16 => self.push_str("i16"),
             Type::S32 => self.push_str("i32"),
             Type::S64 => self.push_str("i64"),
-            Type::Float32 => self.push_str("f32"),
-            Type::Float64 => self.push_str("f64"),
+            Type::F32 => self.push_str("f32"),
+            Type::F64 => self.push_str("f64"),
             Type::Char => self.push_str("char"),
             Type::String => match mode {
                 TypeMode::AllBorrowed(lt) => {
@@ -51,7 +51,7 @@ pub trait RustGenerator<'a> {
                     }
                     self.push_str("str");
                 }
-                TypeMode::Owned => self.push_str("String"),
+                TypeMode::Owned => self.push_str("wasmtime::component::__internal::String"),
             },
         }
     }
@@ -212,7 +212,7 @@ pub trait RustGenerator<'a> {
                 self.push_str("]");
             }
             TypeMode::Owned => {
-                self.push_str("Vec<");
+                self.push_str("wasmtime::component::__internal::Vec<");
                 self.print_ty(ty, next_mode);
                 self.push_str(">");
             }
