@@ -43,7 +43,7 @@ pub struct CompileCommand {
     #[arg(long, value_name = "TARGET")]
     pub target: Option<String>,
 
-    /// The path of the output compiled module; defaults to <MODULE>.cwasm
+    /// The path of the output compiled module; defaults to `<MODULE>.cwasm`
     #[arg(short = 'o', long, value_name = "OUTPUT")]
     pub output: Option<PathBuf>,
 
@@ -61,7 +61,7 @@ impl CompileCommand {
     pub fn execute(mut self) -> Result<()> {
         self.common.init_logging()?;
 
-        let mut config = self.common.config(self.target.as_deref())?;
+        let mut config = self.common.config(self.target.as_deref(), None)?;
 
         if let Some(path) = self.emit_clif {
             if !path.exists() {
