@@ -95,10 +95,10 @@ fn call_wrapped_func() -> Result<(), Error> {
 
         unsafe {
             let mut args = [
-                Val::I32(1).to_raw(&mut store),
-                Val::I64(2).to_raw(&mut store),
-                Val::F32(3.0f32.to_bits()).to_raw(&mut store),
-                Val::F64(4.0f64.to_bits()).to_raw(&mut store),
+                Val::I32(1).to_raw(&mut store)?,
+                Val::I64(2).to_raw(&mut store)?,
+                Val::F32(3.0f32.to_bits()).to_raw(&mut store)?,
+                Val::F64(4.0f64.to_bits()).to_raw(&mut store)?,
             ];
             f.call_unchecked(&mut store, args.as_mut_ptr(), args.len())?;
         }
