@@ -1,4 +1,5 @@
 use anyhow::bail;
+use std::sync::atomic::Ordering;
 use std::sync::atomic::{AtomicBool, AtomicUsize, Ordering::SeqCst};
 use std::sync::Arc;
 use wasmtime::*;
@@ -385,8 +386,16 @@ fn func_constructors() {
     Func::wrap(&mut store, || -> i64 { 0 });
     Func::wrap(&mut store, || -> f32 { 0.0 });
     Func::wrap(&mut store, || -> f64 { 0.0 });
-    Func::wrap(&mut store, || -> ExternRef { loop {} });
-    Func::wrap(&mut store, || -> Option<ExternRef> { None });
+    Func::wrap(&mut store, || -> Rooted<ExternRef> { loop {} });
+    Func::wrap(&mut store, || -> Option<Rooted<ExternRef>> { None });
+    Func::wrap(&mut store, || -> ManuallyRooted<ExternRef> { loop {} });
+    Func::wrap(&mut store, || -> Option<ManuallyRooted<ExternRef>> { None });
+    Func::wrap(&mut store, || -> Rooted<AnyRef> { loop {} });
+    Func::wrap(&mut store, || -> Option<Rooted<AnyRef>> { None });
+    Func::wrap(&mut store, || -> ManuallyRooted<AnyRef> { loop {} });
+    Func::wrap(&mut store, || -> Option<ManuallyRooted<AnyRef>> { None });
+    Func::wrap(&mut store, || -> I31 { loop {} });
+    Func::wrap(&mut store, || -> Option<I31> { None });
     Func::wrap(&mut store, || -> Func { loop {} });
     Func::wrap(&mut store, || -> Option<Func> { None });
     Func::wrap(&mut store, || -> NoFunc { loop {} });
@@ -397,8 +406,25 @@ fn func_constructors() {
     Func::wrap(&mut store, || -> Result<i64> { loop {} });
     Func::wrap(&mut store, || -> Result<f32> { loop {} });
     Func::wrap(&mut store, || -> Result<f64> { loop {} });
-    Func::wrap(&mut store, || -> Result<ExternRef> { loop {} });
-    Func::wrap(&mut store, || -> Result<Option<ExternRef>> { loop {} });
+    Func::wrap(&mut store, || -> Result<Rooted<ExternRef>> { loop {} });
+    Func::wrap(&mut store, || -> Result<Option<Rooted<ExternRef>>> {
+        loop {}
+    });
+    Func::wrap(&mut store, || -> Result<ManuallyRooted<ExternRef>> {
+        loop {}
+    });
+    Func::wrap(
+        &mut store,
+        || -> Result<Option<ManuallyRooted<ExternRef>>> { loop {} },
+    );
+    Func::wrap(&mut store, || -> Result<Rooted<AnyRef>> { loop {} });
+    Func::wrap(&mut store, || -> Result<Option<Rooted<AnyRef>>> { loop {} });
+    Func::wrap(&mut store, || -> Result<ManuallyRooted<AnyRef>> { loop {} });
+    Func::wrap(&mut store, || -> Result<Option<ManuallyRooted<AnyRef>>> {
+        loop {}
+    });
+    Func::wrap(&mut store, || -> Result<I31> { loop {} });
+    Func::wrap(&mut store, || -> Result<Option<I31>> { loop {} });
     Func::wrap(&mut store, || -> Result<Func> { loop {} });
     Func::wrap(&mut store, || -> Result<Option<Func>> { loop {} });
     Func::wrap(&mut store, || -> Result<NoFunc> { loop {} });
@@ -485,19 +511,30 @@ fn signatures_match() {
 
     let f = Func::wrap(
         &mut store,
-        |_: f32, _: f64, _: i32, _: i64, _: i32, _: Option<ExternRef>, _: Option<Func>| -> f64 {
-            loop {}
-        },
+        |_: f32,
+         _: f64,
+         _: i32,
+         _: i64,
+         _: i32,
+         _: Option<Rooted<ExternRef>>,
+         _: Option<ManuallyRooted<ExternRef>>,
+         _: Option<Rooted<AnyRef>>,
+         _: Option<ManuallyRooted<AnyRef>>,
+         _: Option<Func>|
+         -> f64 { loop {} },
     );
 
-    assert_eq!(f.ty(&store).params().len(), 7);
+    assert_eq!(f.ty(&store).params().len(), 10);
     assert!(f.ty(&store).params().nth(0).unwrap().is_f32());
     assert!(f.ty(&store).params().nth(1).unwrap().is_f64());
     assert!(f.ty(&store).params().nth(2).unwrap().is_i32());
     assert!(f.ty(&store).params().nth(3).unwrap().is_i64());
     assert!(f.ty(&store).params().nth(4).unwrap().is_i32());
     assert!(f.ty(&store).params().nth(5).unwrap().is_externref());
-    assert!(f.ty(&store).params().nth(6).unwrap().is_funcref());
+    assert!(f.ty(&store).params().nth(6).unwrap().is_externref());
+    assert!(f.ty(&store).params().nth(7).unwrap().is_anyref());
+    assert!(f.ty(&store).params().nth(8).unwrap().is_anyref());
+    assert!(f.ty(&store).params().nth(9).unwrap().is_funcref());
 
     assert_eq!(f.ty(&store).results().len(), 1);
     assert!(f.ty(&store).results().nth(0).unwrap().is_f64());
@@ -506,6 +543,8 @@ fn signatures_match() {
 #[test]
 #[cfg_attr(miri, ignore)]
 fn import_works() -> Result<()> {
+    let _ = env_logger::try_init();
+
     static HITS: AtomicUsize = AtomicUsize::new(0);
 
     let wasm = wat::parse_str(
@@ -513,9 +552,9 @@ fn import_works() -> Result<()> {
             (import "" "" (func))
             (import "" "" (func (param i32) (result i32)))
             (import "" "" (func (param i32) (param i64)))
-            (import "" "" (func (param i32 i64 i32 f32 f64 externref funcref)))
+            (import "" "" (func (param i32 i64 i32 f32 f64 externref externref funcref anyref anyref i31ref)))
 
-            (func (export "run") (param externref funcref)
+            (func (export "run") (param externref externref funcref)
                 call 0
                 i32.const 0
                 call 1
@@ -531,15 +570,24 @@ fn import_works() -> Result<()> {
                 f64.const 500
                 local.get 0
                 local.get 1
+                local.get 2
+                (ref.i31 (i32.const 36))
+                (ref.i31 (i32.const 42))
+                (ref.i31 (i32.const 0x1234))
                 call 3
             )
         "#,
     )?;
+
     let mut config = Config::new();
     config.wasm_reference_types(true);
+    config.wasm_function_references(true);
+    config.wasm_gc(true);
+
     let engine = Engine::new(&config)?;
     let mut store = Store::new(&engine, ());
     let module = Module::new(&engine, &wasm)?;
+
     let imports = [
         Func::wrap(&mut store, || {
             assert_eq!(HITS.fetch_add(1, SeqCst), 0);
@@ -565,39 +613,63 @@ fn import_works() -> Result<()> {
              c: i32,
              d: f32,
              e: f64,
-             f: Option<ExternRef>,
-             g: Option<Func>| {
+             f: Option<Rooted<ExternRef>>,
+             g: Option<ManuallyRooted<ExternRef>>,
+             h: Option<Func>,
+             i: Option<Rooted<AnyRef>>,
+             j: Option<ManuallyRooted<AnyRef>>,
+             k: Option<I31>|
+             -> Result<()> {
                 assert_eq!(a, 100);
                 assert_eq!(b, 200);
                 assert_eq!(c, 300);
                 assert_eq!(d, 400.0);
                 assert_eq!(e, 500.0);
                 assert_eq!(
-                    f.as_ref().unwrap().data().downcast_ref::<String>().unwrap(),
+                    f.as_ref()
+                        .unwrap()
+                        .data(&caller)?
+                        .downcast_ref::<String>()
+                        .unwrap(),
                     "hello"
                 );
+                assert_eq!(
+                    g.as_ref()
+                        .unwrap()
+                        .data(&caller)?
+                        .downcast_ref::<String>()
+                        .unwrap(),
+                    "goodbye"
+                );
+                assert_eq!(
+                    i.unwrap().as_i31(&caller).unwrap().unwrap(),
+                    I31::wrapping_u32(36)
+                );
+                assert_eq!(
+                    j.unwrap().as_i31(&caller).unwrap().unwrap(),
+                    I31::wrapping_u32(42)
+                );
+                assert_eq!(k, Some(I31::wrapping_u32(0x1234)));
                 let mut results = [Val::I32(0)];
-                g.as_ref()
+                h.as_ref()
                     .unwrap()
                     .call(&mut caller, &[], &mut results)
                     .unwrap();
                 assert_eq!(results[0].unwrap_i32(), 42);
                 assert_eq!(HITS.fetch_add(1, SeqCst), 3);
+                Ok(())
             },
         )
         .into(),
     ];
+
     let instance = Instance::new(&mut store, &module, &imports)?;
     let run = instance.get_func(&mut store, "run").unwrap();
+    let hello = Val::ExternRef(Some(ExternRef::new(&mut store, "hello".to_string())?));
+    let goodbye = Val::ExternRef(Some(ExternRef::new(&mut store, "goodbye".to_string())?));
     let funcref = Val::FuncRef(Some(Func::wrap(&mut store, || -> i32 { 42 })));
-    run.call(
-        &mut store,
-        &[
-            Val::ExternRef(Some(ExternRef::new("hello".to_string()))),
-            funcref,
-        ],
-        &mut [],
-    )?;
+    run.call(&mut store, &[hello, goodbye, funcref], &mut [])?;
+
     assert_eq!(HITS.load(SeqCst), 4);
     Ok(())
 }
@@ -646,8 +718,34 @@ fn get_from_wrapper() {
     assert!(f.typed::<(), f32>(&store).is_ok());
     let f = Func::wrap(&mut store, || -> f64 { loop {} });
     assert!(f.typed::<(), f64>(&store).is_ok());
-    let f = Func::wrap(&mut store, || -> Option<ExternRef> { loop {} });
-    assert!(f.typed::<(), Option<ExternRef>>(&store).is_ok());
+    let f = Func::wrap(&mut store, || -> Rooted<ExternRef> { loop {} });
+    assert!(f.typed::<(), Rooted<ExternRef>>(&store).is_ok());
+    let f = Func::wrap(&mut store, || -> Option<Rooted<ExternRef>> { loop {} });
+    assert!(f.typed::<(), Option<Rooted<ExternRef>>>(&store).is_ok());
+    let f = Func::wrap(&mut store, || -> ManuallyRooted<ExternRef> { loop {} });
+    assert!(f.typed::<(), ManuallyRooted<ExternRef>>(&store).is_ok());
+    let f = Func::wrap(&mut store, || -> Option<ManuallyRooted<ExternRef>> {
+        loop {}
+    });
+    assert!(f
+        .typed::<(), Option<ManuallyRooted<ExternRef>>>(&store)
+        .is_ok());
+    let f = Func::wrap(&mut store, || -> Rooted<AnyRef> { loop {} });
+    assert!(f.typed::<(), Rooted<AnyRef>>(&store).is_ok());
+    let f = Func::wrap(&mut store, || -> Option<Rooted<AnyRef>> { loop {} });
+    assert!(f.typed::<(), Option<Rooted<AnyRef>>>(&store).is_ok());
+    let f = Func::wrap(&mut store, || -> ManuallyRooted<AnyRef> { loop {} });
+    assert!(f.typed::<(), ManuallyRooted<AnyRef>>(&store).is_ok());
+    let f = Func::wrap(&mut store, || -> Option<ManuallyRooted<AnyRef>> { loop {} });
+    assert!(f
+        .typed::<(), Option<ManuallyRooted<AnyRef>>>(&store)
+        .is_ok());
+    let f = Func::wrap(&mut store, || -> I31 { loop {} });
+    assert!(f.typed::<(), I31>(&store).is_ok());
+    let f = Func::wrap(&mut store, || -> Option<I31> { loop {} });
+    assert!(f.typed::<(), Option<I31>>(&store).is_ok());
+    let f = Func::wrap(&mut store, || -> Func { loop {} });
+    assert!(f.typed::<(), Func>(&store).is_ok());
     let f = Func::wrap(&mut store, || -> Option<Func> { loop {} });
     assert!(f.typed::<(), Option<Func>>(&store).is_ok());
 
@@ -662,8 +760,32 @@ fn get_from_wrapper() {
     assert!(f.typed::<f32, ()>(&store).is_ok());
     let f = Func::wrap(&mut store, |_: f64| {});
     assert!(f.typed::<f64, ()>(&store).is_ok());
-    let f = Func::wrap(&mut store, |_: Option<ExternRef>| {});
-    assert!(f.typed::<Option<ExternRef>, ()>(&store).is_ok());
+    let f = Func::wrap(&mut store, |_: Rooted<ExternRef>| {});
+    assert!(f.typed::<Rooted<ExternRef>, ()>(&store).is_ok());
+    let f = Func::wrap(&mut store, |_: Option<Rooted<ExternRef>>| {});
+    assert!(f.typed::<Option<Rooted<ExternRef>>, ()>(&store).is_ok());
+    let f = Func::wrap(&mut store, |_: ManuallyRooted<ExternRef>| {});
+    assert!(f.typed::<ManuallyRooted<ExternRef>, ()>(&store).is_ok());
+    let f = Func::wrap(&mut store, |_: Option<ManuallyRooted<ExternRef>>| {});
+    assert!(f
+        .typed::<Option<ManuallyRooted<ExternRef>>, ()>(&store)
+        .is_ok());
+    let f = Func::wrap(&mut store, |_: Rooted<AnyRef>| {});
+    assert!(f.typed::<Rooted<AnyRef>, ()>(&store).is_ok());
+    let f = Func::wrap(&mut store, |_: Option<Rooted<AnyRef>>| {});
+    assert!(f.typed::<Option<Rooted<AnyRef>>, ()>(&store).is_ok());
+    let f = Func::wrap(&mut store, |_: ManuallyRooted<AnyRef>| {});
+    assert!(f.typed::<ManuallyRooted<AnyRef>, ()>(&store).is_ok());
+    let f = Func::wrap(&mut store, |_: Option<ManuallyRooted<AnyRef>>| {});
+    assert!(f
+        .typed::<Option<ManuallyRooted<AnyRef>>, ()>(&store)
+        .is_ok());
+    let f = Func::wrap(&mut store, |_: I31| {});
+    assert!(f.typed::<I31, ()>(&store).is_ok());
+    let f = Func::wrap(&mut store, |_: Option<I31>| {});
+    assert!(f.typed::<Option<I31>, ()>(&store).is_ok());
+    let f = Func::wrap(&mut store, |_: Func| {});
+    assert!(f.typed::<Func, ()>(&store).is_ok());
     let f = Func::wrap(&mut store, |_: Option<Func>| {});
     assert!(f.typed::<Option<Func>, ()>(&store).is_ok());
 }
@@ -1341,12 +1463,12 @@ fn calls_with_funcref_and_externref() -> anyhow::Result<()> {
     linker.func_wrap(
         "",
         "witness",
-        |mut caller: Caller<'_, ()>, func: Option<Func>, externref: Option<ExternRef>| {
+        |mut caller: Caller<'_, ()>, func: Option<Func>, externref: Option<Rooted<ExternRef>>| {
             if func.is_some() {
                 assert_my_funcref(&mut caller, func.as_ref())?;
             }
             if externref.is_some() {
-                assert_my_externref(externref.as_ref());
+                assert_my_externref(&caller, externref);
             }
             Ok(())
         },
@@ -1354,13 +1476,13 @@ fn calls_with_funcref_and_externref() -> anyhow::Result<()> {
     let instance = linker.instantiate(&mut store, &module)?;
 
     let typed = instance
-        .get_typed_func::<(Option<Func>, Option<ExternRef>), (Option<ExternRef>, Option<Func>)>(
+        .get_typed_func::<(Option<Func>, Option<Rooted<ExternRef>>), (Option<Rooted<ExternRef>>, Option<Func>)>(
             &mut store, "f",
         )?;
     let untyped = typed.func();
 
     let my_funcref = Func::wrap(&mut store, || 100u32);
-    let my_externref = ExternRef::new(99u32);
+    let my_externref = ExternRef::new(&mut store, 99u32)?;
     let mut results = [Val::I32(0), Val::I32(0)];
 
     fn assert_my_funcref(mut store: impl AsContextMut, func: Option<&Func>) -> Result<()> {
@@ -1369,8 +1491,11 @@ fn calls_with_funcref_and_externref() -> anyhow::Result<()> {
         assert_eq!(func.typed::<(), u32>(&store)?.call(&mut store, ())?, 100);
         Ok(())
     }
-    fn assert_my_externref(externref: Option<&ExternRef>) {
-        assert_eq!(externref.unwrap().data().downcast_ref(), Some(&99u32));
+    fn assert_my_externref(store: impl AsContext, externref: Option<Rooted<ExternRef>>) {
+        assert_eq!(
+            externref.unwrap().data(&store).unwrap().downcast_ref(),
+            Some(&99u32)
+        );
     }
 
     // funcref=null, externref=null
@@ -1399,7 +1524,7 @@ fn calls_with_funcref_and_externref() -> anyhow::Result<()> {
 
     // funcref=null, externref=Some
     let (a, b) = typed.call(&mut store, (None, Some(my_externref.clone())))?;
-    assert_my_externref(a.as_ref());
+    assert_my_externref(&store, a);
     assert!(b.is_none());
     untyped.call(
         &mut store,
@@ -1409,12 +1534,12 @@ fn calls_with_funcref_and_externref() -> anyhow::Result<()> {
         ],
         &mut results,
     )?;
-    assert_my_externref(results[0].unwrap_externref());
+    assert_my_externref(&store, results[0].unwrap_externref().copied());
     assert!(results[1].unwrap_funcref().is_none());
 
     // funcref=Some, externref=Some
     let (a, b) = typed.call(&mut store, (Some(my_funcref), Some(my_externref.clone())))?;
-    assert_my_externref(a.as_ref());
+    assert_my_externref(&store, a);
     assert_my_funcref(&mut store, b.as_ref())?;
     untyped.call(
         &mut store,
@@ -1424,7 +1549,7 @@ fn calls_with_funcref_and_externref() -> anyhow::Result<()> {
         ],
         &mut results,
     )?;
-    assert_my_externref(results[0].unwrap_externref());
+    assert_my_externref(&store, results[0].unwrap_externref().copied());
     assert_my_funcref(&mut store, results[1].unwrap_funcref())?;
 
     Ok(())
@@ -1475,7 +1600,7 @@ fn typed_concrete_param() -> anyhow::Result<()> {
     assert!(e.contains("argument type mismatch for reference to concrete type"));
     assert!(e.contains(
         "type mismatch: expected (type (func)), \
-         found (type (func (param (ref null (concrete VMSharedTypeIndex(0))))))"
+         found (type (func (param (ref null (concrete func VMSharedTypeIndex(0))))))"
     ));
 
     // And dynamic checks also work with a non-nullable super type.
@@ -1489,12 +1614,15 @@ fn typed_concrete_param() -> anyhow::Result<()> {
     assert!(e.contains("argument type mismatch for reference to concrete type"));
     assert!(e.contains(
         "type mismatch: expected (type (func)), \
-         found (type (func (param (ref null (concrete VMSharedTypeIndex(0))))))"
+         found (type (func (param (ref null (concrete func VMSharedTypeIndex(0))))))"
     ));
 
     // Calling `typed` with a type that is not a supertype nor a subtype fails
     // the initial type check.
-    let e = f.typed::<Option<ExternRef>, ()>(&store).err().unwrap();
+    let e = f
+        .typed::<Option<Rooted<ExternRef>>, ()>(&store)
+        .err()
+        .unwrap();
     let e = format!("{e:?}");
     assert!(e.contains("type mismatch with parameters"));
     assert!(e.contains("type mismatch: expected func, found extern"));
@@ -1543,15 +1671,15 @@ fn typed_concrete_result() -> anyhow::Result<()> {
     let e = format!("{e:?}");
     assert!(e.contains("type mismatch with results"));
     assert!(e.contains(
-        "type mismatch: expected (ref nofunc), found (ref (concrete VMSharedTypeIndex(0)))"
+        "type mismatch: expected (ref nofunc), found (ref (concrete func VMSharedTypeIndex(0)))"
     ));
 
     // Nor some unrelated type that it is neither a subtype or supertype of.
-    let e = f.typed::<(), ExternRef>(&store).err().unwrap();
+    let e = f.typed::<(), Rooted<ExternRef>>(&store).err().unwrap();
     let e = format!("{e:?}");
     assert!(e.contains("type mismatch with results"));
     assert!(e.contains(
-        "type mismatch: expected (ref extern), found (ref (concrete VMSharedTypeIndex(0)))"
+        "type mismatch: expected (ref extern), found (ref (concrete func VMSharedTypeIndex(0)))"
     ));
 
     Ok(())
@@ -1645,7 +1773,10 @@ fn call_wasm_passing_subtype_func_param() -> anyhow::Result<()> {
     let h_ty = FuncType::new(
         &engine,
         None,
-        Some(ValType::Ref(RefType::new(true, HeapType::Concrete(g_ty)))),
+        Some(ValType::Ref(RefType::new(
+            true,
+            HeapType::ConcreteFunc(g_ty),
+        ))),
     );
     let h = Func::new(&mut store, h_ty, move |_caller, _params, results| {
         results[0] = Val::FuncRef(Some(g.clone()));
@@ -1929,3 +2060,75 @@ fn typed_v128_imports() -> anyhow::Result<()> {
 
     Ok(())
 }
+
+#[test]
+#[cfg_attr(miri, ignore)]
+fn wrap_and_typed_i31ref() -> Result<()> {
+    let mut config = Config::new();
+    config.wasm_function_references(true);
+    config.wasm_gc(true);
+
+    let engine = Engine::new(&config)?;
+    let mut store = Store::new(&engine, ());
+
+    static HITS: AtomicUsize = AtomicUsize::new(0);
+    let mut linker = Linker::new(&engine);
+    linker.func_wrap("env", "i31ref", |x: Option<I31>| -> Option<I31> {
+        assert_eq!(HITS.fetch_add(1, Ordering::SeqCst), 0);
+        x
+    })?;
+    linker.func_wrap("env", "ref-i31", |x: I31| -> I31 {
+        assert_eq!(HITS.fetch_add(1, Ordering::SeqCst), 1);
+        x
+    })?;
+
+    let module = Module::new(
+        &engine,
+        r#"
+            (module
+                (import "env" "i31ref" (func (param i31ref) (result i31ref)))
+                (import "env" "ref-i31" (func (param (ref i31)) (result (ref i31))))
+
+                (func (export "i31ref") (param i31ref) (result i31ref)
+                    local.get 0
+                    call 0
+                )
+
+                (func (export "ref-i31") (param (ref i31)) (result (ref i31))
+                    local.get 0
+                    call 1
+                )
+            )
+        "#,
+    )?;
+
+    let instance = linker.instantiate(&mut store, &module)?;
+
+    let i31ref = instance.get_typed_func::<Option<I31>, Option<I31>>(&mut store, "i31ref")?;
+    let x = i31ref.call(&mut store, Some(I31::wrapping_u32(42)))?;
+    assert_eq!(x, Some(I31::wrapping_u32(42)));
+
+    let ref_i31 = instance.get_typed_func::<I31, I31>(&mut store, "ref-i31")?;
+    let x = ref_i31.call(&mut store, I31::wrapping_u32(0x1234))?;
+    assert_eq!(x, I31::wrapping_u32(0x1234));
+
+    assert_eq!(HITS.load(Ordering::SeqCst), 2);
+    Ok(())
+}
+
+#[test]
+fn call_func_with_funcref_both_typed_and_untyped() -> Result<()> {
+    let mut config = Config::new();
+    config.wasm_function_references(true);
+    config.wasm_gc(true);
+
+    let engine = Engine::new(&config)?;
+    let mut store = Store::new(&engine, ());
+
+    let f1 = Func::wrap(&mut store, |_: Option<Func>| {});
+    let f2 = Func::wrap(&mut store, || {});
+
+    f1.typed::<Func, ()>(&mut store)?.call(&mut store, f2)?;
+    f1.call(&mut store, &[Val::FuncRef(Some(f2))], &mut [])?;
+    Ok(())
+}
