@@ -7,6 +7,7 @@ mod code_too_large;
 mod component_model;
 mod coredump;
 mod debug;
+mod defaults;
 mod epoch_interruption;
 mod externals;
 mod fuel;
@@ -36,6 +37,7 @@ mod store;
 mod table;
 mod threads;
 mod traps;
+mod types;
 mod wait_notify;
 mod wasi_testsuite;
 mod wast;
@@ -96,7 +98,7 @@ pub(crate) fn small_pool_config() -> wasmtime::PoolingAllocationConfig {
     // When testing, we may choose to start with MPK force-enabled to ensure
     // we use that functionality.
     if std::env::var("WASMTIME_TEST_FORCE_MPK").is_ok() {
-        config.memory_protection_keys(wasmtime_runtime::MpkEnabled::Enable);
+        config.memory_protection_keys(wasmtime::MpkEnabled::Enable);
     }
 
     #[cfg(feature = "async")]
