@@ -1,8 +1,7 @@
 use rayon::prelude::*;
 use std::sync::atomic::{AtomicU32, Ordering::SeqCst};
-use std::time::{Duration, Instant};
+use std::time::Duration;
 use wasmtime::*;
-use wasmtime_runtime::mpk;
 
 fn module(engine: &Engine) -> Result<Module> {
     let mut wat = format!("(module\n");
@@ -247,7 +246,7 @@ fn guards_present_pooling() -> Result<()> {
 #[test]
 #[cfg_attr(miri, ignore)]
 fn guards_present_pooling_mpk() -> Result<()> {
-    if !mpk::is_supported() {
+    if !wasmtime::PoolingAllocationConfig::are_memory_protection_keys_available() {
         println!("skipping `guards_present_pooling_mpk` test; mpk is not supported");
         return Ok(());
     }
@@ -588,7 +587,7 @@ fn shared_memory_basics() -> Result<()> {
     assert_eq!(memory.atomic_wait64(8, 1, None), Ok(WaitResult::Mismatch));
 
     // timeout
-    let near_future = Instant::now() + Duration::new(0, 100);
+    let near_future = Duration::new(0, 100);
     assert_eq!(
         memory.atomic_wait32(8, 0, Some(near_future)),
         Ok(WaitResult::TimedOut)
@@ -609,8 +608,8 @@ fn shared_memory_wait_notify() -> Result<()> {
 
     let engine = Engine::default();
     let memory = SharedMemory::new(&engine, MemoryType::shared(1, 1))?;
-    let data = unsafe { &*(memory.data().as_ptr() as *const AtomicU32) };
-    let locked = unsafe { &*(memory.data().as_ptr().add(4) as *const AtomicU32) };
+    let data = unsafe { AtomicU32::from_ptr(memory.data().as_ptr().cast_mut().cast()) };
+    let locked = unsafe { AtomicU32::from_ptr(memory.data().as_ptr().add(4).cast_mut().cast()) };
 
     // Note that `SeqCst` is used here to not think much about the orderings
     // here, and it also somewhat more closely mirrors what's happening in wasm.
