@@ -45,7 +45,9 @@ fn custom_limiter_detect_os_oom_failure() -> Result<()> {
 
     // Default behavior of on-demand memory allocation so that a
     // memory grow will hit Linux for a larger mmap.
-    let engine = Engine::default();
+    let mut config = Config::new();
+    config.wasm_reference_types(false);
+    let engine = Engine::new(&config)?;
     let linker = Linker::new(&engine);
     let module = Module::new(&engine, r#"(module (memory (export "m") 0))"#).unwrap();
 
