@@ -2,9 +2,7 @@ use super::regs;
 use crate::{
     abi::{align_to, ABIOperand, ABIParams, ABIResults, ABISig, ParamsOrReturns, ABI},
     isa::{reg::Reg, CallingConvention},
-    masm::OperandSize,
 };
-use smallvec::SmallVec;
 use wasmtime_environ::{WasmHeapType, WasmValType};
 
 /// Helper environment to track argument-register
@@ -86,18 +84,6 @@ impl ABI for X64ABI {
         16
     }
 
-    fn ret_addr_offset() -> u8 {
-        // 1 8-byte slot.
-        // ┌──────────┬
-        // │   Ret    │
-        // │   Addr   │
-        // ├──────────┼ * offset
-        // │          │
-        // │   FP     │
-        // └──────────┴
-        8
-    }
-
     fn word_bits() -> u8 {
         64
     }
@@ -163,22 +149,10 @@ impl ABI for X64ABI {
         regs::scratch_xmm()
     }
 
-    fn fp_reg() -> Reg {
-        regs::rbp()
-    }
-
-    fn sp_reg() -> Reg {
-        regs::rsp()
-    }
-
     fn vmctx_reg() -> Reg {
         regs::vmctx()
     }
 
-    fn callee_saved_regs(call_conv: &CallingConvention) -> SmallVec<[(Reg, OperandSize); 18]> {
-        regs::callee_saved(call_conv)
-    }
-
     fn stack_slot_size() -> u8 {
         Self::word_bytes()
     }
